"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const configuration_1 = require("@atomist/automation-client/lib/configuration");
const ApolloGraphClient_1 = require("@atomist/automation-client/lib/graph/ApolloGraphClient");
const Deferred_1 = require("@atomist/automation-client/lib/internal/util/Deferred");
const port_1 = require("@atomist/automation-client/lib/util/port");
// tslint:disable-next-line:import-blacklist
const axios_1 = require("axios");
const chalk_1 = require("chalk");
const express = require("express");
const inquirer = require("inquirer");
const js_sha256_1 = require("js-sha256");
const _ = require("lodash");
const opn = require("open");
const os = require("os");
const cliConfig_1 = require("./cliConfig");
const print = require("./print");
const UserQuery = `query User {
  user {
    principal {
      sub
      pid
    }
  }
}`;
const PersonByIdentityQuery = `query PersonByIdentity {
  personByIdentity {
    team {
      id
      name
    }
  }
}`;
const CreateApiKeyMutation = `mutation createKey($description: String!){
  createApiKey(description: $description) {
    id
    key
    description
    createdAt
    lastUsed
    owner {
      id
    }
  }
}`;
/**
 * Set up local configuration with Atomist api key and workspaces
 * @param opts
 */
function config(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const cfgPath = configuration_1.userConfigPath();
        const userCfg = cliConfig_1.resolveUserConfig();
        const defaultCfg = configuration_1.defaultConfiguration();
        const cfg = configuration_1.mergeConfigs(defaultCfg, userCfg);
        let apiKey = opts.createApiKey !== true ? (opts.apiKey || cfg.apiKey) : undefined;
        // No api key; config and create a new key
        if (!apiKey) {
            try {
                apiKey = yield createApiKey(cfg);
            }
            catch (e) {
                print.error(`Failed to create API key: ${e.message}`);
                return 1;
            }
        }
        try {
            // Validate api key
            yield validateApiKey(apiKey, cfg);
        }
        catch (e) {
            print.error(`Failed to validate API key: ${e.message}`);
            return 1;
        }
        userCfg.apiKey = apiKey;
        yield configuration_1.writeUserConfig(userCfg);
        let workspaceIds;
        if (!opts.workspaceId) {
            try {
                // Retrieve list and config workspaces
                workspaceIds = yield configureWorkspaces(apiKey, cfg);
            }
            catch (e) {
                print.error(`Failed to configure workspaces: ${e.message}`);
                return 1;
            }
        }
        else {
            workspaceIds = opts.workspaceId.split(/\s+/);
        }
        userCfg.workspaceIds = workspaceIds;
        yield configuration_1.writeUserConfig(userCfg);
        print.log(`Successfully wrote configuration ${chalk_1.default.green(cfgPath)}`);
        print.log(`You are ready to connect a software delivery machine.
To create one, try 'atomist create sdm'.
More info: https://docs.atomist.com/quick-start/#create-a-software-delivery-machine`);
        return 0;
    });
}
exports.config = config;
/**
 * Initiate a login flow using a selected auth provider to create a new api key
 * @param cfg
 */
function createApiKey(cfg) {
    return __awaiter(this, void 0, void 0, function* () {
        let questions = [
            {
                type: "input",
                name: "apiKey",
                transformer: maskString,
                message: `Enter your ${chalk_1.default.cyan("api key")} from ${chalk_1.default.yellow("https://app.atomist.com/apikeys")}
    or hit ${chalk_1.default.cyan("<ENTER>")} to login with Atomist:`,
            },
        ];
        let answers = yield inquirer.prompt(questions);
        if (!answers.apiKey) {
            const providers = yield axios_1.default.get(`${cfg.endpoints.auth}/providers`);
            let authUrl;
            if (providers.data.length > 1) {
                print.log(`Select one of the following authentication providers\navailable to login with Atomist:`);
                questions = [
                    {
                        type: "list",
                        name: "provider",
                        message: "Authentication Provider",
                        choices: providers.data.map((p) => ({
                            name: p.display_name,
                            value: p.login_url,
                        })),
                    },
                ];
                answers = yield inquirer.prompt(questions);
                authUrl = answers.provider;
            }
            else {
                authUrl = providers.data[0].login_url;
            }
            let spinner = createSpinner(`Waiting for login flow to finish in your browser`);
            const state = nonce(20);
            const verifier = nonce();
            const code = js_sha256_1.sha256(verifier);
            const port = yield port_1.scanFreePort();
            const url = `${authUrl}?state=${state}&redirect-uri=http://127.0.0.1:${port}/callback&code-challenge=${code}`;
            yield opn(url, { wait: false });
            const app = express();
            const callback = new Deferred_1.Deferred();
            app.get("/callback", (req, res) => __awaiter(this, void 0, void 0, function* () {
                if (state !== req.query.state) {
                    callback.reject("State parameter not correct after authentication. Abort!");
                    // res.status(500).json({ message: "State parameter not correct after authentication" });
                    res.redirect("https://atomist.com/error-oauth.html");
                    return;
                }
                try {
                    const token = yield axios_1.default.post(`${cfg.endpoints.auth}/token`, {
                        code: req.query.code,
                        verifier,
                        grant_type: "pkce",
                    });
                    callback.resolve({ jwt: token.data.access_token });
                    res.redirect("https://atomist.com/success-oauth.html");
                }
                catch (e) {
                    callback.reject(new Error(`Authentication failed: ${e.message}`));
                    // res.status(500).json({ message: e.message });
                    res.redirect("https://atomist.com/error-oauth.html");
                }
            }));
            const server = app.listen(port);
            let jwt;
            try {
                jwt = (yield callback.promise).jwt;
            }
            finally {
                server.close();
                spinner.stop(true);
            }
            const graphClient = new ApolloGraphClient_1.ApolloGraphClient(cfg.endpoints.graphql.replace("/team", ""), { Authorization: `Bearer ${jwt}` });
            spinner = createSpinner(`Creating new API key`);
            const result = yield graphClient.mutate({
                mutation: CreateApiKeyMutation,
                variables: {
                    description: `Generated by Atomist CLI on ${os.hostname()} by ${os.userInfo().username}`,
                },
            });
            spinner.stop(true);
            return result.createApiKey.key;
        }
        else {
            return answers.apiKey;
        }
    });
}
/**
 * Validate a given api key by making a backend call to the GraphQL endpoint
 * @param apiKey
 * @param cfg
 */
function validateApiKey(apiKey, cfg) {
    return __awaiter(this, void 0, void 0, function* () {
        const spinner = createSpinner("Validating API key");
        const graphClient = new ApolloGraphClient_1.ApolloGraphClient(cfg.endpoints.graphql.replace("/team", ""), { Authorization: `Bearer ${apiKey}` });
        try {
            const providers = yield axios_1.default.get(`${cfg.endpoints.auth}/providers`);
            const result = yield graphClient.query({
                query: UserQuery,
            });
            spinner.stop(true);
            // If there is no workspace yet, there is also no record returned
            const sub = _.get(result, "user.principal.sub");
            const pid = _.get(result, "user.principal.pid");
            const provider = providers.data.find((p) => p.id === pid);
            if (!!sub && !!pid) {
                print.log(`Logged in as ${chalk_1.default.green(sub)} using ${chalk_1.default.green(_.get(provider, "display_name", "n/a"))}`);
            }
            else {
                print.log(`Logged in`);
            }
        }
        catch (e) {
            if (!!e.networkError && e.networkError.statusCode === 401) {
                print.log(`
API key is invalid. Run ${chalk_1.default.cyan("atomist config --create-api-key")} to obtain a new API key.
`);
            }
            throw e;
        }
        finally {
            spinner.stop(true);
        }
    });
}
exports.validateApiKey = validateApiKey;
/**
 * Read the list of workspaces and let the user choose to which workspaces to connect to
 * @param apiKey
 * @param cfg
 */
function configureWorkspaces(apiKey, cfg, multiple = true) {
    return __awaiter(this, void 0, void 0, function* () {
        const graphClient = new ApolloGraphClient_1.ApolloGraphClient(cfg.endpoints.graphql.replace("/team", ""), { Authorization: `Bearer ${apiKey}` });
        const result = yield graphClient.query({
            query: PersonByIdentityQuery,
            options: automation_client_1.QueryNoCacheOptions,
        });
        const workspaces = _.get(result, "personByIdentity") || [];
        if (workspaces.length === 0) {
            print.log(`No workspaces available. Run ${chalk_1.default.cyan("atomist workspace create")}`);
            return [];
        }
        if (multiple) {
            print.log(`Select one or more workspaces:`);
        }
        else {
            print.log(`Select a workspace:`);
        }
        const questions = [
            {
                type: multiple ? "checkbox" : "list",
                name: "workspaceIds",
                message: "Workspaces",
                choices: workspaces.sort((p1, p2) => p1.team.name.localeCompare(p2.team.name))
                    .map(p => ({
                    name: `${p.team.id} - ${p.team.name}`,
                    value: p.team.id,
                    checked: cfg.workspaceIds.includes(p.team.id),
                    short: p.team.id,
                })),
                validate: (value) => {
                    if (value.length < 1) {
                        return chalk_1.default.red(multiple ? "Please select at least one workspace" : "Please select one workspace");
                    }
                    return true;
                },
            },
        ];
        const answers = yield inquirer.prompt(questions);
        return automation_client_1.toStringArray(answers.workspaceIds) || [];
    });
}
exports.configureWorkspaces = configureWorkspaces;
function createSpinner(text) {
    const Spinner = require("cli-spinner").Spinner;
    const spinner = new Spinner(`${text} ${chalk_1.default.yellow("%s")} `);
    spinner.setSpinnerDelay(100);
    spinner.setSpinnerString("⠋⠙⠹⠸⠼⠴⠦⠧⠇⠏");
    spinner.start();
    return spinner;
}
exports.createSpinner = createSpinner;
function nonce(length = 40) {
    const crypto = require("crypto");
    return crypto
        .randomBytes(Math.ceil((length * 3) / 4))
        .toString("base64") // convert to base64 format
        .slice(0, length) // return required number of characters
        .replace(/\+/g, "0") // replace '+' with '0'
        .replace(/\//g, "0"); // replace '/' with '0'
}
exports.nonce = nonce;
/**
 * Mask secret string.
 * @param secret string to mask
 * @return masked string
 */
function maskString(s) {
    if (s.length > 10) {
        return s.charAt(0) + "*".repeat(s.length - 2) + s.charAt(s.length - 1);
    }
    return "*".repeat(s.length);
}
exports.maskString = maskString;
//# sourceMappingURL=config.js.map