"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const sdm_1 = require("@atomist/sdm");
const chalk_1 = require("chalk");
const fs = require("fs-extra");
const inquirer = require("inquirer");
const path = require("path");
const config_1 = require("./config");
const print = require("./print");
/**
 * Search for an SDM extension pack.
 *
 * @param opts see InstallOptions
 * @return integer return value
 */
function install(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const keywords = opts.keywords.filter(k => !!k);
        let spinner = config_1.createSpinner(`Searching extension packs${keywords.length > 0 ? " for " : ""}${chalk_1.default.cyan(keywords.join(", "))}`);
        let packs;
        try {
            const registryArgs = opts.registry ? ["--registry", opts.registry] : [];
            const result = yield sdm_1.execPromise("npm", [
                "search",
                "--json",
                ...registryArgs,
                "atomist",
                "SDM",
                "extension",
                "pack",
                ...opts.keywords.filter(k => !!k)
            ], {});
            packs = JSON.parse(result.stdout);
            spinner.stop(true);
        }
        catch (e) {
            spinner.stop(true);
            print.error(`Failed to search registry: ${e.message}`);
            return 1;
        }
        if (packs.length > 0) {
            const questions = [
                {
                    type: "list",
                    name: "package",
                    message: "Extension Packs",
                    choices: packs
                        .map(p => ({
                        name: `${p.name} ${chalk_1.default.yellow(p.description)} ${chalk_1.default.reset("by")} ${chalk_1.default.gray(p.maintainers.map(m => m.username).join(", "))}`,
                        value: p,
                        short: p.name,
                    })),
                },
            ];
            const answers = yield inquirer.prompt(questions);
            spinner = config_1.createSpinner(`Installing extension pack ${chalk_1.default.cyan(answers.package.name)} in ${opts.cwd}`);
            try {
                yield sdm_1.execPromise("npm", ["install", answers.package.name], { cwd: opts.cwd });
                spinner.stop(true);
                print.log(`Successfully installed extension pack ${chalk_1.default.cyan(answers.package.name)} by ${chalk_1.default.gray(answers.package.maintainers.map((m) => m.username).join(", "))}`);
                const p = path.join(opts.cwd, "node_modules", answers.package.name, "package.json");
                const pj = JSON.parse((yield fs.readFile(p)).toString());
                print.log(`Visit ${chalk_1.default.yellow(pj.homepage)} for install and usage instructions`);
            }
            catch (e) {
                spinner.stop(true);
                print.error(`Failed to install extension pack '${answers.package.name}': ${e.message}`);
                return 1;
            }
        }
        else {
            print.log(`No SDM extension packs found for: ${opts.keywords.join(", ")}`);
            return 1;
        }
        return 0;
    });
}
exports.install = install;
//# sourceMappingURL=install.js.map