"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const sdm_pack_k8s_1 = require("@atomist/sdm-pack-k8s");
const fs = require("fs-extra");
const inquirer = require("inquirer");
const yaml = require("js-yaml");
const config_1 = require("./config");
const print = require("./print");
/**
 * Encrypt or decrypt secret data values.
 *
 * @param opts see KubeCryptOptions
 * @return integer return value, 0 if successful, non-zero otherwise
 */
function kubeCrypt(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        let secret;
        const literalProp = `literal-${automation_client_1.guid()}`;
        if (opts.literal) {
            secret = wrapLiteral(opts.literal, literalProp);
        }
        else if (opts.file) {
            try {
                const secretString = yield fs.readFile(opts.file, "utf8");
                secret = yield yaml.safeLoad(secretString);
            }
            catch (e) {
                print.error(`Failed to load secret spec from file '${opts.file}': ${e.message}`);
                return 2;
            }
        }
        else {
            const answers = yield inquirer.prompt([{
                    type: "input",
                    name: "literal",
                    message: `Enter literal string to be ${opts.action}ed:`,
                }]);
            secret = wrapLiteral(answers.literal, literalProp);
        }
        if (!opts.secretKey) {
            const answers = yield inquirer.prompt([{
                    type: "input",
                    name: "secretKey",
                    message: `Enter encryption key:`,
                    transformer: config_1.maskString,
                    validate: v => v.length < 1 ? "Secret key must have non-zero length" : true,
                }]);
            opts.secretKey = answers.secretKey;
        }
        const action = (opts.action === "decrypt") ? sdm_pack_k8s_1.decryptSecret : sdm_pack_k8s_1.encryptSecret;
        try {
            const transformed = yield action(secret, opts.secretKey);
            if (transformed.data[literalProp]) {
                print.log(transformed.data[literalProp]);
            }
            else if (/\.ya?ml$/.test(opts.file)) {
                print.log(yaml.safeDump(transformed));
            }
            else {
                print.log(JSON.stringify(transformed, undefined, 2));
            }
        }
        catch (e) {
            print.error(`Failed to ${opts.action} secret: ${e.message}`);
            return 3;
        }
        return 0;
    });
}
exports.kubeCrypt = kubeCrypt;
function wrapLiteral(literal, prop) {
    const secret = {
        apiVersion: "v1",
        data: {},
        kind: "Secret",
        type: "Opaque",
    };
    secret.data[prop] = literal;
    return secret;
}
//# sourceMappingURL=kubeCrypt.js.map