"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const sdm_pack_k8s_1 = require("@atomist/sdm-pack-k8s");
const fs = require("fs-extra");
const path = require("path");
const print = require("./print");
/**
 * Fetch Kubernetes resources from the currently configured Kuberenetes
 * cluster, remove Kubernetes system-provided properties, and write the
 * resource specifications to files.
 *
 * @param opts see KubeFetchOptions
 * @return integer return value, 0 if successful, non-zero otherwise
 */
function kubeFetch(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        let outDir;
        if (opts.outputDir) {
            try {
                yield fs.ensureDir(opts.outputDir);
            }
            catch (e) {
                print.error(`Failed to create output directory '${opts.outputDir}': ${e.message}`);
                return 1;
            }
            outDir = opts.outputDir;
        }
        else {
            outDir = process.cwd();
        }
        const fileFormat = (opts.outputFormat && opts.outputFormat.toLowerCase() === "json") ? "json" : "yaml";
        let options;
        if (opts.optionsFile) {
            try {
                options = yield fs.readJson(opts.optionsFile);
            }
            catch (e) {
                print.error(`Failed to read Kubernetes fetch options file '${opts.optionsFile}': ${e.message}`);
                return 2;
            }
        }
        const errs = [];
        print.log("Fetching Kubernetes resources...");
        try {
            const resources = yield sdm_pack_k8s_1.kubernetesFetch(options);
            for (const spec of resources) {
                const specPath = path.join(outDir, sdm_pack_k8s_1.kubernetesSpecFileBasename(spec) + "." + fileFormat);
                print.log(`Writing Kubernetes spec: ${specPath}`);
                try {
                    const specString = yield sdm_pack_k8s_1.kubernetesSpecStringify(spec, { format: fileFormat, secretKey: opts.secretKey });
                    yield fs.writeFile(specPath, specString);
                }
                catch (e) {
                    errs.push(`'${specPath}': ${e.message}`);
                }
            }
        }
        catch (e) {
            print.error(`Failed to fetch Kubernetes resources: ${e.message}`);
            return 3;
        }
        if (errs.length > 0) {
            const s = (errs.length > 1) ? "s" : "";
            print.error(`Failed to write Kubernetes spec file${s}: ${errs.join(", ")}`);
            return 10 + errs.length;
        }
        return 0;
    });
}
exports.kubeFetch = kubeFetch;
//# sourceMappingURL=kubeFetch.js.map