"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const sdm_1 = require("@atomist/sdm");
const api_1 = require("@atomist/sdm-pack-k8s/lib/kubernetes/api");
const config_1 = require("@atomist/sdm-pack-k8s/lib/kubernetes/config");
const chalk_1 = require("chalk");
const cli_highlight_1 = require("cli-highlight");
const fs = require("fs-extra");
const inquirer = require("inquirer");
const yaml = require("js-yaml");
const stringify = require("json-stringify-safe");
const os = require("os");
const path = require("path");
const request = require("request");
const cliConfig_1 = require("./cliConfig");
const print = require("./print");
/**
 * Deploy Atomist Kubernetes utilities into a Kubernetes cluster.
 *
 * @param opts see [[KubeInstallOptions]]
 * @return integer return value, 0 if successful, non-zero otherwise
 */
// tslint:disable-next-line:cyclomatic-complexity
function kubeInstall(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const ns = opts.ns;
        const yes = opts.yes;
        let dryRun = opts.dryRun;
        const cliConfig = cliConfig_1.resolveCliConfig();
        const apiKey = cliConfig.apiKey;
        if (!apiKey) {
            print.error(`No API key set in user configuration, run 'atomist config' first`);
            return Promise.resolve(1);
        }
        const workspaceIds = cliConfig.workspaceIds;
        if (!workspaceIds || workspaceIds.length < 1) {
            print.error(`No workspace IDs set in user configuration, run 'atomist config' first`);
            return Promise.resolve(1);
        }
        let context;
        try {
            const kubeConfigPath = path.join(os.homedir(), ".kube", "config");
            const kubeConfigString = yield fs.readFile(kubeConfigPath, "utf8");
            const kubeConfig = yaml.safeLoad(kubeConfigString);
            context = kubeConfig["current-context"];
        }
        catch (e) {
            print.warn(`Failed to get current context from Kubernetes config: ${e.message}`);
        }
        let url = opts.url;
        if (context === "minikube" && !url) {
            try {
                const ipResult = yield sdm_1.execPromise("minikube", ["ip"]);
                url = `http://${ipResult.stdout.trim()}`;
            }
            catch (e) {
                print.warn(`Failed to get minikube IP address, not setting URL: ${e.message}`);
            }
        }
        const environment = opts.env || context || "kubernetes";
        if (dryRun === undefined && yes === undefined) {
            const target = context ? `context ${chalk_1.default.cyan(context)}` : `cluster ${chalk_1.default.cyan(environment)}`;
            const questions = [
                {
                    type: "list",
                    name: "dryRun",
                    message: `Deploy Atomist Kubernetes utilities to ${target}:`,
                    default: "yes",
                    choices: [
                        {
                            name: "Yes",
                            value: "yes",
                            short: "Yes",
                        },
                        {
                            name: "Dry-run (print Kubernetes specs but do not apply them)",
                            value: "dry-run",
                            short: "Dry-run",
                        },
                        {
                            name: "No (exits immediately)",
                            value: "no",
                            short: "No",
                        },
                    ],
                },
            ];
            const answers = yield inquirer.prompt(questions);
            if (answers.dryRun === "no") {
                return 0;
            }
            else if (answers.dryRun === "dry-run") {
                dryRun = true;
            }
            else {
                dryRun = false;
            }
        }
        let specs;
        try {
            specs = yield fetchSpecs(ns);
        }
        catch (e) {
            print.error(e.message);
            return 1;
        }
        const webhooks = kubeWebhookUrls(workspaceIds);
        const k8sConfig = k8sSdmConfig({ apiKey, environment, url, workspaceIds });
        specs.push(encodeSecret("k8vent", ns || "k8vent", { environment, webhooks }));
        specs.push(encodeSecret("k8s-sdm", ns || "sdm", { "client.config.json": stringify(k8sConfig) }));
        if (dryRun) {
            specs.forEach(spec => {
                print.log("---");
                print.log(cli_highlight_1.highlight(yaml.safeDump(spec), { language: "yaml" }).replace(/\n$/, ""));
            });
            return 0;
        }
        let client;
        try {
            const kc = config_1.loadKubeConfig();
            client = kc.makeApiClient(api_1.K8sObjectApi);
        }
        catch (e) {
            print.error(`Failed to create Kubernetes client: ${errMsg(e)}`);
            return 2;
        }
        for (const spec of specs) {
            try {
                yield applySpec(client, spec);
            }
            catch (e) {
                print.error(`Failed to apply ${specSlug(spec)} spec: ${errMsg(e)}`);
                return 3;
            }
        }
        print.log("");
        print.log(`Successfully installed Atomist Kubernetes utilities into your cluster`);
        print.log(`Please confirm correct startup of k8s-sdm by running:`);
        print.log(`  $ ${chalk_1.default.yellow("kubectl get pod -n " + (ns || "sdm"))}`);
        return 0;
    });
}
exports.kubeInstall = kubeInstall;
/**
 * Fetch the appropriate Kubernetes specs from GitHub and parse them.
 *
 * @param ns Namespace resources are to be deployed to, if not cluster-wide
 * @return Resource specs that need to be upserted
 */
function fetchSpecs(ns) {
    return __awaiter(this, void 0, void 0, function* () {
        const specTails = {
            "k8s-sdm": {
                ns: "/assets/kubectl/namespace-scoped.yaml",
                cluster: "/assets/kubectl/cluster-wide.yaml",
            },
            "k8vent": {
                ns: "/kube/kubectl/namespace-scoped.yaml",
                cluster: "/kube/kubectl/cluster-wide.yaml",
            },
        };
        const specs = [];
        for (const specSrc of Object.keys(specTails)) {
            const specBaseUrl = ghBaseRawUrl(specSrc);
            const specUrl = specBaseUrl + ((ns) ? specTails[specSrc].ns : specTails[specSrc].cluster);
            try {
                print.log(`Fetching ${chalk_1.default.cyan(specSrc)} specs: ${specUrl}`);
                const result = yield requestPromise(specUrl);
                specs.push(...processSpecs(result.body, ns));
            }
            catch (e) {
                e.message = `Failed to download and parse spec '${specUrl}': ${errMsg(e)}`;
                throw e;
            }
        }
        return specs;
    });
}
exports.fetchSpecs = fetchSpecs;
/**
 * Convert workspace IDs to Atomist Kubernetes webhook URLs.
 *
 * @param workspaceIds array of Atomist workspace/team IDs
 * @return comma-delimited list of webhook URLs
 */
function kubeWebhookUrls(workspaceIds) {
    const base = `${automation_client_1.webhookBaseUrl()}/atomist/kube/teams`;
    return workspaceIds.map(id => `${base}/${id}`).join(",");
}
exports.kubeWebhookUrls = kubeWebhookUrls;
/**
 * Return base raw content GitHub.com URL for atomist repos.
 */
function ghBaseRawUrl(repo) {
    return `https://raw.githubusercontent.com/atomist/${repo}/master`;
}
/**
 * Create appropriate configuration for a k8s-sdm.
 */
function k8sSdmConfig(opts) {
    const cfg = {
        apiKey: opts.apiKey,
        environment: opts.environment,
        logging: { level: "debug" },
        name: `@atomist/k8s-sdm_${opts.environment}`,
        sdm: {
            k8s: {
                options: {
                    addCommands: true,
                    registerCluster: true,
                },
            },
            kubernetes: {
                provider: {
                    url: opts.url,
                },
            },
        },
        workspaceIds: opts.workspaceIds,
    };
    return cfg;
}
exports.k8sSdmConfig = k8sSdmConfig;
function requestPromise(uri) {
    return new Promise((resolve, reject) => {
        request(uri, (error, response, body) => {
            if (error) {
                reject(error);
            }
            else {
                if (response.statusCode >= 200 && response.statusCode <= 299) {
                    resolve({ response, body });
                }
                else {
                    reject({ response, body });
                }
            }
        });
    });
}
function errMsg(e) {
    if (e.message) {
        return e.message;
    }
    else if (e.body && e.body.message) {
        return e.body.message;
    }
    else if (e.response && e.response.body && e.response.body.message) {
        return e.response.body.message;
    }
    else if (e.response && e.response.statusMessage) {
        return e.response.body.message;
    }
    else {
        return stringify(e);
    }
}
function processSpecs(raw, ns) {
    const specs = yaml.safeLoadAll(raw);
    if (ns) {
        specs.forEach(s => s.metadata.namespace = ns);
    }
    return specs;
}
exports.processSpecs = processSpecs;
/**
 * Create encoded secret object from key-value pairs.
 *
 * @param secrets key-value pairs of secrets, the values are base64 encoded
 * @return Kubernetes secret object
 */
function encodeSecret(name, ns, secrets) {
    const kubeSecret = {
        apiVersion: "v1",
        kind: "Secret",
        type: "Opaque",
        metadata: {
            name,
            namespace: ns,
        },
        data: {},
    };
    for (const secret of Object.keys(secrets)) {
        kubeSecret.data[secret] = Buffer.from(secrets[secret]).toString("base64");
    }
    return kubeSecret;
}
exports.encodeSecret = encodeSecret;
/**
 * Return informative string for spec.
 */
function specSlug(spec) {
    return [spec.kind, spec.metadata.namespace, spec.metadata.name].join("/").replace("//", "/");
}
exports.specSlug = specSlug;
/**
 * Create or update Kubernetes resource.
 */
function applySpec(client, spec) {
    return __awaiter(this, void 0, void 0, function* () {
        const slug = specSlug(spec);
        try {
            yield client.read(spec);
        }
        catch (e) {
            print.log(`Creating ${slug}`);
            yield client.create(spec);
            return;
        }
        print.log(`Updating ${slug}`);
        yield client.patch(spec);
        return;
    });
}
exports.applySpec = applySpec;
//# sourceMappingURL=kubeInstall.js.map