"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const child_process_1 = require("@atomist/automation-client/lib/util/child_process");
const fs = require("fs-extra");
const gitUrlParse = require("git-url-parse");
const _ = require("lodash");
const os = require("os");
const path = require("path");
const print = require("./print");
const start_1 = require("./start");
/**
 * Files to check and copy from the seed into the cloned client/SDM repo
 */
const FilesToCopy = [
    "package.json",
    "tsconfig.json",
    "tslint.json",
];
/**
 * Clone, prepare and start an SDM/client project from a remote Git or Gist location
 */
function repositoryStart(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const optsToUse = _.merge({
            index: "index.ts",
            sha: "master",
            local: false,
            profile: undefined,
            watch: false,
            debug: false,
            seedUrl: "https://github.com/atomist-seeds/empty-sdm.git",
        }, opts);
        let cwd = optsToUse.cwd;
        if (!!optsToUse.cloneUrl) {
            const gitUrl = gitUrlParse(optsToUse.cloneUrl);
            cwd = path.join(os.homedir(), ".atomist", "sdm", gitUrl.owner, gitUrl.name);
            const seed = path.join(os.homedir(), ".atomist", "cache", `sdm-${automation_client_1.guid().slice(0, 7)}`);
            // Git clone
            try {
                yield clone(optsToUse, cwd);
            }
            catch (e) {
                print.error(`Failed to clone/checkout repository: ${e.message}`);
                return 5;
            }
            // Move the provided file into the index.ts
            if (!!optsToUse.index && optsToUse.index !== "index.ts") {
                yield copyIndexTs(optsToUse, cwd);
            }
            // Clone the seed if there are some needed files missing in the repository
            try {
                if (FilesToCopy.some(f => !fs.pathExistsSync(path.join(cwd, f)))) {
                    yield cloneSeed(optsToUse, seed);
                }
            }
            catch (e) {
                print.error(`Failed to checkout seed: ${e.message}`);
                return 10;
            }
            // Copy over package.json, tsconfig.json in case they are missing
            try {
                yield copyFiles(optsToUse, cwd, seed);
            }
            catch (e) {
                print.error(`Failed to copy seed files into clone: ${e.message}`);
                return 15;
            }
            // Delete the seed clone as we don't need it any more
            try {
                yield fs.remove(seed);
            }
            catch (e) {
                print.warn(`Failed to remove seed: ${e.message}`);
            }
        }
        // Finally call start on the SDM/client
        return start_1.start(Object.assign(Object.assign({}, optsToUse), { cwd }));
    });
}
exports.repositoryStart = repositoryStart;
function clone(optsToUse, cwd) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!(yield fs.pathExists(cwd))) {
            print.info("Cloning repository...");
            yield child_process_1.execPromise("git", ["clone", optsToUse.cloneUrl, cwd]);
            print.info("Finished");
        }
        else {
            print.info("Updating repository...");
            for (const f of FilesToCopy) {
                yield fs.remove(path.join(cwd, f));
            }
            yield child_process_1.execPromise("git", ["reset", "--hard"], { cwd });
            yield child_process_1.execPromise("git", ["pull"], { cwd });
            print.info("Finished");
        }
        if (!!optsToUse.sha) {
            print.info(`Checking out '${optsToUse.sha}'...`);
            yield child_process_1.execPromise("git", ["checkout", optsToUse.sha], { cwd });
            print.info("Finished");
        }
    });
}
function cloneSeed(optsToUse, seed) {
    return __awaiter(this, void 0, void 0, function* () {
        print.info("Cloning seed...");
        yield child_process_1.execPromise("git", ["clone", optsToUse.seedUrl, seed]);
        print.info("Finished");
    });
}
function copyIndexTs(optsToUse, cwd) {
    return __awaiter(this, void 0, void 0, function* () {
        print.info(`Preparing '${optsToUse.index}'...`);
        const from = path.join(cwd, optsToUse.index.replace(".ts", "").replace(".js", ""));
        // Rewrite the index.ts to export the configuration from provided file to not break relative imports
        const indexTs = `import { configuration as cfg } from "${from}";

export const configuration = cfg;
`;
        yield fs.writeFile(path.join(cwd, "index.ts"), indexTs);
        print.info("Finished");
    });
}
function copyFiles(optsToUse, cwd, seed) {
    return __awaiter(this, void 0, void 0, function* () {
        const gitUrl = gitUrlParse(optsToUse.cloneUrl);
        for (const f of FilesToCopy) {
            const fp = path.join(cwd, f);
            const sp = path.join(seed, f);
            if (!(yield fs.pathExists(fp)) && (yield fs.pathExists(sp))) {
                print.info(`Creating '${f}'`);
                yield fs.copyFile(sp, fp);
                if (f === "package.json") {
                    const pj = yield fs.readJson(fp);
                    // This is special handling required to support gist clone urls
                    if ((gitUrl.owner.length === 0 || gitUrl.owner === "git")
                        && optsToUse.cloneUrl.includes("gist")) {
                        pj.name = `@gist/${gitUrl.name.slice(0, 7)}`;
                    }
                    else {
                        pj.name = `@${gitUrl.owner}/${gitUrl.name}`;
                    }
                    const sha = yield child_process_1.execPromise("git", ["log", "--pretty=format:%h", "-n", "1"], { cwd });
                    pj.version = `0.1.0-${sha.stdout.trim()}`;
                    yield fs.writeJson(fp, pj, { replacer: undefined, spaces: 2 });
                }
            }
        }
    });
}
//# sourceMappingURL=repositoryStart.js.map