"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const sdm_1 = require("@atomist/sdm");
const chalk_1 = require("chalk");
const fs = require("fs-extra");
const path = require("path");
const config_1 = require("./config");
const print = require("./print");
function getAtomistDependencies(dependencies) {
    return Object.keys(dependencies).filter(key => key.startsWith("@atomist/"));
}
function hasPackageJson(directory) {
    return fs.existsSync(path.join(directory, "package.json"));
}
function updateDependenciesToTag(dependencies, versionTag, spinner) {
    return __awaiter(this, void 0, void 0, function* () {
        const dependenciesToUpdate = dependencies
            .filter((d) => __awaiter(this, void 0, void 0, function* () {
            const moduleVersion = yield getModuleVersion(`${d}@${versionTag}`);
            return moduleVersion !== undefined;
        }))
            .map(d => `${d}@${versionTag}`);
        spinner.setSpinnerTitle(`Updating ${dependenciesToUpdate.length} dependencies to ${versionTag} ${chalk_1.default.yellow("%s")}`);
        yield sdm_1.execPromise("npm", [
            "install",
            "--save",
            ...dependenciesToUpdate,
        ]);
    });
}
function getModuleVersion(module) {
    return __awaiter(this, void 0, void 0, function* () {
        const log = new sdm_1.StringCapturingProgressLog();
        const result = yield sdm_1.spawnLog("npm", ["show", module, "version"], {
            logCommand: false,
            log,
        });
        if (result.code === 0) {
            return log.log.trim();
        }
        return undefined;
    });
}
function updateDevDependenciesToTag(dependencies, versionTag, spinner) {
    return __awaiter(this, void 0, void 0, function* () {
        const dependenciesToUpdate = dependencies.map(d => `${d}@${versionTag}`);
        spinner.setSpinnerTitle(`Updating dev dependencies to ${versionTag} ${chalk_1.default.yellow("%s")}`);
        yield sdm_1.execPromise("npm", [
            "install",
            "--save-dev",
            ...dependenciesToUpdate,
        ]);
    });
}
/**
 * Search for an SDM extension pack.
 *
 * @param opts see InstallOptions
 * @return integer return value
 */
function updateSdm(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        if (hasPackageJson(opts.cwd)) {
            const spinner = config_1.createSpinner(`Updating dependencies in package.json`);
            try {
                const packageJson = yield fs.readJson(path.join(opts.cwd, "package.json"));
                const extractedDependencies = getAtomistDependencies(packageJson.dependencies);
                const extractedDevDependencies = getAtomistDependencies(packageJson.devDependencies);
                yield updateDependenciesToTag(extractedDependencies, opts.versionTag, spinner);
                yield updateDevDependenciesToTag(extractedDevDependencies, opts.versionTag, spinner);
                spinner.stop(true);
                return 0;
            }
            catch (e) {
                print.error(`Error while updating package.json`);
                return 1;
            }
        }
        else {
            print.error(`Current directory does not contain a package.json`);
            return 1;
        }
    });
}
exports.updateSdm = updateSdm;
//# sourceMappingURL=updateSdm.js.map