"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const metadataReading_1 = require("./internal/metadata/metadataReading");
const string_1 = require("./internal/util/string");
const scan_1 = require("./scan");
const constructionUtils_1 = require("./util/constructionUtils");
var QuestionStyle;
(function (QuestionStyle) {
    QuestionStyle["Dialog"] = "dialog";
    QuestionStyle["Threaded"] = "threaded";
    QuestionStyle["Unthreaded"] = "unthreaded";
    QuestionStyle["DialogAction"] = "dialog_action";
})(QuestionStyle = exports.QuestionStyle || (exports.QuestionStyle = {}));
/**
 * Given a regex used for an intent, add anchors to it if missing.  Additionally, wraps the new anchors in match groups along with the original
 * intent for access later.
 */
function addRegExIntentAnchors(intent) {
    if (!(intent instanceof RegExp)) {
        throw new Error(`Intent must be an instance of RegExp!`);
    }
    let regex = intent.source;
    if (!regex.startsWith("^")) {
        regex = "^[\\s\\S]*" + regex;
    }
    if (!regex.endsWith("$")) {
        regex = regex + "[\\s\\S]*$";
    }
    return regex;
}
exports.addRegExIntentAnchors = addRegExIntentAnchors;
/**
 * Parse command handler intent.  If the incoming intent is a RegExp convert it to a string pattern.  Otherwise, pass existing intent through.
 */
function parseRegExIntent(intent) {
    if (typeof intent === "string" && intent.length > 0) {
        return intent;
    }
    else if ((Array.isArray(intent) && (intent.length < 1 || intent.every(i => typeof i === "string")))) {
        return intent;
    }
    else if (intent instanceof RegExp) {
        return addRegExIntentAnchors(intent);
    }
    else if (intent === undefined || intent === null || (typeof intent === "string" && intent === "")) {
        throw new Error(`Intent cannot be undefined, null, or empty!`);
    }
    else {
        throw new Error(`Unknown Intent type of ${typeof intent} supplied!`);
    }
}
exports.parseRegExIntent = parseRegExIntent;
/**
 * Create a HandleCommand instance with the appropriate metadata wrapping
 * the given function
 */
function commandHandlerFrom(h, factory, name = h.name || `Command${string_1.generateHash(h.toString())}`, description = name, intent = [], tags = [], autoSubmit = false, question) {
    const updatedIntent = parseRegExIntent(intent);
    const handler = new FunctionWrappingCommandHandler(name, description, h, factory, tags, updatedIntent, autoSubmit, question);
    scan_1.registerCommand(handler);
    return handler;
}
exports.commandHandlerFrom = commandHandlerFrom;
class FunctionWrappingCommandHandler {
    constructor(name, description, h, parametersFactory, _tags = [], _intent = [], autoSubmit = false, _question) {
        this.name = name;
        this.description = description;
        this.h = h;
        this.parametersFactory = parametersFactory;
        this._tags = _tags;
        this._intent = _intent;
        this.autoSubmit = autoSubmit;
        this._question = _question;
        const newParamInstance = this.freshParametersInstance();
        const md = metadataReading_1.metadataFromInstance(newParamInstance);
        this.parameters = md.parameters;
        this.mapped_parameters = md.mapped_parameters;
        this.values = md.values;
        this.secrets = md.secrets;
        this.intent = string_1.toStringArray(_intent);
        this.tags = string_1.toStringArray(_tags).map(t => ({ name: t, description: t }));
        this.auto_submit = autoSubmit;
        this.question = (!!_question ? _question.toString() : undefined);
    }
    /* tslint:enAble:variable-name */
    freshParametersInstance() {
        return constructionUtils_1.toFactory(this.parametersFactory)();
    }
    handle(ctx, params) {
        return this.h(ctx, params);
    }
}
//# sourceMappingURL=onCommand.js.map