"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const ActionResult_1 = require("../../action/ActionResult");
const configuration_1 = require("../../configuration");
const httpClient_1 = require("../../spi/http/httpClient");
const logger_1 = require("../../util/logger");
const AbstractRemoteRepoRef_1 = require("./AbstractRemoteRepoRef");
const BasicAuthCredentials_1 = require("./BasicAuthCredentials");
const RepoId_1 = require("./RepoId");
/**
 * RemoteRepoRef implementation for BitBucket server (not BitBucket Cloud)
 */
class BitBucketServerRepoRef extends AbstractRemoteRepoRef_1.AbstractRemoteRepoRef {
    /**
     * Construct a new BitBucketServerRepoRef
     * @param {string} remoteBase remote base, including scheme
     * @param {string} owner
     * @param {string} repo
     * @param {boolean} isProject
     * @param {string} sha
     * @param {string} path
     * @param {string} branch
     * @param {string}apiUrl
     */
    constructor(remoteBase, owner, repo, isProject = true, sha, path, branch, apiUrl) {
        super(RepoId_1.ProviderType.bitbucket, remoteBase, apiUrl || `${noTrailingSlash(remoteBase)}/rest/api/1.0/`, owner, repo, sha, path, branch);
        this.isProject = isProject;
        this.kind = "bitbucketserver";
        this.ownerType = isProject ? "projects" : "users";
        logger_1.logger.debug("Constructed BitBucketServerRepoRef: %j", this);
    }
    createRemote(creds, description, visibility) {
        const url = `${this.scheme}${this.apiBase}/${this.apiBasePathComponent}`;
        const data = {
            name: this.repo,
            scmId: "git",
            forkable: "true",
        };
        logger_1.logger.debug("Making request to BitBucket '%s' to create repo, data=%j", url, data);
        return configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory()).create(url).exchange(url, {
            method: httpClient_1.HttpMethod.Post,
            body: data,
            headers: Object.assign({ "Content-Type": "application/json" }, usernameColonPassword(creds)),
        })
            .then(response => ({
            success: true,
            target: this,
            response,
        }))
            .catch(error => {
            error.message = `Error attempting to create repository ${JSON.stringify(this)}}: ${error.message}`;
            logger_1.logger.error(error.message);
            throw error;
        });
    }
    deleteRemote(creds) {
        const url = `${this.scheme}${this.apiBase}/${this.apiPathComponent}`;
        logger_1.logger.debug(`Making request to '${url}' to delete repo`);
        return configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory()).create(url).exchange(url, {
            method: httpClient_1.HttpMethod.Delete,
            headers: Object.assign({}, usernameColonPassword(creds)),
        })
            .then(response => ({
            success: true,
            target: this,
            response,
        }))
            .catch(error => {
            error.message = `Error attempting to delete repository: ${error.message}`;
            logger_1.logger.error(error.message);
            throw error;
        });
    }
    setUserConfig(credentials, project) {
        return Promise.resolve(ActionResult_1.successOn(this));
    }
    raisePullRequest(creds, title, body, head, base, reviewers) {
        return __awaiter(this, void 0, void 0, function* () {
            const url = `${this.scheme}${this.apiBase}/${this.apiPathComponent}/pull-requests`;
            logger_1.logger.debug(`Making request to '${url}' to raise PR`);
            const repoId = yield this.getRepoId(creds);
            // Figure out reviewers
            const allReviewers = yield this.getDefaultReviewers(creds, repoId, head, base);
            if (reviewers) {
                if (reviewers.some(r => r.type !== RepoId_1.PullRequestReviewerType.individual)) {
                    throw new Error(`Bitbucket only supports reviewer type of individual!  Found ` +
                        JSON.stringify([...new Set(reviewers.map(r => RepoId_1.PullRequestReviewerType[r.type]))]));
                }
                allReviewers.push(...reviewers.map(r => r.name));
            }
            // Build payload
            const data = {
                title,
                description: body,
                fromRef: {
                    id: head,
                },
                toRef: {
                    id: base,
                },
                reviewers: allReviewers.map(r => {
                    return {
                        user: {
                            name: r,
                        },
                    };
                }),
            };
            return configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory()).create(url).exchange(url, {
                method: httpClient_1.HttpMethod.Post,
                body: data,
                headers: Object.assign({ "Content-Type": "application/json" }, usernameColonPassword(creds)),
            })
                .then(response => ({
                success: true,
                target: this,
                response,
            }))
                .catch(error => {
                error.message = `Error attempting to raise PR: ${error.message}`;
                logger_1.logger.error(error.message);
                throw error;
            });
        });
    }
    getDefaultReviewers(creds, repoId, head, base) {
        return __awaiter(this, void 0, void 0, function* () {
            const restApiPath = `/rest/default-reviewers/1.0/${this.apiBasePathComponent}${this.repo}/reviewers`;
            const url = `${this.scheme}${noTrailingSlash(this.remoteBase)}${restApiPath}`;
            const queryParams = `sourceRepoId=${repoId}&targetRepoId=${repoId}&sourceRefId=${head}&targetRefId=${base}`;
            const urlWithQueryParams = `${url}?${queryParams}`;
            const apiResponse = yield configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory())
                .create(url).exchange(urlWithQueryParams, {
                method: httpClient_1.HttpMethod.Get,
                headers: Object.assign({}, usernameColonPassword(creds)),
            });
            return apiResponse.body.map(reviewer => reviewer.name);
        });
    }
    getRepoId(creds) {
        return __awaiter(this, void 0, void 0, function* () {
            const url = `${this.scheme}${this.apiBase}/${this.apiPathComponent}`;
            const apiResponse = yield configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory()).create(url).exchange(url, {
                method: httpClient_1.HttpMethod.Get,
                headers: Object.assign({}, usernameColonPassword(creds)),
            });
            return apiResponse.body.id;
        });
    }
    get url() {
        return `${this.scheme}${this.remoteBase}/${this.ownerType}/${this.owner}/repos/${this.repo}`;
    }
    get pathComponent() {
        return `scm/${this.maybeTilde}${this.owner}/${this.repo}`;
    }
    get maybeTilde() {
        return this.isProject ? "" : "~";
    }
    get apiBasePathComponent() {
        return `projects/${this.maybeTilde}${this.owner}/repos/`;
    }
    get apiPathComponent() {
        return this.apiBasePathComponent + this.repo;
    }
}
exports.BitBucketServerRepoRef = BitBucketServerRepoRef;
function usernameColonPassword(creds) {
    if (BasicAuthCredentials_1.isBasicAuthCredentials(creds)) {
        return {
            Authorization: `Basic ${Buffer.from(creds.username + ":" + creds.password).toString("base64")}`,
        };
    }
    else {
        return {};
    }
}
function noTrailingSlash(s) {
    return s.replace(/\/$/, "");
}
//# sourceMappingURL=BitBucketServerRepoRef.js.map