import { ActionResult } from "../../action/ActionResult";
import { Configurable } from "../../project/git/Configurable";
import { HttpResponse } from "../../spi/http/httpClient";
import { AbstractRemoteRepoRef } from "./AbstractRemoteRepoRef";
import { ProjectOperationCredentials } from "./ProjectOperationCredentials";
import { PullRequestReviewer, RepoRef } from "./RepoId";
export declare const GitHubDotComBase = "https://api.github.com";
/**
 * GitHub repo ref
 */
export declare class GitHubRepoRef extends AbstractRemoteRepoRef {
    /**
     * Create a GitHubRepoRef instance.
     * @param params Object with the following properties:
     *                 owner: Repo owner
     *                 repo: Repo name
     *                 sha: Commit SHA to checkout
     *                 rawApiBase: Full GitHub API base URL
     *                 path: Path within the Git repository to use as project root
     *                 branch: Branch to checkout
     *                 rawRemoteBase: Full GitHub remote base URL
     */
    static from(params: {
        owner: string;
        repo: string;
        sha?: string;
        rawApiBase?: string;
        path?: string;
        branch?: string;
    }): GitHubRepoRef;
    readonly kind: string;
    readonly apiBase: string;
    /**
     * Create a GitHubRepoRef instance.  It may be easier to use [[GitHubRepoRef.from]].
     * @param owner Repo owner
     * @param repo Repo name
     * @param sha Commit SHA to checkout
     * @param rawApiBase Full GitHub API base URL
     * @param path Path within the Git repository to use as project root
     * @param branch Branch to checkout
     * @param rawRemoteBase Full GitHub remote base URL
     */
    constructor(owner: string, repo: string, sha?: string, rawApiBase?: string, path?: string, branch?: string, rawRemoteBase?: string);
    createRemote(creds: ProjectOperationCredentials, description: string, visibility: string): Promise<ActionResult<this>>;
    setUserConfig(credentials: ProjectOperationCredentials, project: Configurable): Promise<ActionResult<any>>;
    getPr(credentials: ProjectOperationCredentials, head: string): Promise<any>;
    /**
     * Used to assign reviewers to existing Pull Requests
     * https://developer.github.com/v3/pulls/review_requests/#create-a-review-request
     * @param {ProjectOperationCredentials} credentials
     * @param {string} prNumber
     * @param {PullRequestReviewer[]} reviewers
     */
    addReviewersToPullRequest(credentials: ProjectOperationCredentials, prNumber: string, reviewers: PullRequestReviewer[]): Promise<HttpResponse<any>>;
    /**
     * Used to create a new Pull Request
     * https://developer.github.com/v3/pulls/#create-a-pull-request
     * @param {ProjectOperationCredentials} credentials
     * @param {string} title
     * @param {string} body
     * @param {string} head Source branch
     * @param {string} base Target branch
     * @param {PullRequestReviewer[]} reviewers
     */
    raisePullRequest(credentials: ProjectOperationCredentials, title: string, body: string, head: string, base: string, reviewers?: PullRequestReviewer[]): Promise<ActionResult<this>>;
    deleteRemote(creds: ProjectOperationCredentials): Promise<ActionResult<this>>;
}
export declare function isGitHubRepoRef(rr: RepoRef): rr is GitHubRepoRef;
export declare function beautifyPullRequestBody(body: string): string;
//# sourceMappingURL=GitHubRepoRef.d.ts.map