"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const ActionResult_1 = require("../../action/ActionResult");
const configuration_1 = require("../../configuration");
const httpClient_1 = require("../../spi/http/httpClient");
const gitHub_1 = require("../../util/gitHub");
const logger_1 = require("../../util/logger");
const AbstractRemoteRepoRef_1 = require("./AbstractRemoteRepoRef");
const validationPatterns_1 = require("./params/validationPatterns");
const ProjectOperationCredentials_1 = require("./ProjectOperationCredentials");
const RepoId_1 = require("./RepoId");
exports.GitHubDotComBase = "https://api.github.com";
const FullGitHubDotComBase = "https://api.github.com/";
/**
 * GitHub repo ref
 */
class GitHubRepoRef extends AbstractRemoteRepoRef_1.AbstractRemoteRepoRef {
    /**
     * Create a GitHubRepoRef instance.  It may be easier to use [[GitHubRepoRef.from]].
     * @param owner Repo owner
     * @param repo Repo name
     * @param sha Commit SHA to checkout
     * @param rawApiBase Full GitHub API base URL
     * @param path Path within the Git repository to use as project root
     * @param branch Branch to checkout
     * @param rawRemoteBase Full GitHub remote base URL
     */
    constructor(owner, repo, sha, rawApiBase = exports.GitHubDotComBase, path, branch, rawRemoteBase) {
        super(rawApiBase === exports.GitHubDotComBase || rawApiBase === FullGitHubDotComBase ? RepoId_1.ProviderType.github_com : RepoId_1.ProviderType.ghe, rawRemoteBase || apiBaseToRemoteBase(rawApiBase), rawApiBase, owner, repo, sha, path, branch);
        this.kind = "github";
    }
    /**
     * Create a GitHubRepoRef instance.
     * @param params Object with the following properties:
     *                 owner: Repo owner
     *                 repo: Repo name
     *                 sha: Commit SHA to checkout
     *                 rawApiBase: Full GitHub API base URL
     *                 path: Path within the Git repository to use as project root
     *                 branch: Branch to checkout
     *                 rawRemoteBase: Full GitHub remote base URL
     */
    static from(params) {
        if (params.sha && !params.sha.match(validationPatterns_1.GitShaRegExp.pattern)) {
            throw new Error("You provided an invalid SHA: " + params.sha);
        }
        return new GitHubRepoRef(params.owner, params.repo, params.sha, params.rawApiBase, params.path, params.branch);
    }
    createRemote(creds, description, visibility) {
        if (!ProjectOperationCredentials_1.isTokenCredentials(creds)) {
            throw new Error("Only token auth supported");
        }
        return gitHub_1.createRepo(creds.token, this, description, visibility === "private")
            .then(() => ActionResult_1.successOn(this));
    }
    setUserConfig(credentials, project) {
        // Only permit one retry trying to lookup user info
        const config = Object.assign(Object.assign({}, headers(credentials)), { retry: { retries: 1 } });
        const httpClient = configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory())
            .create(`${this.scheme}${this.apiBase}`);
        return Promise.all([
            httpClient.exchange(`${this.scheme}${this.apiBase}/user`, config),
            httpClient.exchange(`${this.scheme}${this.apiBase}/user/public_emails`, config),
        ])
            .then(results => {
            const name = results[0].body.name || results[0].body.login;
            let email = results[0].body.email;
            if (!email && results[1].body && results[1].body.length > 0) {
                email = results[1].body[0].email;
            }
            if (name && email) {
                return project.setUserConfig(name, email);
            }
            else {
                return project.setUserConfig("Atomist Bot", "bot@atomist.com");
            }
        })
            .catch(() => project.setUserConfig("Atomist Bot", "bot@atomist.com"))
            .then(ActionResult_1.successOn);
    }
    getPr(credentials, head) {
        return __awaiter(this, void 0, void 0, function* () {
            const config = headers(credentials);
            const url = `${this.scheme}${this.apiBase}/repos/${this.owner}/${this.repo}/pulls?state=open&head=${this.owner}:${head}`;
            const httpClient = configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory())
                .create(url);
            return (yield httpClient.exchange(url, config)).body;
        });
    }
    /**
     * Used to assign reviewers to existing Pull Requests
     * https://developer.github.com/v3/pulls/review_requests/#create-a-review-request
     * @param {ProjectOperationCredentials} credentials
     * @param {string} prNumber
     * @param {PullRequestReviewer[]} reviewers
     */
    addReviewersToPullRequest(credentials, prNumber, reviewers) {
        return __awaiter(this, void 0, void 0, function* () {
            const url = `${this.scheme}${this.apiBase}/repos/${this.owner}/${this.repo}`;
            const httpClient = configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory())
                .create(url);
            return httpClient.exchange(`${url}/pulls/${prNumber}/requested_reviewers`, Object.assign({ body: {
                    reviewers: reviewers.filter(r => r.type === RepoId_1.PullRequestReviewerType.individual).map(i => i.name),
                    team_reviewers: reviewers.filter(r => r.type === RepoId_1.PullRequestReviewerType.team).map(t => t.name),
                }, method: httpClient_1.HttpMethod.Post }, headers(credentials)));
        });
    }
    /**
     * Used to create a new Pull Request
     * https://developer.github.com/v3/pulls/#create-a-pull-request
     * @param {ProjectOperationCredentials} credentials
     * @param {string} title
     * @param {string} body
     * @param {string} head Source branch
     * @param {string} base Target branch
     * @param {PullRequestReviewer[]} reviewers
     */
    raisePullRequest(credentials, title, body, head, base, reviewers) {
        return __awaiter(this, void 0, void 0, function* () {
            const url = `${this.scheme}${this.apiBase}/repos/${this.owner}/${this.repo}`;
            const httpClient = configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory())
                .create(url);
            // Check if PR already exists on the branch
            const pr = yield this.getPr(credentials, head);
            if (!!pr && pr.length > 0) {
                try {
                    yield httpClient.exchange(`${url}/pulls/${pr[0].number}`, Object.assign({ body: { title }, method: httpClient_1.HttpMethod.Patch }, headers(credentials)));
                    yield httpClient.exchange(`${url}/issues/${pr[0].number}/comments`, Object.assign({ body: beautifyPullRequestBody(body), method: httpClient_1.HttpMethod.Post }, headers(credentials)));
                    return {
                        target: this,
                        success: true,
                    };
                }
                catch (e) {
                    logger_1.logger.error(`Error attempting to add PR comment. ${url}  ${e}`);
                    throw e;
                }
            }
            else {
                try {
                    yield httpClient.exchange(`${url}/pulls`, Object.assign({ body: {
                            title,
                            body: beautifyPullRequestBody(body),
                            head,
                            base,
                        }, method: httpClient_1.HttpMethod.Post }, headers(credentials)));
                }
                catch (e) {
                    logger_1.logger.error(`Error Attempting to Raise PR.  ${url} ${e}`);
                    throw e;
                }
                try {
                    if (reviewers) {
                        /**
                         * If there are reviewers retrieve the PR that was created and update the required reviewers
                         */
                        const thisPr = yield this.getPr(credentials, head);
                        yield this.addReviewersToPullRequest(credentials, thisPr[0].number, reviewers);
                    }
                }
                catch (e) {
                    logger_1.logger.error(`Error Attempting to Assign Reviewer(s) to PR.  ${url} ${e}`);
                    throw e;
                }
                return {
                    target: this,
                    success: true,
                };
            }
        });
    }
    deleteRemote(creds) {
        const url = `${this.scheme}${this.apiBase}/repos/${this.owner}/${this.repo}`;
        const httpClient = configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory())
            .create(url);
        return httpClient.exchange(url, Object.assign(Object.assign({}, headers(creds)), { method: httpClient_1.HttpMethod.Delete }))
            .then(() => ActionResult_1.successOn(this));
    }
}
exports.GitHubRepoRef = GitHubRepoRef;
function isGitHubRepoRef(rr) {
    const maybe = rr;
    return maybe && !!maybe.apiBase && maybe.kind === "github";
}
exports.isGitHubRepoRef = isGitHubRepoRef;
function headers(credentials) {
    if (!ProjectOperationCredentials_1.isTokenCredentials(credentials)) {
        throw new Error("Only token auth supported");
    }
    return {
        headers: {
            Authorization: `token ${credentials.token}`,
        },
    };
}
function apiBaseToRemoteBase(rawApiBase) {
    if (rawApiBase.includes("api.github.com")) {
        return "https://github.com";
    }
    if (rawApiBase.includes("api/v3")) {
        return rawApiBase.substring(0, rawApiBase.indexOf("api/v3"));
    }
    return rawApiBase;
}
// exported for testing
function beautifyPullRequestBody(body) {
    const tagRegEx = /(\[[-\w]+:[-\w:=\/\.]+\])/gm;
    let tagMatches = tagRegEx.exec(body);
    const tags = [];
    while (!!tagMatches) {
        tags.push(tagMatches[1]);
        tagMatches = tagRegEx.exec(body);
    }
    if (tags.length > 0) {
        const newBody = body.replace(/\[[-\w]+:[-\w:=\/\.]+\]/g, "")
            .replace(/\n\s*\n\s*\n/g, "\n\n")
            .trim();
        return `${newBody}

---
<details>
  <summary>Tags</summary>
<br/>
${tags.sort().map(t => `<code>${t}</code>`).join("<br/>")}
</details>`;
    }
    return body;
}
exports.beautifyPullRequestBody = beautifyPullRequestBody;
//# sourceMappingURL=GitHubRepoRef.js.map