"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const url = require("url");
const ActionResult_1 = require("../../action/ActionResult");
const configuration_1 = require("../../configuration");
const httpClient_1 = require("../../spi/http/httpClient");
const logger_1 = require("../../util/logger");
const AbstractRemoteRepoRef_1 = require("./AbstractRemoteRepoRef");
const validationPatterns_1 = require("./params/validationPatterns");
const RepoId_1 = require("./RepoId");
exports.GitlabDotComApiBase = "https://gitlab.com/api/v4";
exports.GitlabDotComRemoteUrl = "https://gitlab.com/";
/**
 * Repository reference implementation for Gitlab
 */
class GitlabRepoRef extends AbstractRemoteRepoRef_1.AbstractRemoteRepoRef {
    constructor(owner, repo, sha, apiBase = exports.GitlabDotComApiBase, gitlabRemoteUrl = exports.GitlabDotComRemoteUrl, path, branch) {
        super(apiBase === exports.GitlabDotComApiBase ? RepoId_1.ProviderType.gitlab_com : RepoId_1.ProviderType.gitlab_enterprise, gitlabRemoteUrl, apiBase, owner, repo, sha, path, branch);
        this.kind = "gitlab";
    }
    static from(params) {
        if (params.sha && !params.sha.match(validationPatterns_1.GitShaRegExp.pattern)) {
            throw new Error("You provided an invalid SHA: " + params.sha);
        }
        return new GitlabRepoRef(params.owner, params.repo, params.sha, params.rawApiBase, params.gitlabRemoteUrl, params.path, params.branch);
    }
    createRemote(creds, description, visibility) {
        return __awaiter(this, void 0, void 0, function* () {
            const gitlabUrl = `${this.scheme}${this.apiBase}/projects`;
            const httpClient = configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory()).create(gitlabUrl);
            const namespace = yield this.getNamespaceForOwner(this.owner, creds);
            return httpClient.exchange(gitlabUrl, {
                method: httpClient_1.HttpMethod.Post,
                body: {
                    name: `${this.repo}`,
                    visibility,
                    namespace_id: namespace,
                },
                headers: {
                    "Private-Token": creds.privateToken,
                    "Content-Type": "application/json",
                },
            }).then(response => {
                return {
                    target: this,
                    success: true,
                    response,
                };
            }).catch(err => {
                logger_1.logger.error(`Error attempting to create remote project, status code ${err.response.status}. ` +
                    `The response was ${JSON.stringify(err.response.data)}`);
                return Promise.reject(err);
            });
        });
    }
    deleteRemote(creds) {
        const gitlabUrl = `${this.scheme}${this.apiBase}/project/${this.owner}%2f${this.repo}`;
        const httpClient = configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory()).create(gitlabUrl);
        logger_1.logger.debug(`Making request to '${url}' to delete repo`);
        return httpClient.exchange(gitlabUrl, {
            method: httpClient_1.HttpMethod.Delete,
            headers: {
                "Private-Token": creds.privateToken,
                "Content-Type": "application/json",
            },
        }).then(response => {
            return {
                target: this,
                success: true,
                response,
            };
        }).catch(err => {
            logger_1.logger.error("Error attempting to delete repository: " + err);
            return Promise.reject(err);
        });
    }
    setUserConfig(credentials, project) {
        return Promise.resolve(ActionResult_1.successOn(this));
    }
    /**
     * Retrieve the Gitlab Server version
     *
     * @param {ProjectOperationCredentials} credentials
     * @returns {version: string, revision: string} Revision info
     */
    getGitlabVersion(credentials) {
        return __awaiter(this, void 0, void 0, function* () {
            const gitlabUrl = `${this.scheme}${this.apiBase}/version`;
            const httpClient = configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory()).create(gitlabUrl);
            logger_1.logger.debug(`Making request to '${gitlabUrl}' to get Gitlab Version`);
            const result = yield httpClient.exchange(gitlabUrl, {
                method: httpClient_1.HttpMethod.Get,
                headers: {
                    "Private-Token": credentials.privateToken,
                    "Content-Type": "application/json",
                },
            });
            return result.body;
        });
    }
    /**
     * Retrieve the id of the supplied group
     * @param {ProjectOperationCredentials} credentials
     * @param {string} name
     * @returns {number} The id of the requested group
     */
    getGroupId(credentials, name) {
        return __awaiter(this, void 0, void 0, function* () {
            const gitlabUrl = `${this.scheme}${this.apiBase}/groups/${encodeURI(name)}?with_projects=false`;
            const httpClient = configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory()).create(gitlabUrl);
            try {
                const result = yield httpClient.exchange(gitlabUrl, {
                    method: httpClient_1.HttpMethod.Get,
                    headers: {
                        "Private-Token": credentials.privateToken,
                        "Content-Type": "application/json",
                    },
                    retry: {
                        retries: 1,
                    },
                });
                return result.body.id;
            }
            catch (err) {
                if (err.response.status === 404) {
                    throw new Error(`Failed to resolve ID for group ${name}.  Does it exist?`);
                }
                else {
                    throw err;
                }
            }
        });
    }
    /**
     * Retrieve the id of the supplied user
     * @param {ProjectOperationCredentials} credentials
     * @param {string} name
     * @returns {number} The id of the user
     */
    getUserId(credentials, name) {
        return __awaiter(this, void 0, void 0, function* () {
            const gitlabUrl = `${this.scheme}${this.apiBase}/users?username=${encodeURI(name)}`;
            const httpClient = configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory()).create(gitlabUrl);
            const result = yield httpClient.exchange(gitlabUrl, {
                method: httpClient_1.HttpMethod.Get,
                headers: {
                    "Private-Token": credentials.privateToken,
                    "Content-Type": "application/json",
                },
            });
            if (result.body === undefined || Array.isArray(result.body) && result.body.length === 0) {
                throw new Error(`Failed retrieve Gitlab user id for ${name}; does user exist?`);
            }
            return result.body[0].id;
        });
    }
    /**
     * Convert reviewers from names to ids
     *
     * @param {ProjectOperationCredentials} credentials
     * @param {PullRequestReviewer[]} reviewers
     * @returns {users: number[], groups: number[]} The user and group ids requested for MR approval
     */
    resolveApprovers(credentials, reviewers) {
        return __awaiter(this, void 0, void 0, function* () {
            let userIds;
            let groupIds;
            try {
                userIds = yield Promise.all(reviewers.filter(r => r.type === RepoId_1.PullRequestReviewerType.individual).map((i) => __awaiter(this, void 0, void 0, function* () { return this.getUserId(credentials, i.name); })));
            }
            catch (err) {
                throw new Error(`Couldn't resolve ID(s) for supplied Merge Request approval users. ` + err);
            }
            try {
                groupIds = yield Promise.all(reviewers.filter(r => r.type === RepoId_1.PullRequestReviewerType.team).map((g) => __awaiter(this, void 0, void 0, function* () { return this.getGroupId(credentials, g.name); })));
            }
            catch (err) {
                throw new Error(`Couldn't resolve ID(s) for supplied Merge Request approval groups. ` + err);
            }
            return { users: userIds, groups: groupIds };
        });
    }
    /**
     * Add Approvers to Merge Request
     *
     * @param {ProjectOperationCredentials} credentials
     * @param {PullRequestReviewer[]} reviewers
     * @param {number} mrNumber The iid of the merge request
     * @param {number} projectId The project id in Gitlab
     */
    addApproversToMergeRequest(credentials, reviewers, mrNumber, projectId) {
        return __awaiter(this, void 0, void 0, function* () {
            const gitlabVersion = yield this.getGitlabVersion(credentials);
            const versionDetails = gitlabVersion.version.split(".").map(v => parseInt(v, undefined));
            const httpClient = configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory())
                .create(`${this.scheme}${this.apiBase}`);
            if (versionDetails[0] < 10 || (versionDetails[0] === 10 && versionDetails[1] < 6)) {
                /**
                 * For versions less then 10.6 there is no support for this
                 */
                throw new Error(`Cannot set merge request approvers, Gitlab version ${gitlabVersion.version} not supported!`);
            }
            else if (versionDetails[0] < 12 || (versionDetails[0] === 12 && versionDetails[1] < 3)) {
                /**
                 * For Gitlab versions less then 12.3, use PUT /projects/:id/merge_requests/:merge_request_iid/approvers API
                 */
                // Set the approvers
                const gitlabUrl = `${this.scheme}${this.apiBase}/projects/${projectId}/merge_requests/${mrNumber}/approvers`;
                const approvers = yield this.resolveApprovers(credentials, reviewers);
                yield httpClient.exchange(gitlabUrl, {
                    method: httpClient_1.HttpMethod.Put,
                    body: {
                        id: projectId,
                        iid: mrNumber,
                        approver_ids: approvers.users,
                        approver_group_ids: approvers.groups,
                    },
                    headers: {
                        "Private-Token": credentials.privateToken,
                        "Content-Type": "application/json",
                    },
                });
                // Set the required approvals equal to the requested reviewers/groups
                const gitlabApprovalUrl = `${this.scheme}${this.apiBase}/projects/${projectId}/merge_requests/${mrNumber}/approvals`;
                yield httpClient.exchange(gitlabApprovalUrl, {
                    method: httpClient_1.HttpMethod.Post,
                    body: {
                        id: projectId,
                        iid: mrNumber,
                        approvals_required: approvers.groups.length + approvers.users.length,
                    },
                    headers: {
                        "Private-Token": credentials.privateToken,
                        "Content-Type": "application/json",
                    },
                });
            }
            else {
                /**
                 * For Gitlab version 12.3 or higher, use POST /projects/:id/merge_requests/:merge_request_iid/approval_rules
                 *
                 * Note: This will override the project level rule(s)!
                 */
                const gitlabUrl = `${this.scheme}${this.apiBase}/projects/${projectId}/merge_requests/${mrNumber}/approval_rules`;
                const approvers = yield this.resolveApprovers(credentials, reviewers);
                yield httpClient.exchange(gitlabUrl, {
                    method: httpClient_1.HttpMethod.Post,
                    body: {
                        id: projectId,
                        merge_request_iid: mrNumber,
                        name: `Atomist Generated Rule`,
                        approvals_required: approvers.groups.length + approvers.users.length,
                        user_ids: approvers.users,
                        group_ids: approvers.groups,
                    },
                    headers: {
                        "Private-Token": credentials.privateToken,
                        "Content-Type": "application/json",
                    },
                });
            }
        });
    }
    raisePullRequest(credentials, title, body, head, base, reviewers) {
        return __awaiter(this, void 0, void 0, function* () {
            const gitlabUrl = `${this.scheme}${this.apiBase}/projects/${this.owner}%2f${this.repo}/merge_requests`;
            const httpClient = configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory()).create(gitlabUrl);
            logger_1.logger.debug(`Making request to '${gitlabUrl}' to raise PR`);
            let response;
            try {
                response = yield httpClient.exchange(gitlabUrl, {
                    method: httpClient_1.HttpMethod.Post,
                    body: {
                        id: `${this.owner}%2f${this.repo}`,
                        title,
                        description: body,
                        source_branch: head,
                        target_branch: base,
                    },
                    headers: {
                        "Private-Token": credentials.privateToken,
                        "Content-Type": "application/json",
                    },
                });
            }
            catch (err) {
                throw new Error(`Failed to raise PR, status code ${err.response.status}. ` +
                    `The response was ${JSON.stringify(err.response.data)}`);
            }
            if (reviewers) {
                try {
                    yield this.addApproversToMergeRequest(credentials, reviewers, response.body.iid, response.body.project_id);
                }
                catch (err) {
                    throw new Error(`Failed to add reviewers to Merge Request. ${err}`);
                }
            }
            return {
                target: this,
                success: true,
            };
        });
    }
    getNamespaceForOwner(owner, creds) {
        const gitlabUrl = `${this.scheme}${this.apiBase}/namespaces?search=${encodeURI(owner)}`;
        const httpClient = configuration_1.configurationValue("http.client.factory", httpClient_1.defaultHttpClientFactory()).create(gitlabUrl);
        return httpClient.exchange(gitlabUrl, {
            method: httpClient_1.HttpMethod.Get,
            headers: {
                "Private-Token": creds.privateToken,
                "Content-Type": "application/json",
            },
        }).then(response => {
            const namespaces = response.body;
            const ownerNamespace = namespaces.filter(n => n.name === owner)[0];
            if (!!ownerNamespace) {
                return ownerNamespace.id;
            }
            else {
                return Promise.reject("Cannot find Gitlab namespace with name " + owner);
            }
        });
    }
}
exports.GitlabRepoRef = GitlabRepoRef;
//# sourceMappingURL=GitlabRepoRef.js.map