"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const base64_1 = require("../internal/util/base64");
const GitHubRepoRef_1 = require("../operations/common/GitHubRepoRef");
const httpClient_1 = require("../spi/http/httpClient");
const http_1 = require("./http");
const logger_1 = require("./logger");
/**
 * Return a deep link to the file location
 * @param {GitHubRepoRef} grr
 * @param {SourceLocation} sourceLocation
 * @return {string}
 */
function deepLink(grr, sourceLocation) {
    return `${grr.scheme}${grr.remoteBase}/${grr.owner}/${grr.repo}/blob/${grr.sha}` +
        (!!sourceLocation ? `/${sourceLocation.path}` : "") +
        (!!sourceLocation && !!sourceLocation.lineFrom1 ? `#L${sourceLocation.lineFrom1}` : "");
}
exports.deepLink = deepLink;
/**
 * Check whether the given file, including path, exists
 * in the GitHub repo
 * @param token GitHub token. Don't try to auth if it's undefined or null
 * @param user
 * @param repo
 * @param path
 * @return {Promise<boolean|T>}
 */
function hasFile(token, user, repo, path) {
    return __awaiter(this, void 0, void 0, function* () {
        // We only care if it returns 200. Otherwise it isn't there
        try {
            yield filePromise(token, user, repo, path);
            return true;
        }
        catch (e) {
            return false;
        }
    });
}
exports.hasFile = hasFile;
/**
 * Return file content, or undefined if it's not found
 * @param {string} token
 * @param {string} user
 * @param {string} repo
 * @param {string} path
 * @return {Promise<string>}
 */
function fileContent(token, user, repo, path) {
    return filePromise(token, user, repo, path)
        .then(d => base64_1.decode(d.body.content))
        .catch(() => {
        return undefined;
    });
}
exports.fileContent = fileContent;
function filePromise(token, user, repo, path) {
    const url = `${GitHubRepoRef_1.GitHubDotComBase}/repos/${user}/${repo}/contents/${path}`;
    logger_1.logger.debug(`Request to '${url}' to check for file existence]`);
    // We only care if it returns 200. Otherwise it isn't there
    return http_1.httpClient(url).exchange(url, Object.assign(Object.assign({ method: httpClient_1.HttpMethod.Get }, authHeaders(token)), { retry: { retries: 0 } }));
}
function raiseIssue(token, rr, issue) {
    const grr = GitHubRepoRef_1.isGitHubRepoRef(rr) ? rr : new GitHubRepoRef_1.GitHubRepoRef(rr.owner, rr.repo, rr.sha);
    const url = `${grr.scheme}${grr.apiBase}/repos/${rr.owner}/${rr.repo}/issues`;
    logger_1.logger.debug(`Request to '${url}' to raise issue`);
    return http_1.httpClient(url).exchange(url, Object.assign({ method: httpClient_1.HttpMethod.Post, body: issue }, authHeaders(token)));
}
exports.raiseIssue = raiseIssue;
function createCommitComment(token, rr, comment) {
    const url = `${rr.scheme}${rr.apiBase}/repos/${rr.owner}/${rr.repo}/commits/${rr.sha}/comments`;
    logger_1.logger.debug(`Request to '${url}' to create comment`);
    return http_1.httpClient(url).exchange(url, Object.assign({ method: httpClient_1.HttpMethod.Post, body: comment }, authHeaders(token)));
}
exports.createCommitComment = createCommitComment;
function createRepo(token, rr, description, priv) {
    return __awaiter(this, void 0, void 0, function* () {
        const client = http_1.httpClient(`${rr.scheme}${rr.apiBase}`);
        let repoUrl;
        // check if owner is a user
        const userOrOrg = yield client.exchange(`${rr.scheme}${rr.apiBase}/users/${rr.owner}`, Object.assign({ method: httpClient_1.HttpMethod.Get }, authHeaders(token)));
        if (userOrOrg.body.type === "User") {
            repoUrl = `${rr.scheme}${rr.apiBase}/user/repos`;
        }
        else if (userOrOrg.body.type === "Organization") {
            repoUrl = `${rr.scheme}${rr.apiBase}/orgs/${rr.owner}/repos`;
        }
        const payload = {
            name: rr.repo,
            description,
            private: priv,
        };
        logger_1.logger.debug(`Request to '${repoUrl}' to create repo`);
        return client.exchange(repoUrl, Object.assign(Object.assign({ method: httpClient_1.HttpMethod.Post, body: payload }, authHeaders(token)), { retry: { retries: 0 } }));
    });
}
exports.createRepo = createRepo;
function authHeaders(token) {
    return token ? {
        headers: {
            Authorization: `token ${token}`,
        },
    }
        : {};
}
//# sourceMappingURL=gitHub.js.map