"use strict";
/*
 * Copyright © 2018 Atomist, Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const spawn = require("cross-spawn");
exports.spawn = spawn;
const os = require("os");
const path = require("path");
const strip_ansi_1 = require("strip-ansi");
const treeKill = require("tree-kill");
const child_process_1 = require("./child_process");
const logger_1 = require("./logger");
/**
 * Default ErrorFinder that regards return code 0 as success
 * @param {number} code
 * @return {boolean}
 * @deprecated use @atomist/sdm version
 */
exports.SuccessIsReturn0ErrorFinder = code => code !== 0;
/**
 * Spawn a process, log its output, and wait for it to exit,
 * asynchronously.  It is spawned using cross-spawn.
 *
 * @param {SpawnCommand} spawnCommand command to run
 * @param options standard spawn options
 * @param log log to write output to
 * @param {Partial<SpawnWatchOptions>} spOpts
 * @return {Promise<ChildProcessResult>}
 * @deprecated use @atomist/sdm spawnAndLog
 */
function spawnAndWatch(spawnCommand, options, log, spOpts = {}) {
    return __awaiter(this, void 0, void 0, function* () {
        const childProcess = spawn(spawnCommand.command, spawnCommand.args || [], options);
        if (spOpts.logCommand === false) {
            logger_1.logger.debug(`${options.cwd || path.resolve(".")} > ${stringifySpawnCommand(spawnCommand)} (pid '${childProcess.pid}')`);
        }
        else {
            log.write(`/--\n`);
            log.write(`${options.cwd || path.resolve(".")} > ${stringifySpawnCommand(spawnCommand)} (pid '${childProcess.pid}')\n`);
            log.write(`\\--\n`);
        }
        return watchSpawned(childProcess, log, spOpts);
    });
}
exports.spawnAndWatch = spawnAndWatch;
/**
 * Handle the result of a spawned process, streaming back
 * output to log
 * @param childProcess
 * @param log to write stdout and stderr to
 * @param opts: Options for error parsing, ANSI code stripping etc.
 * @return {Promise<ChildProcessResult>}
 * @deprecated use @atomist/sdm spawnAndLog
 */
function watchSpawned(childProcess, log, opts = {}) {
    return __awaiter(this, void 0, void 0, function* () {
        let timer;
        let running = true;
        if (opts.timeout) {
            timer = setTimeout(() => {
                if (running) {
                    logger_1.logger.warn("Spawn timeout expired. Killing command with pid '%s'", childProcess.pid);
                    child_process_1.killProcess(childProcess.pid);
                }
            }, opts.timeout);
        }
        return new Promise((resolve, reject) => {
            const optsToUse = Object.assign({ errorFinder: exports.SuccessIsReturn0ErrorFinder, stripAnsi: false }, opts);
            if (!optsToUse.errorFinder) {
                // The caller specified undefined, which is an error. Ignore them, for they know not what they do.
                optsToUse.errorFinder = exports.SuccessIsReturn0ErrorFinder;
            }
            function sendToLog(data) {
                const formatted = optsToUse.stripAnsi ? strip_ansi_1.default(data.toString()) : data.toString();
                log.write(formatted);
            }
            childProcess.stdout.on("data", sendToLog);
            childProcess.stderr.on("data", sendToLog);
            childProcess.addListener("exit", (code, signal) => {
                running = false;
                logger_1.logger.debug("Spawn exit with pid '%d': code '%d', signal '%d'", childProcess.pid, code, signal);
                clearTimeout(timer);
                resolve({
                    error: optsToUse.errorFinder(code, signal, log),
                    code,
                    childProcess,
                });
            });
            childProcess.addListener("error", err => {
                running = false;
                // Process could not be spawned or killed
                logger_1.logger.warn("Spawn failure: %s", err);
                clearTimeout(timer);
                reject(err);
            });
        });
    });
}
/**
 * toString for a SpawnCommand. Used for logging.
 * @param {SpawnCommand} sc
 * @return {string}
 * @deprecated use childProcessString
 */
function stringifySpawnCommand(sc) {
    return `${sc.command}${(!!sc.args) ? " '" + sc.args.join("' '") + "'" : ""}`;
}
exports.stringifySpawnCommand = stringifySpawnCommand;
/**
 * Convenience function to create a spawn command from a sentence such
 * as "npm run compile" Does not respect quoted arguments.  Use
 * spawnAndWatch passing it the command and argument array if your
 * command arguments have spaces, etc.
 *
 * @param {string} sentence command and argument string
 * @param options
 * @return {SpawnCommand}
 * @deprecated just pass the proper arguments to @atomist/sdm spawnAndLog
 */
function asSpawnCommand(sentence, options = {}) {
    const split = sentence.split(" ");
    return {
        command: split[0],
        args: split.slice(1),
        options,
    };
}
exports.asSpawnCommand = asSpawnCommand;
/**
 * Kill the child process and wait for it to shut down. This can take
 * a while as child processes may have shut down hooks.  On win32,
 * tree-kill is used and the Promise is rejected if the process(es) do(es)
 * not exit within `wait` milliseconds.  On other platforms, first the
 * child process is sent the default signal, SIGTERM.  After `wait`
 * milliseconds, it is sent SIGKILL.  After another `wait`
 * milliseconds, an error is thrown.
 *
 * @param {module:child_process.ChildProcess} childProcess
 * @param wait the number of milliseconds to wait before sending SIGKILL and
 *             then erroring, default is 30000 ms
 * @return {Promise<any>}
 * @deprecated use @atomist/sdm killAndWait
 */
function poisonAndWait(childProcess, wait = 30000) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise((resolve, reject) => {
            const pid = childProcess.pid;
            const termTimer = setTimeout(() => {
                if (os.platform() === "win32") {
                    reject(new Error(`Failed to tree-kill child process ${pid} in ${wait} ms`));
                }
                else {
                    logger_1.logger.debug(`Child process ${pid} did not exit in ${wait} ms, sending SIGKILL`);
                    childProcess.kill("SIGKILL");
                }
            }, wait);
            const killTimer = (os.platform() === "win32") ? undefined : setTimeout(() => {
                reject(new Error(`Failed to kill child process ${pid}`));
            }, 2 * wait);
            childProcess.on("close", clearAndResolve(pid, resolve, termTimer, killTimer));
            childProcess.on("exit", (code, signal) => {
                logger_1.logger.debug(`Child process ${pid} exited with code '${code}' and signal '${signal}'`);
            });
            childProcess.on("error", clearAndReject(pid, reject, termTimer, killTimer));
            child_process_1.killProcess(pid);
        });
    });
}
exports.poisonAndWait = poisonAndWait;
/**
 * Cross-platform kill.  On win32, tree-kill is used and signal is
 * ignored since win32 does not support different signals.  On other
 * platforms, ChildProcess.kill(signal) is used.
 *
 * @param cp child process to kill
 * @param signal optional signal, Node.js default is used if not provided
 * @deprecated use killProcess
 */
function crossKill(cp, signal) {
    if (os.platform() === "win32") {
        logger_1.logger.debug(`Calling tree-kill on child process ${cp.pid}`);
        treeKill(cp.pid);
    }
    else {
        const sig = (signal) ? `signal ${signal}` : "default signal";
        logger_1.logger.debug(`Sending ${sig} to child process ${cp.pid}`);
        cp.kill(signal);
    }
}
exports.crossKill = crossKill;
/**
 * Clear provided timers and resolve a promise.
 * @deprecated not used by spawnPromise
 */
function clearAndResolve(pid, resolve, ...timers) {
    return (code, signal) => {
        logger_1.logger.debug(`Child process ${pid} closed with code '${code}' and signal '${signal}'`);
        clearTimers(timers);
        resolve();
    };
}
/**
 * Clear provided timers and reject a promise.
 * @deprecated not used by spawnPromise
 */
function clearAndReject(pid, reject, ...timers) {
    return (reason) => {
        logger_1.logger.error(`Child process ${pid} errored: ${reason.message}`);
        clearTimers(timers);
        reject(reason);
    };
}
/**
 * Clear provided timers.  It checks to make sure the timers are
 * defined before clearing them.
 *
 * @param timers the timers to clear.
 * @deprecated only used by deprecated functions
 */
function clearTimers(timers) {
    timers.filter(t => !!t).map(clearTimeout);
}
//# sourceMappingURL=spawn.js.map