"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const Concat_1 = require("./matchers/Concat");
const ChangeSet_1 = require("./internal/ChangeSet");
const DefaultInputState_1 = require("./internal/DefaultInputState");
const ExactMatch_1 = require("./internal/ExactMatch");
const InputStateManager_1 = require("./internal/InputStateManager");
const MicrogrammarSpecParser_1 = require("./internal/MicrogrammarSpecParser");
const MicrogrammarUpdates_1 = require("./internal/MicrogrammarUpdates");
const Whitespace_1 = require("./internal/Whitespace");
const MatchReport_1 = require("./MatchReport");
const StringInputStream_1 = require("./spi/StringInputStream");
/**
 * Holds a set of updatable matches
 */
class Updatable {
    constructor(hits, content) {
        const mut = new MicrogrammarUpdates_1.MicrogrammarUpdates();
        this.cs = new ChangeSet_1.ChangeSet(content);
        this.matches = hits.map(m => mut.updatableMatch(m, this.cs));
    }
    updated() {
        return this.cs.updated();
    }
}
exports.Updatable = Updatable;
/**
 * Central class for microgrammar usage.
 * Represents a microgrammar that we can use to match input
 * in a string or stream.
 * Modifications are tracked and we can get an updated string
 * afterwards.
 */
class Microgrammar {
    constructor(matcher) {
        this.matcher = matcher;
        this.definitions = this.matcher.definitions;
    }
    /**
     * Make this match transparently updatable using property mutation
     * @param match match to make updatable
     * @param content the match is within
     * @return {T&MatchUpdater}
     */
    static updatableMatch(match, content) {
        return new MicrogrammarUpdates_1.MicrogrammarUpdates().updatableMatch(match, content);
    }
    /**
     * Make these matches transparently updatable using property mutation
     * @param matches matches
     * @param content content the matches are within
     * @return {Updatable<T>}
     */
    static updatable(matches, content) {
        return new Updatable(matches, content);
    }
    /**
     * Create a microgrammar with typed properties according
     * to the given interface.
     * If the definitions aren't nested, infer string type
     * @param {TermsDefinition<T>} definitions
     * @return {Microgrammar<T>}
     */
    static fromDefinitions(definitions) {
        return new Microgrammar(Concat_1.Concat.of(definitions));
    }
    /**
     * Create a microgrammar with inferred interface taking properties of type "any" from definitions.
     * Use fromDefinitions for stronger typing.
     * If the definitions aren't nested, infer string type
     * @param {TermsDefinition<T>} definitions
     * @return {Microgrammar<T>}
     */
    static fromDefinitionsAs(definitions) {
        return this.fromDefinitions(definitions);
    }
    /**
     * Create a microgrammar with string variables.
     * String is of form "method ${name}(): ${returnType}".
     * Definitions should be provided for each string variable.
     * Use fromDefinitions to achieve nesting or non-string typing.
     * If the definitions aren't nested, infer string type
     * @return {Microgrammar<T>}
     */
    static fromString(spec, components = {}, options = {}) {
        return new Microgrammar(new MicrogrammarSpecParser_1.MicrogrammarSpecParser().fromString(spec, components, options));
    }
    /**
     * Create a microgrammar with string variables with automatic typing as in
     * fromDefinitionsAs.
     * String is of form "method ${name}(): ${returnType}".
     * Definitions should be provided for each string variable.
     * Use fromDefinitions to achieve nesting or non-string typing.
     * If the definitions aren't nested, infer string type
     * @return {Microgrammar<T>}
     */
    static fromStringAs(spec, components = {}, options = {}) {
        return this.fromString(spec, components, options);
    }
    /**
     * Generator for matching the given input.
     * @param {string | InputStream} input
     * @param {{}} parseContext
     * @param {Listeners} l
     * @return {Iterable<PatternMatch>}
     */
    matchIterator(input, parseContext = {}, l) {
        const mrIterable = matchesIn(this, input, parseContext, l);
        const newIt = function* () {
            for (const elt of mrIterable) {
                yield elt.toPatternMatch();
            }
        };
        return newIt();
    }
    matchReportIterator(input, parseContext = {}, l) {
        return matchesIn(this, input, parseContext, l);
    }
    /**
     * Convenience method to find matches without the ability to update them
     * @param input
     * @param stopAfterMatch() function that can cause matching to stop after a given match.
     * Often used to stop after one.
     * @param parseContext context for the whole parsing operation
     * @param l listeners observing input characters as they are read
     * @return {PatternMatch[]}
     */
    findMatches(input, parseContext, l, stopAfterMatch = () => false) {
        const lm = new LazyMatcher(this.matcher, stopAfterMatch);
        lm.consume(input, parseContext, l);
        return lm.matches;
    }
    /**
     * Parallel to findMatches, but returns a Promise
     * @param {string | InputStream} input
     * @param {{}} parseContext
     * @return {Promise<Array<T & PatternMatch>>}
     */
    findMatchesAsync(input, parseContext) {
        return __awaiter(this, void 0, void 0, function* () {
            return (yield this.findMatchReportsAsync(input, parseContext)).map(mr => mr.toPatternMatch());
        });
    }
    findMatchReportsAsync(input, parseContext) {
        return __awaiter(this, void 0, void 0, function* () {
            const matches = [];
            for (const m of matchesIn(this.matcher, input, parseContext)) {
                matches.push(m);
            }
            return matches;
        });
    }
    /**
     * Convenient method to find the first match, or null if not found.
     * Stops searching after the first match.
     * @param input
     * @param l listeners observing input characters as they are read
     * @returns {PatternMatch[]}
     */
    firstMatch(input, l) {
        const found = this.findMatches(input, {}, l, pm => true);
        return found.length > 0 ? found[0] : null;
    }
    /**
     * Return a match if it explains the whole of the input.
     * This style of usage is more like a traditional parser,
     * building an AST for a whole file.
     * @param input
     * @param parseContext context for the whole parsing operation
     * @param l listeners observing input characters as they are read
     * @return {PatternMatch&T}
     * @Deprecated prefer perfectMatch
     */
    exactMatch(input, parseContext = {}, l) {
        return MatchReport_1.toPatternMatchOrDismatchReport(this.perfectMatch(input, parseContext, l));
    }
    /**
     * Return a MatchReport. if it is successful, call .toValueStructure<T>() to get the match value.
     * Otherwise, call .toExplanationTree() to get a report of why it didn't match.
     * @param input
     * @param parseContext context for the whole parsing operation
     * @param l listeners observing input characters as they are read
     */
    perfectMatch(input, parseContext = {}, l) {
        return ExactMatch_1.perfectMatch(this.matcher, input, parseContext, l);
    }
}
exports.Microgrammar = Microgrammar;
/**
 * Single use, usually stateful, class for matching input.
 * Offers the ability to observe a match, as well as match one,
 * and to change the matcher in use depending on observation and matching.
 * E.g. it's possible to choose to start matching pattern B after finding pattern A,
 * or after *seeing* pattern A, even if starting off matching something else.
 * This enables us, for example, to parse XML, with the observer watching element
 * open and close to maintain the current path, while the matcher matches anything we want.
 */
class MatchingMachine {
    /**
     * Create a new stateful matching machine
     * @param initialMatcher matcher to start using. This can be changed by the callback methods in this class
     * @param o optional observer
     */
    constructor(initialMatcher, o) {
        this.matcher = Concat_1.toMatchingLogic(initialMatcher);
        if (!!o) {
            this.observer = Concat_1.toMatchingLogic(o);
        }
        this.omg = this.observer ? Microgrammar.fromDefinitions(this.observer) : undefined;
    }
    /**
     * Stream-oriented matching. The observer can match in parallel with the main matcher.
     * @param input
     * @param parseContext context for the whole parsing operation
     * @param l listeners observing input characters as they are read
     */
    consume(input, parseContext = {}, l) {
        let currentMatcher = this.matcher;
        const stream = toInputStream(input);
        const stateManager = new InputStateManager_1.InputStateManager(stream);
        let currentInputState = new DefaultInputState_1.DefaultInputState(stateManager, 0, l);
        while (currentMatcher && !currentInputState.exhausted()) {
            currentInputState = Whitespace_1.readyToMatch(currentInputState, this.matcher.$consumeWhiteSpaceBetweenTokens === true, currentMatcher, this.observer).state;
            const previousIs = currentInputState;
            const tryMatch = currentMatcher.matchPrefixReport(currentInputState, {}, parseContext);
            // We can't accept empty matches as genuine at this level:
            // For example, if the matcher is just a Rep or Alt
            if (MatchReport_1.isSuccessfulMatchReport(tryMatch) && tryMatch.matched !== "") {
                const match = tryMatch.toPatternMatch();
                // Enrich with the name
                match.$name = match.$matcherId;
                currentMatcher = Concat_1.toMatchingLogic(this.onMatch(match)); // this sneakily puts it in class instance state
                currentInputState = currentInputState.consume(match.$matched, `Microgrammar after match on [${match.$matched} from [${match.$matcherId}]`);
            }
            else {
                // We didn't match. Discard the current input character and try again
                if (!currentInputState.exhausted()) {
                    currentInputState = currentInputState.advance();
                }
            }
            if (this.observer) {
                // There are two cases: If we matched, we need to look multiple times in the input
                if (MatchReport_1.isSuccessfulMatchReport(tryMatch) && this.omg) {
                    const matches = this.omg.findMatches(tryMatch.matched);
                    for (const m of matches) {
                        currentMatcher = Concat_1.toMatchingLogic(this.observeMatch(m));
                    }
                }
                else {
                    const observerMatch = this.observer.matchPrefixReport(previousIs, {}, parseContext);
                    if (MatchReport_1.isSuccessfulMatchReport(observerMatch)) {
                        currentMatcher = Concat_1.toMatchingLogic(this.observeMatch(observerMatch.toPatternMatch()));
                    }
                }
            }
            // We can advance the window
            stateManager.dropLeft(currentInputState.offset);
        } // while
    }
    /**
     * Observe a match. The return can change the matcher in use, or return the current matcher.
     * @param pm pattern to observe
     * @returns {MatchingLogic}
     */
    observeMatch(pm) {
        return this.matcher;
    }
    /**
     * React to a match. The return can change the matcher, or return the current matcher.
     * @param pm matcher
     */
    onMatch(pm) {
        return this.matcher;
    }
}
exports.MatchingMachine = MatchingMachine;
function toInputStream(input) {
    return (typeof input === "string") ?
        new StringInputStream_1.StringInputStream(input) :
        input;
}
class LazyMatcher extends MatchingMachine {
    constructor(ml, stopAfterMatch) {
        super(ml);
        this.stopAfterMatch = stopAfterMatch;
        this.matches = [];
    }
    onMatch(pm) {
        this.matches.push(pm);
        return this.stopAfterMatch(pm) ? undefined : this.matcher;
    }
}
/**
 * Generator for matching the given input.
 * @param matcher
 * @param {string | InputStream} input
 * @param {{}} parseContext
 * @param {Listeners} l
 * @return {Iterable<PatternMatch>}
 */
function* matchesIn(matcher, input, parseContext = {}, l) {
    const matchingLogic = Concat_1.toMatchingLogic(matcher);
    const stream = toInputStream(input);
    const stateManager = new InputStateManager_1.InputStateManager(stream);
    let currentInputState = new DefaultInputState_1.DefaultInputState(stateManager, 0, l);
    while (matchingLogic && !currentInputState.exhausted()) {
        currentInputState = Whitespace_1.readyToMatch(currentInputState, matchingLogic.$consumeWhiteSpaceBetweenTokens === true, matchingLogic).state;
        const tryMatch = matchingLogic.matchPrefixReport(currentInputState, {}, parseContext);
        // We can't accept empty matches as genuine at this level:
        // For example, if the matcher is just a Rep or Alt
        if (MatchReport_1.isSuccessfulMatchReport(tryMatch) && tryMatch.matched !== "") {
            const m = tryMatch.toPatternMatch();
            // Enrich with the name
            m.$name = m.$matcherId;
            yield tryMatch;
            //  = toMatchingLogic(this.onMatch(match));
            currentInputState = currentInputState.consume(m.$matched, `Microgrammar after match on [${m.$matched} from [${m.$matcherId}]`);
        }
        else {
            // We didn't match. Discard the current input character and try again
            if (!currentInputState.exhausted()) {
                currentInputState = currentInputState.advance();
            }
        }
        // We can advance the window
        stateManager.dropLeft(currentInputState.offset);
    }
}
exports.matchesIn = matchesIn;
//# sourceMappingURL=Microgrammar.js.map