"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const failedMatchReport_1 = require("./internal/matchReport/failedMatchReport");
const terminalMatchReport_1 = require("./internal/matchReport/terminalMatchReport");
const wrappingMatchReport_1 = require("./internal/matchReport/wrappingMatchReport");
const Concat_1 = require("./matchers/Concat");
const MatchReport_1 = require("./MatchReport");
/**
 * Optional match on the given matcher
 * @param o matcher
 * @return {Opt}
 */
function optional(o) {
    return new Opt(o);
}
exports.optional = optional;
class Opt {
    /**
     * Optional match
     * @param o matching logic
     */
    constructor(o) {
        this.parseNodeName = "Optional";
        this.matcher = Concat_1.toMatchingLogic(o);
    }
    get $id() {
        return `Opt[${this.matcher.$id}]`;
    }
    matchPrefix(is, thisMatchContext, parseContext) {
        return MatchReport_1.toMatchPrefixResult(this.matchPrefixReport(is, thisMatchContext, parseContext));
    }
    matchPrefixReport(is, thisMatchContext, parseContext) {
        if (is.exhausted()) {
            // console.log(`Match from Opt on exhausted stream`);
            return terminalMatchReport_1.successfulMatchReport(this, {
                matched: "",
                offset: is.offset,
                parseNodeName: this.parseNodeName,
                valueRepresented: { value: undefined },
                reason: "Input stream is exhausted. Good thing this was optional.",
            });
        }
        const maybe = this.matcher.matchPrefixReport(is, thisMatchContext, parseContext);
        if (MatchReport_1.isSuccessfulMatchReport(maybe)) {
            return wrappingMatchReport_1.wrappingMatchReport(this, { parseNodeName: this.parseNodeName, inner: maybe });
        }
        return new WrappingEmptyMatchReport(this, this.parseNodeName, maybe);
    }
}
exports.Opt = Opt;
class WrappingEmptyMatchReport {
    constructor(matcher, parseNodeName, inner) {
        this.matcher = matcher;
        this.parseNodeName = parseNodeName;
        this.inner = inner;
        this.successful = true;
        this.kind = "real";
        this.matched = "";
        this.offset = inner.offset;
        this.endingOffset = inner.offset;
    }
    toPatternMatch() {
        const pm = {
            $matcherId: this.matcher.$id,
            $matched: this.matched,
            $offset: this.offset,
            $value: undefined,
            matchedStructure: () => undefined,
        };
        // hack for compatibility with isSuccessfulMatch
        pm.$successfulMatch = true;
        return pm;
    }
    toParseTree() {
        return {
            $name: this.parseNodeName,
            $value: this.matched,
            $offset: this.offset,
            $children: [],
        };
    }
    toValueStructure() {
        return undefined;
    }
    toExplanationTree() {
        return {
            reason: "Did not match, but that's OK; it's optional.",
            $name: this.parseNodeName,
            $value: this.matched,
            $offset: this.offset,
            $children: [this.inner.toExplanationTree()],
            successful: true,
        };
    }
}
/**
 * Match the first of these matchers that matches. Equivalent to an Alt (alternate)
 * @param a first matcher
 * @param b second matcher
 * @param matchers any further matchers: varargs
 * @returns {Alt}
 */
function firstOf(a, b, ...matchers) {
    return new Alt(a, b, ...matchers);
}
exports.firstOf = firstOf;
/**
 * Matches first match of 2 or more matchers.
 */
class Alt {
    constructor(a, b, ...matchers) {
        this.parseNodeName = "Alternative";
        const matchObjects = [a, b].concat(matchers);
        this.matchers = matchObjects.map(Concat_1.toMatchingLogic);
    }
    // tslint:disable-next-line:member-ordering
    get $id() {
        return `Alt(${this.matchers.map(m => m.$id).join(",")})`;
    }
    matchPrefix(is, thisMatchContext, parseContext) {
        return MatchReport_1.toMatchPrefixResult(this.matchPrefixReport(is, thisMatchContext, parseContext));
    }
    matchPrefixReport(is, thisMatchContext, parseContext) {
        if (is.exhausted()) {
            return failedMatchReport_1.failedMatchReport(this, {
                offset: is.offset,
                parseNodeName: this.parseNodeName,
                reason: "Input exhausted",
            });
        }
        const failedMatches = [];
        for (const matcher of this.matchers) {
            const m = matcher.matchPrefixReport(is, thisMatchContext, parseContext);
            if (MatchReport_1.isSuccessfulMatchReport(m)) {
                return wrappingMatchReport_1.wrappingMatchReport(this, {
                    inner: m,
                    additional: failedMatches,
                    parseNodeName: this.parseNodeName,
                });
            }
            else if (MatchReport_1.isFailedMatchReport(m)) {
                failedMatches.push(m);
            }
        }
        return failedMatchReport_1.failedMatchReport(this, {
            parseNodeName: this.parseNodeName,
            offset: is.offset,
            children: failedMatches,
            reason: "All alternatives failed",
        });
    }
}
exports.Alt = Alt;
/**
 * Add a condition with a function that verifies that even if we found a match
 * we are happy with it: For example, we like the value it contains.
 * Also capable of vetoing match if the input state is problematic before the potential match
 */
function when(o, matchTest, inputStateTest = () => true) {
    const output = new WhenMatcher(Concat_1.toMatchingLogic(o), matchTest, inputStateTest);
    for (const prop in o) {
        if (o.hasOwnProperty(prop)) {
            output[prop] = o[prop];
        }
    }
    return output;
}
exports.when = when;
class WhenMatcher {
    constructor(inner, matchTest, inputStateTest) {
        this.inner = inner;
        this.matchTest = matchTest;
        this.inputStateTest = inputStateTest;
        this.$id = `When[${inner.$id}]`;
        this.canStartWith = inner.canStartWith;
    }
    matchPrefixReport(is, thisMatchContext, parseContext) {
        if (!this.inputStateTest(is)) {
            return failedMatchReport_1.failedMatchReport(this, {
                parseNodeName: "When",
                offset: is.offset,
                reason: "Input state test returned false",
            });
        }
        const result = this.inner.matchPrefixReport(is, thisMatchContext, parseContext);
        if (!MatchReport_1.isSuccessfulMatchReport(result)) {
            return wrappingMatchReport_1.wrappingFailedMatchReport(this, {
                offset: is.offset,
                inner: result,
            });
        }
        if (!this.matchTest(result.toPatternMatch())) {
            return wrappingMatchReport_1.wrappingFailedMatchReport(this, {
                offset: is.offset,
                inner: result,
                reason: "Match test returned false",
            });
        }
        return terminalMatchReport_1.successfulMatchReport(this, {
            matched: result.matched, offset: result.offset,
            valueRepresented: { value: result.toPatternMatch().$value },
            parseNodeName: "When",
        });
    }
    matchPrefix(a, b, c) { return MatchReport_1.toMatchPrefixResult(this.matchPrefixReport(a, b, c)); }
}
//# sourceMappingURL=Ops.js.map