"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const failedMatchReport_1 = require("./internal/matchReport/failedMatchReport");
const terminalMatchReport_1 = require("./internal/matchReport/terminalMatchReport");
const MatchReport_1 = require("./MatchReport");
/**
 * Match a literal string
 */
class Literal {
    constructor(literal) {
        this.literal = literal;
        this.$id = `Literal[${this.literal}]`;
        this.parseNodeName = "Literal";
    }
    matchPrefix(is) {
        return MatchReport_1.toMatchPrefixResult(this.matchPrefixReport(is));
    }
    matchPrefixReport(is) {
        const peek = is.peek(this.literal.length);
        return (peek === this.literal) ?
            terminalMatchReport_1.successfulMatchReport(this, {
                matched: this.literal,
                parseNodeName: this.parseNodeName,
                offset: is.offset,
            }) :
            failedMatchReport_1.failedMatchReport(this, {
                offset: is.offset,
                matched: commonPortion(this.literal, peek),
                reason: `Did not match literal [${this.literal}]: saw [${peek}]`,
                parseNodeName: this.parseNodeName,
            });
    }
    canStartWith(char) {
        return this.literal[0] === char;
    }
    get requiredPrefix() {
        return this.literal;
    }
}
exports.Literal = Literal;
function commonPortion(str1, str2) {
    let i = 0;
    while (i < str1.length && i < str2.length && str1[i] === str2[i]) {
        i++;
    }
    return str1.slice(0, i);
}
function isLiteral(ml) {
    return ml && ml.literal !== undefined;
}
exports.isLiteral = isLiteral;
const LOOK_AHEAD_SIZE = 100;
/**
 * Support for regex matching. Subclasses can convert the value to
 * whatever type they require.
 */
class AbstractRegex {
    /**
     * Match a regular expression
     * @param regex JavaScript regex to match. Don't use an end anchor.
     * Start anchor will be added if not already there
     * @param lookahead number of characters to pull from the input to try to match.
     * We'll keep grabbing more if a match is found for the whole string
     */
    constructor(regex, lookahead = LOOK_AHEAD_SIZE, parseNodeName) {
        this.lookahead = lookahead;
        this.regex = regex.source.charAt(0) !== "^" ? new RegExp("^" + regex.source) : regex;
        this.parseNodeName = parseNodeName || "Regex";
    }
    get $id() {
        return `Regex: ${this.regex.source}`;
    }
    matchPrefix(is) {
        const output = MatchReport_1.toMatchPrefixResult(this.matchPrefixReport(is));
        return output;
    }
    matchPrefixReport(is) {
        let results;
        let lookAt;
        let charactersToSee = 0;
        function theRegexMatchedSomething() {
            return !!results && !!results[0];
        }
        function matchedEverythingWeLookedAt() {
            return results[0] === lookAt;
        }
        function thereIsMoreToRead() {
            return lookAt.length === charactersToSee;
        }
        // Keep asking for more input if we have matched all of the input in
        // our lookahead buffer
        do {
            charactersToSee += this.lookahead;
            lookAt = is.peek(charactersToSee);
            results = this.regex.exec(lookAt);
        } while (theRegexMatchedSomething() && matchedEverythingWeLookedAt() && thereIsMoreToRead());
        if (theRegexMatchedSomething()) {
            const matched = results[0];
            return terminalMatchReport_1.successfulMatchReport(this, {
                matched,
                offset: is.offset,
                parseNodeName: this.parseNodeName,
                valueRepresented: { value: this.toValue(matched) },
                reason: "Matched RegExp: " + this.regex.toString(),
            });
        }
        else {
            return failedMatchReport_1.failedMatchReport(this, {
                offset: is.offset,
                parseNodeName: this.parseNodeName,
                reason: `Did not match regex /${this.regex.source}/ in [${lookAt}]`,
            });
        }
    }
}
exports.AbstractRegex = AbstractRegex;
/**
 * Match a regular expression.
 */
class Regex extends AbstractRegex {
    /**
     * Create wrapping a native JavaScript regular expression
     * @param regex Regular expression to wrap.
     * Do not use an end anchor. If a start anchor isn't provided it will be added
     */
    constructor(regex) {
        super(regex);
    }
    toValue(s) {
        return s;
    }
}
exports.Regex = Regex;
class MatchInteger extends AbstractRegex {
    constructor() {
        super(/(?:0|[1-9]\d*)/);
    }
    canStartWith(c) {
        return !isNaN(+c);
    }
    toValue(s) {
        return parseInt(s, 10);
    }
}
exports.MatchInteger = MatchInteger;
/**
 * Match an integer. Allows 0  but not leading 0 if more digits
 */
exports.Integer = new MatchInteger();
class MatchFloat extends AbstractRegex {
    constructor() {
        super(/[+-]?\d*[\.]?\d+/);
    }
    toValue(s) {
        return parseFloat(s);
    }
}
exports.MatchFloat = MatchFloat;
/**
 * Match a float.
 */
exports.Float = new MatchFloat();
class MatchLowercaseBoolean extends AbstractRegex {
    constructor() {
        super(/^false|true/);
    }
    toValue(s) {
        return s === "true";
    }
}
exports.MatchLowercaseBoolean = MatchLowercaseBoolean;
/**
 * Match a LowercaseBoolean.
 */
exports.LowercaseBoolean = new MatchLowercaseBoolean();
//# sourceMappingURL=Primitives.js.map