"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Concat_1 = require("./matchers/Concat");
const failedMatchReport_1 = require("./internal/matchReport/failedMatchReport");
const terminalMatchReport_1 = require("./internal/matchReport/terminalMatchReport");
const wrappingMatchReport_1 = require("./internal/matchReport/wrappingMatchReport");
const Whitespace_1 = require("./internal/Whitespace");
const MatchReport_1 = require("./MatchReport");
/**
 * Match zero or more of these
 * @param o matcher
 * @return {Rep1}
 */
function zeroOrMore(o) {
    return new Rep(o);
}
exports.zeroOrMore = zeroOrMore;
/**
 * Match at least one of these
 * @param o matcher
 * @return {Rep1}
 */
function atLeastOne(o) {
    return new Rep1(o);
}
exports.atLeastOne = atLeastOne;
/**
 * Handle repetition, with or without a separator.
 * Prefer subclasses for simplicity and clarity.
 * By default, match zero or more times without a separator
 */
class Repetition {
    /**
     * Generic rep support. Normally use subclasses.
     * @param o matcher
     * @param min mininum number of times the matcher must match for this to be considered a match. Default 0
     * @param sep if this is provided it indicates that this is a rep sep and it is the delimiter
     */
    constructor(o, min = 0, sep) {
        this.min = min;
        this.sep = sep;
        this.$consumeWhiteSpaceBetweenTokens = true;
        this.matcher = Concat_1.toMatchingLogic(o);
        if (sep) {
            this.sepMatcher = Concat_1.toMatchingLogic(sep);
        }
    }
    get $id() {
        return `Rep[${this.matcher.$id}:min=${this.min},sep=[${this.sep}]`;
    }
    consumeWhiteSpace(consumeWhiteSpaceBetweenTokens) {
        this.$consumeWhiteSpaceBetweenTokens = consumeWhiteSpaceBetweenTokens;
        return this;
    }
    canStartWith(char) {
        return (this.min === 0) ?
            true :
            !this.matcher.canStartWith || this.matcher.canStartWith(char);
    }
    get requiredPrefix() {
        return (this.min === 0) ?
            undefined :
            this.matcher.requiredPrefix;
    }
    matchPrefix(is, thisMatchContext, parseContext) {
        return MatchReport_1.toMatchPrefixResult(this.matchPrefixReport(is, thisMatchContext, parseContext));
    }
    matchPrefixReport(is, thisMatchContext, parseContext) {
        let currentInputState = is;
        const successfulMatches = [];
        let failingMatchReport;
        let matched = "";
        const consumeWhitespace = () => {
            const eat = Whitespace_1.readyToMatch(currentInputState, this.$consumeWhiteSpaceBetweenTokens);
            if (!!eat.skipped) {
                successfulMatches.push({
                    kind: "whitespace",
                    inner: whitespaceChildMatch(eat.skipped, this, currentInputState.offset),
                });
            }
            currentInputState = eat.state;
            matched += eat.skipped;
        };
        while (!currentInputState.exhausted()) {
            consumeWhitespace();
            const report = this.matcher.matchPrefixReport(currentInputState, thisMatchContext, parseContext);
            if (!MatchReport_1.isSuccessfulMatchReport(report)) {
                failingMatchReport = { kind: "rep", inner: report };
                break;
            }
            else {
                if (report.matched === "") {
                    throw new Error(`Matcher with id ${this.matcher.$id} within rep matched the empty string.\n` +
                        `I do not think this grammar means what you think it means`);
                }
                currentInputState = currentInputState.consume(report.matched, `Rep matched [${report.matched}]`);
                successfulMatches.push({ kind: "rep", inner: report });
                matched += report.matched;
            }
            if (this.sepMatcher) {
                consumeWhitespace();
                const sepMatchReport = this.sepMatcher.matchPrefixReport(currentInputState, thisMatchContext, parseContext);
                if (!MatchReport_1.isSuccessfulMatchReport(sepMatchReport)) {
                    failingMatchReport = { kind: "sep", inner: sepMatchReport };
                    break;
                }
                else {
                    // successful
                    successfulMatches.push({ kind: "sep", inner: report });
                    currentInputState = currentInputState.consume(sepMatchReport.matched, `Rep separator`);
                    matched += sepMatchReport.matched;
                }
            }
        }
        const successfulRepetitionCount = successfulMatches.filter(k => k.kind === "rep").length;
        return (successfulRepetitionCount >= this.min) ?
            new SuccessfulRepMatchReport(this, "Rep", successfulMatches, is.offset, matched, failingMatchReport) :
            new FailedRepMatchReport(this, "Rep", successfulMatches, is.offset, matched, `Required ${this.min} matches but only found ${successfulRepetitionCount}`, failingMatchReport);
    }
}
exports.Repetition = Repetition;
function whitespaceChildMatch(skipped, matcher, offset) {
    const matchReport = terminalMatchReport_1.successfulMatchReport(matcher, {
        parseNodeName: "Whitespace",
        matched: skipped,
        offset,
    });
    return matchReport;
}
class SuccessfulRepMatchReport {
    constructor(matcher, parseNodeName, successfulMatchReports, originalOffset, matched, innerFailedMatchReport) {
        this.matcher = matcher;
        this.parseNodeName = parseNodeName;
        this.successfulMatchReports = successfulMatchReports;
        this.matched = matched;
        this.innerFailedMatchReport = innerFailedMatchReport;
        this.successful = true;
        this.kind = "real";
        this.innerMatchReports = () => {
            return wrapMatchReports(this.matcher, this.successfulMatchReports, this.innerFailedMatchReport);
        };
        this.offset = originalOffset;
        this.endingOffset = originalOffset + matched.length;
    }
    toPatternMatch() {
        const value = this.successfulMatchReports.filter(k => k.kind === "rep").map(k => k.inner.toPatternMatch().$value);
        return {
            $value: value,
            $matcherId: this.matcher.$id,
            $matched: this.matched,
            $offset: this.offset,
            matchedStructure: () => {
                return this.toValueStructure();
            },
        };
    }
    toParseTree() {
        return {
            $name: this.parseNodeName,
            $value: this.matched,
            $offset: this.offset,
            $children: this.innerMatchReports().filter(MatchReport_1.isSuccessfulMatchReport).map(r => r.toParseTree()),
        };
    }
    toValueStructure() {
        return this.successfulMatchReports.filter(k => k.kind === "rep").map(k => k.inner.toValueStructure());
    }
    toExplanationTree() {
        return {
            successful: true,
            $name: this.parseNodeName,
            $value: this.matched,
            $offset: this.offset,
            $children: this.innerMatchReports().map(r => r.toExplanationTree()),
        };
    }
}
function wrapMatchReports(matcher, successfulMatchReports, innerFailedMatchReport) {
    const wrapOne = (sm) => {
        if (sm.kind === "whitespace") {
            // these don't need wrapped
            return sm.inner;
        }
        return wrappingMatchReport_1.wrappingMatchReport(matcher, {
            inner: sm.inner,
            parseNodeName: sm.kind === "rep" ? "Repetition" : "Separator",
        });
    };
    const wrappedMatchReports = successfulMatchReports.map(wrapOne);
    if (failedMatchReport_1.failedMatchReport) {
        wrappedMatchReports.push(wrappingMatchReport_1.wrappingFailedMatchReport(matcher, {
            inner: innerFailedMatchReport.inner,
            parseNodeName: innerFailedMatchReport.kind === "rep" ? "Repetition" : "Separator",
            reason: "Looks like this is the end of the repetition",
        }));
    }
    return wrappedMatchReports;
}
class FailedRepMatchReport {
    constructor(matcher, parseNodeName, successfulMatchReports, originalOffset, matched, reason, innerFailedMatchReport) {
        this.matcher = matcher;
        this.parseNodeName = parseNodeName;
        this.successfulMatchReports = successfulMatchReports;
        this.matched = matched;
        this.reason = reason;
        this.innerFailedMatchReport = innerFailedMatchReport;
        this.successful = false;
        this.kind = "real";
        this.innerMatchReports = () => {
            return wrapMatchReports(this.matcher, this.successfulMatchReports, this.innerFailedMatchReport);
        };
        this.offset = originalOffset;
        this.endingOffset = originalOffset + matched.length;
    }
    toExplanationTree() {
        return {
            successful: false,
            reason: this.reason,
            $name: this.parseNodeName,
            $value: this.matched,
            $offset: this.offset,
            $children: this.innerMatchReports().map(r => r.toExplanationTree()),
        };
    }
}
/**
 * Match 0 or more times, without a separator
 */
class Rep extends Repetition {
    constructor(o) {
        super(o, 0);
    }
}
exports.Rep = Rep;
/**
 * Match 1 or more times
 */
class Rep1 extends Repetition {
    constructor(o) {
        super(o, 1);
    }
}
exports.Rep1 = Rep1;
class RepSep extends Repetition {
    constructor(o, sep) {
        super(o, 0, sep);
    }
}
exports.RepSep = RepSep;
class Rep1Sep extends Repetition {
    constructor(o, sep) {
        super(o, 1, sep);
    }
}
exports.Rep1Sep = Rep1Sep;
//# sourceMappingURL=Rep.js.map