"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const MatchReport_1 = require("../MatchReport");
const terminalMatchReport_1 = require("./matchReport/terminalMatchReport");
const wrappingMatchReport_1 = require("./matchReport/wrappingMatchReport");
const Whitespace_1 = require("./Whitespace");
/**
 * Inspired by SNOBOL BREAK: http://www.snobol4.org/docs/burks/tutorial/ch4.htm
 * In SNOBOL.
 * Used internally in Concat. Also exposed to end users in Skip functions.
 * BREAK(S) matches "up to but not including" any character in S.
 * The string matched must always be followed in the subject by a character in S.
 * Unlike SPAN and NOTANY, BREAK will match the null string.
 * This implementation goes beyond SNOBOL, in that it allows the break to consume the end token,
 * and also allows the ability to specify a pattern that must not be found.
 */
class Break {
    /**
     * Consume input until (or until and including) the terminal match
     * @param terminateOn desired terminal match
     * @param consumeTerminal default false. Whether to consume the terminal match. If this is set to true,
     * the value of the match will be the match of the terminal pattern. Otherwise it will the preceding,
     * skipped content.
     * @param badMatcher pattern we don't want to see before the desired determinal match.
     * If we see this pattern before, the match breaks.
     */
    constructor(terminateOn, consumeTerminal = false, badMatcher) {
        this.terminateOn = terminateOn;
        this.consumeTerminal = consumeTerminal;
        this.badMatcher = badMatcher;
        this.parseNodeName = "Break";
    }
    get $id() {
        return `Break[${this.terminateOn.$id}]`;
    }
    canStartWith(char) {
        return (this.consumeTerminal && !this.badMatcher && this.terminateOn.canStartWith) ?
            this.terminateOn.canStartWith(char) :
            true;
    }
    get requiredPrefix() {
        return (this.consumeTerminal && !this.badMatcher) ?
            this.terminateOn.requiredPrefix :
            undefined;
    }
    matchPrefix(is, thisMatchContext, parseContext) {
        return MatchReport_1.toMatchPrefixResult(this.matchPrefixReport(is, thisMatchContext, parseContext));
    }
    matchPrefixReport(is, thisMatchContext, parseContext) {
        let currentIs = is;
        let matched = "";
        const originalOffset = is.offset;
        // Apply optimization if possible where we skip to the terminal if we're consuming it and not avoiding a bad match
        if (this.consumeTerminal && !this.badMatcher) {
            const skipped = Whitespace_1.readyToMatch(currentIs, false, this.terminateOn);
            matched += skipped.skipped;
            currentIs = skipped.state;
        }
        let terminalMatch = this.terminateOn.matchPrefixReport(currentIs, thisMatchContext, parseContext);
        while (!currentIs.exhausted() && !MatchReport_1.isSuccessfulMatchReport(terminalMatch)) { // if it fits, it sits
            // But we can't match the bad match if it's defined
            if (this.badMatcher) {
                // technically every time the bad matcher didn't match could be a child of this one's ExplanationTree.
                // but that would be ridiculous.
                const badMatchReport = this.badMatcher.matchPrefixReport(currentIs, thisMatchContext, parseContext);
                if (MatchReport_1.isSuccessfulMatchReport(badMatchReport)) {
                    return wrappingMatchReport_1.wrappingFailedMatchReport(this, {
                        offset: is.offset,
                        matched,
                        parseNodeName: this.parseNodeName,
                        inner: badMatchReport,
                    });
                }
            }
            matched += currentIs.peek(1);
            currentIs = currentIs.advance();
            if (!currentIs.exhausted()) {
                terminalMatch = this.terminateOn.matchPrefixReport(currentIs, thisMatchContext, parseContext);
            }
        }
        // We have found the terminal if we get here... or the end of the input.
        if (this.consumeTerminal && MatchReport_1.isSuccessfulMatchReport(terminalMatch)) {
            return new BreakWithTerminalMatchReport(this, this.parseNodeName, terminalMatch, is.offset, matched + terminalMatch.matched);
        }
        // todo: if terminalMatch ended this, it needs to be a child
        return terminalMatchReport_1.successfulMatchReport(this, {
            matched,
            offset: is.offset,
            parseNodeName: this.parseNodeName,
        });
    }
}
exports.Break = Break;
class BreakWithTerminalMatchReport extends wrappingMatchReport_1.SuccessfulMatchReportWrapper {
    constructor(matcher, parseNodeName, inner, offsetOverride, matchedOverride) {
        super(matcher, parseNodeName, inner);
        this.offsetOverride = offsetOverride;
        this.matchedOverride = matchedOverride;
    }
    get offset() {
        return this.offsetOverride;
    }
    get matched() {
        return this.matchedOverride;
    }
    toPatternMatch() {
        const terminal = this.inner.toPatternMatch();
        // historically, the pattern match returned by Break violates the invariant of $offset is the start of $matched
        return Object.assign({}, terminal, { $matched: this.matched });
    }
}
function isBreak(thing) {
    return !!thing.terminateOn;
}
exports.isBreak = isBreak;
//# sourceMappingURL=Break.js.map