"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Concat_1 = require("../matchers/Concat");
const Primitives_1 = require("../Primitives");
const MatchReport_1 = require("../MatchReport");
const MicrogrammarParseError_1 = require("../MicrogrammarParseError");
const Break_1 = require("./Break");
const CompleteFromStringOptions_1 = require("./CompleteFromStringOptions");
const ExactMatch_1 = require("./ExactMatch");
const SpecGrammar_1 = require("./SpecGrammar");
/**
 * Convenient function to create a microgrammar from a spec within another grammar
 * @param spec string spec
 * @param components
 * @param options
 * @returns {Concat}
 */
function fromString(spec, components = {}, options = {}) {
    return new MicrogrammarSpecParser().fromString(spec, components, options);
}
exports.fromString = fromString;
/**
 * Parses microgrammars expressed as strings.
 */
class MicrogrammarSpecParser {
    constructor() {
        this.anonFieldCount = 0;
    }
    fromString(spec, components, options) {
        const optionsToUse = CompleteFromStringOptions_1.completeWithDefaults(options);
        const specToUse = this.preprocess(spec, optionsToUse);
        const report = ExactMatch_1.perfectMatch(SpecGrammar_1.specGrammar(optionsToUse), specToUse);
        if (!MatchReport_1.isSuccessfulMatchReport(report)) {
            throw new MicrogrammarParseError_1.MicrogrammarParseError(`Unable to parse microgrammar: ${specToUse}`, report.toExplanationTree());
        }
        const matcherSequence1 = this.definitionSpecsFromMicrogrammarSpec(report.toValueStructure(), 
        // tslint:disable-next-line:no-boolean-literal-compare
        components.$consumeWhiteSpaceBetweenTokens !== false);
        const matcherSequence2 = this.populateSpecifiedElements(components, matcherSequence1);
        const matcherSequence3 = this.inferUnspecifiedElements(matcherSequence2);
        const definitions = this.definitionsFromSpecs(specToUse, matcherSequence3);
        const concat = Concat_1.Concat.of(definitions);
        // Copy config to Concat
        for (const key in components) {
            if (key.charAt(0) === "$") {
                concat[key] = components[key];
            }
        }
        return concat;
    }
    /**
     * Given a spec, replace all the DiscardToken instances with a named,
     * but unbound, matcher spec
     * @param spec spec to preprocess before parsing
     * @param optionsToUse options
     * @returns {string}
     */
    preprocess(spec, optionsToUse) {
        const split = spec.split(optionsToUse.ellipsis);
        let joined = "";
        for (let i = 0; i < split.length; i++) {
            if (i > 0) {
                joined += `${optionsToUse.componentPrefix}{_discard${i}}`;
            }
            joined += split[i];
        }
        return joined;
    }
    definitionSpecsFromMicrogrammarSpec(match, consumeWhiteSpaceBetweenTokens) {
        // flatMap would work better here
        const matcherSequence1 = [];
        match.these.forEach(t => {
            if (t.literal.length > 0) {
                matcherSequence1.push({ anonymous: this.matcherForLiteral(t.literal, consumeWhiteSpaceBetweenTokens) });
            }
            matcherSequence1.push({ reference: t.element.elementName });
        });
        matcherSequence1.push({ anonymous: this.matcherForLiteral(match.trailing, consumeWhiteSpaceBetweenTokens) });
        return matcherSequence1;
    }
    populateSpecifiedElements(elements, definitionSpecs) {
        return definitionSpecs.map(t => {
            if (isReference(t) && elements[t.reference]) {
                return {
                    named: {
                        name: t.reference,
                        matcher: Concat_1.toMatchingLogic(elements[t.reference]),
                    },
                };
            }
            else {
                return t;
            }
        });
    }
    inferUnspecifiedElements(definitionSpecs) {
        return definitionSpecs.map((s, i) => {
            if (isReference(s)) {
                const next = definitionSpecs[i + 1];
                if (!next) {
                    throw new Error("I expect to have some sort of trailing matcher");
                }
                if (isReference(next)) {
                    throw new Error("There are two elements in a row without something more specific in between them");
                }
                const untilTheNextThing = new Break_1.Break(matcherFrom(next));
                return {
                    named: {
                        name: s.reference,
                        matcher: untilTheNextThing,
                    },
                };
            }
            else {
                return s;
            }
        });
    }
    definitionsFromSpecs(id, definitionSpecs) {
        const definitions = { $id: id };
        definitionSpecs.forEach(t => {
            if (isAnonymous(t)) {
                this.addAnonymousToDefinitions(definitions, t.anonymous);
            }
            if (isNamed(t)) {
                definitions[t.named.name] = t.named.matcher;
            }
        });
        return definitions;
    }
    matcherForLiteral(literal, consumeWhiteSpaceBetweenTokens) {
        if (!consumeWhiteSpaceBetweenTokens) {
            return new Primitives_1.Literal(literal);
        }
        const whiteSpaceSeparated = literal.split(/\s/);
        if (whiteSpaceSeparated.length === 1) {
            return new Primitives_1.Literal(literal);
        }
        else {
            const definitions = {};
            whiteSpaceSeparated.forEach(token => {
                if (token.length > 0) {
                    this.addAnonymousToDefinitions(definitions, new Primitives_1.Literal(token));
                }
            });
            return Concat_1.Concat.of(definitions);
        }
    }
    addAnonymousToDefinitions(definitions, matcher) {
        const arbName = `_${this.anonFieldCount++}`;
        definitions[arbName] = matcher;
    }
}
exports.MicrogrammarSpecParser = MicrogrammarSpecParser;
function isAnonymous(thing) {
    return !!thing.anonymous;
}
function isReference(thing) {
    return !!thing.reference;
}
function isNamed(thing) {
    return !!thing.named;
}
function matcherFrom(definitionSpec) {
    if (isAnonymous(definitionSpec)) {
        return definitionSpec.anonymous;
    }
    else {
        return definitionSpec.named.matcher;
    }
}
//# sourceMappingURL=MicrogrammarSpecParser.js.map