"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ChangeSet_1 = require("./ChangeSet");
const PatternMatch_1 = require("../PatternMatch");
/**
 * Handle low level updating using get/set properties
 */
class MicrogrammarUpdates {
    updatableMatch(match, cs) {
        const changeSet = (typeof cs === "string") ? new ChangeSet_1.ChangeSet(cs) : cs;
        const updating = {
            $changeSet: changeSet,
            newContent() {
                return this.$changeSet.updated();
            },
            replaceAll(newValue) {
                this.$changeSet.change(match, newValue);
            },
        };
        this.addMatchesAsProperties(updating, updating.$changeSet, match);
        return updating;
    }
    /**
     * adds each match as a property to the target, with setters that update the changeset
     * @param target
     * @param cs
     * @param match
     */
    addMatchesAsProperties(target, cs, match) {
        if (PatternMatch_1.isTreePatternMatch(match)) {
            const submatches = match.submatches();
            // tslint:disable-next-line:forin
            for (const key in submatches) {
                const submatch = submatches[key];
                let initialValue;
                if (PatternMatch_1.isTreePatternMatch(submatch) && submatch.submatches() === {}) {
                    initialValue = submatch.$matched; // or $value ? they should both be the string value.
                    // this could also be derived from content + offset, which reduces memory consumption
                }
                else if (PatternMatch_1.isTreePatternMatch(submatch)) {
                    initialValue = {};
                    this.addMatchesAsProperties(initialValue, cs, submatch);
                }
                else if (submatch) {
                    initialValue = submatch.$value;
                }
                const privateProperty = "_" + key;
                // https://stackoverflow.com/questions/12827266/get-and-set-in-typescript
                target[privateProperty] = initialValue;
                Object.defineProperty(target, key, {
                    get() {
                        return (target.$invalidated) ?
                            undefined :
                            this[privateProperty];
                    },
                    set(newValue) {
                        if (target.$invalidated) {
                            throw new Error(`Cannot set [${key}] on [${target}]: invalidated by parent change`);
                        }
                        target[privateProperty] = newValue;
                        cs.change(submatch, newValue);
                        if (PatternMatch_1.isTreePatternMatch(submatch) && submatch.submatches() !== {}) {
                            // The caller has set the value of an entire property block.
                            // Invalidate the properties under it
                            for (const prop of Object.getOwnPropertyNames(target)) {
                                if (typeof target[prop] === "object") {
                                    target[prop].$invalidated = true;
                                }
                            }
                        }
                    },
                    enumerable: true,
                    configurable: true,
                });
            }
        }
        else {
            // console.log(`Not a tree pattern match: ${JSON.stringify(match)}`);
        }
    }
}
exports.MicrogrammarUpdates = MicrogrammarUpdates;
//# sourceMappingURL=MicrogrammarUpdates.js.map