"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const PatternMatch_1 = require("../../PatternMatch");
const failedMatchReport_1 = require("./failedMatchReport");
const wrappingMatchReport_1 = require("./wrappingMatchReport");
function successfulTreeMatchReport(matcher, params) {
    const { offset, matched, parseNodeName, reason, computeEffects, extraProperties, children } = params;
    return new TreeMatchReport(matcher, matched, offset, children || [], reason, parseNodeName || matcher.$id, extraProperties || {}, computeEffects || []);
}
exports.successfulTreeMatchReport = successfulTreeMatchReport;
function namedChild(name, matchReport) {
    return {
        name, matchReport, explicit: true,
    };
}
exports.namedChild = namedChild;
function unnamedChild(matchReport) {
    return { explicit: false, matchReport };
}
exports.unnamedChild = unnamedChild;
class TreeMatchReport {
    constructor(matcher, matched, offset, children, reason, parseNodeName, extraProperties, computeEffects) {
        this.matcher = matcher;
        this.matched = matched;
        this.offset = offset;
        this.children = children;
        this.reason = reason;
        this.parseNodeName = parseNodeName;
        this.extraProperties = extraProperties;
        this.computeEffects = computeEffects;
        this.successful = true;
        this.kind = "real";
        this.endingOffset = offset + matched.length;
    }
    toPatternMatch() {
        // plus, all the extra properties go on.
        const vs = this.toValueStructure();
        const output = {
            $matcherId: this.matcher.$id,
            $matched: this.matched,
            $offset: this.offset,
            matchedStructure: () => {
                return vs;
            },
            $value: vs,
            $valueMatches: {},
        };
        const submatches = {};
        for (const child of this.children) {
            if (!child.explicit) {
                continue;
            }
            if (PatternMatch_1.isSpecialMember(child.name)) {
                continue;
            }
            const childPatternMatch = child.matchReport.toPatternMatch();
            submatches[child.name] = childPatternMatch;
            if (PatternMatch_1.isTreePatternMatch(childPatternMatch)) {
                output[child.name] = childPatternMatch;
            }
            else {
                output[child.name] = child.matchReport.toValueStructure();
                output.$valueMatches[child.name] = childPatternMatch;
            }
        }
        output.submatches = () => submatches;
        this.applyComputeEffects(output);
        return output;
    }
    toParseTree() {
        // Parse tree ignores computations. It shows only what was pulled from the file
        const happiness = this.children.map(c => wrapChild(this.matcher, c, { altered: false }).toParseTree());
        return {
            $name: this.parseNodeName,
            $offset: this.offset,
            $value: this.matched || "",
            $children: happiness,
        };
    }
    toValueStructure() {
        const output = {};
        for (const child of this.children) {
            if (!child.explicit) {
                continue;
            }
            if (PatternMatch_1.isSpecialMember(child.name)) {
                continue;
            }
            output[child.name] = child.matchReport.toValueStructure();
        }
        this.applyComputeEffects(output);
        return output;
    }
    toExplanationTree() {
        const happiness = this.children.map(c => wrapChild(this.matcher, c, computationsThatAlter(c, this.computeEffects, this.extraProperties)).toExplanationTree());
        return Object.assign({}, this.toParseTree(), { successful: true, reason: this.reason, $children: happiness });
    }
    applyComputeEffects(target) {
        this.computeEffects.forEach(ce => {
            [...ce.alteredProperties, ...ce.newProperties]
                .filter(p => !PatternMatch_1.isSpecialMember(p))
                .forEach(p => {
                target[p] = this.extraProperties[p];
            });
        });
    }
}
function computationsThatAlter(c, computeEffects, finalValues) {
    if (!c.explicit) {
        return { altered: false };
    }
    const computationSteps = computeEffects
        .filter(ce => ce.alteredProperties.includes(c.name))
        .map(ce => ce.stepName);
    if (computationSteps.length === 0) {
        return { altered: false };
    }
    return {
        altered: true,
        computationSteps,
        finalValue: finalValues[c.name],
    };
}
function failedTreeMatchReport(matcher, params) {
    const { matched, parseNodeName, reason, successes, failureName, failureReport } = params;
    const failureMatchReport = failureReport || failedMatchReport_1.failedMatchReport(matcher, {
        offset: params.failingOffset, parseNodeName: failureName, reason,
    });
    return new FailedTreeMatchReport(matcher, matched, params.originalOffset, successes || [], { name: failureName, matchReport: failureMatchReport }, reason, parseNodeName || matcher.$id, params.extraProperties, params.computeEffects);
}
exports.failedTreeMatchReport = failedTreeMatchReport;
class FailedTreeMatchReport {
    constructor(matcher, matched, offset, successfulChildren, failedChild, reason, parseNodeName, extraProperties, computeEffects) {
        this.matcher = matcher;
        this.matched = matched;
        this.offset = offset;
        this.successfulChildren = successfulChildren;
        this.failedChild = failedChild;
        this.reason = reason;
        this.parseNodeName = parseNodeName;
        this.extraProperties = extraProperties;
        this.computeEffects = computeEffects;
        this.kind = "real";
        this.successful = false;
    }
    toExplanationTree() {
        const happiness = this.successfulChildren
            .map(c => wrapChild(this.matcher, c, computationsThatAlter(c, this.computeEffects, this.extraProperties))
            .toExplanationTree());
        const failure = this.failedChild.matchReport.toExplanationTree();
        return {
            successful: false,
            reason: this.reason,
            $name: this.parseNodeName,
            $offset: this.offset,
            $value: this.matched || "",
            $children: [...happiness, failure],
        };
    }
}
function wrapChild(matcher, child, alteringComputations) {
    if (!child.explicit) {
        return child.matchReport;
    }
    if (!alteringComputations.altered) {
        return wrappingMatchReport_1.wrappingMatchReport(matcher, { parseNodeName: child.name, inner: child.matchReport });
    }
    return new OverridingMatchReport(matcher, child.name, child.matchReport, alteringComputations);
}
class OverridingMatchReport extends wrappingMatchReport_1.SuccessfulMatchReportWrapper {
    constructor(matcher, parseNodeName, original, alteringComputations) {
        super(matcher, parseNodeName, original);
        this.alteringComputations = alteringComputations;
    }
    toExplanationTree() {
        const children = [this.inner.toExplanationTree()];
        let value;
        if (this.alteringComputations.altered) {
            value = this.alteringComputations.finalValue;
            const alterationExplanation = {
                successful: true,
                $name: "Compute",
                $value: this.alteringComputations.finalValue,
                $children: [],
                $offset: 0,
                reason: "Overwritten by function in " + this.alteringComputations.computationSteps.join(", "),
            };
            children.push(alterationExplanation);
        }
        else {
            this.inner.toValueStructure();
        }
        return {
            successful: true,
            $name: this.parseNodeName,
            $offset: this.offset,
            $value: value,
            $children: children,
            reason: "Value altered by computation",
        };
    }
}
//# sourceMappingURL=treeMatchReport.js.map