"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const stringify = require("json-stringify-safe");
const Break_1 = require("../internal/Break");
const terminalMatchReport_1 = require("../internal/matchReport/terminalMatchReport");
const treeMatchReport_1 = require("../internal/matchReport/treeMatchReport");
const Whitespace_1 = require("../internal/Whitespace");
const MatchReport_1 = require("../MatchReport");
const PatternMatch_1 = require("../PatternMatch");
const Primitives_1 = require("../Primitives");
const treeMatchReport_2 = require("./../internal/matchReport/treeMatchReport");
function isMatchVeto(thing) {
    return PatternMatch_1.isSpecialMember(thing.$id);
}
const methodsOnEveryMatchingLogic = ["$id", "matchPrefix", "canStartWith", "requiredPrefix"];
/**
 * Represents a concatenation of multiple matchers. This is the normal
 * way we compose matches, although this class needn't be used explicitly,
 * as Microgrammars use it, via fromDefinitions or by composition involving
 * an object literal which will be converted to a Concat.
 * Users should only create Concats directly in the unusual case where they need
 * to control whitespace handling in a unique way for that particular Concat.
 */
class Concat {
    constructor(definitions) {
        this.definitions = definitions;
        this.$consumeWhiteSpaceBetweenTokens = true;
        this.$skipGaps = false;
        this.$lazy = false;
        this.matchSteps = [];
        this.parseNodeName = "Concat";
    }
    /**
     * Normal way to create a Concat. If a $lazy field
     * is set to true, the Concat will be lazily initialized, and
     * _init() must be called before use.
     * @param definitions
     * @return {Concat}
     */
    static of(definitions) {
        const concat = new Concat(definitions);
        if (definitions.$lazy !== true) {
            concat._init();
        }
        return concat;
    }
    get _initialized() {
        return !!this.firstMatcher;
    }
    /**
     * Evaluate all members to ready this Concat for use.
     * Only call this function after using the lazy static factory method:
     * _init is called automatically in the case of the regular Concat.of
     * function
     */
    _init() {
        if (!this._initialized) {
            for (const stepName in this.definitions) {
                if (methodsOnEveryMatchingLogic.indexOf(stepName) === -1) {
                    const def = this.definitions[stepName];
                    if (def === undefined || def === null) {
                        throw new Error(`Invalid concatenation: Step [${stepName}] is ${def}`);
                    }
                    if (stepName.charAt(0) === "$") {
                        // It's a config property. Copy it over.
                        this[stepName] = def;
                    }
                    else if (typeof def === "function") {
                        // It's a calculation function
                        if (def.length === 0 && stepName !== "_init") {
                            // A no arg function is invalid
                            throw new Error(`No arg function [${stepName}] is invalid as a matching step`);
                        }
                        if (PatternMatch_1.isSpecialMember(stepName)) {
                            this.matchSteps.push({ $id: stepName, veto: def });
                        }
                        else {
                            this.matchSteps.push({ $id: stepName, compute: def });
                        }
                    }
                    else {
                        // It's a normal matcher
                        const m = toMatchingLogic(def);
                        // If we are skipping gaps, skip between productions
                        const named = new NamedMatcher(stepName, this.$skipGaps ? new Break_1.Break(m, true) : m);
                        this.matchSteps.push(named);
                    }
                }
            }
            this.firstMatcher = this.matchSteps.filter(isMatcher)[0];
        }
    }
    get $id() {
        return (this.definitions.$id) ?
            this.definitions.$id :
            "Concat{" + this.matchSteps.map(m => m.$id).join(",") + "}";
    }
    canStartWith(char) {
        return !this.firstMatcher.canStartWith || this.firstMatcher.canStartWith(char);
    }
    get requiredPrefix() {
        return this.firstMatcher.requiredPrefix;
    }
    matchPrefixReport(initialInputState, thisMatchContext, parseContext) {
        const bindingTarget = {};
        const matches = [];
        let currentInputState = initialInputState;
        let matched = "";
        const computeEffects = [];
        for (const step of this.matchSteps) {
            if (isMatcher(step)) {
                const eat = Whitespace_1.readyToMatch(currentInputState, this.$consumeWhiteSpaceBetweenTokens);
                if (!!eat.skipped) {
                    matches.push(whitespaceChildMatch(eat.skipped, this, currentInputState.offset));
                }
                currentInputState = eat.state;
                matched += eat.skipped;
                const report = step.matchPrefixReport(currentInputState, thisMatchContext, parseContext);
                if (MatchReport_1.isSuccessfulMatchReport(report)) {
                    matches.push(treeMatchReport_2.namedChild(step.$id, report));
                    currentInputState = currentInputState.consume(report.matched, `Concat step [${report.matcher.$id}] matched ${report.matched}`);
                    matched += report.matched;
                    bindingTarget[step.$id] = report.toValueStructure();
                }
                else {
                    return treeMatchReport_2.failedTreeMatchReport(this, {
                        originalOffset: initialInputState.offset,
                        parseNodeName: this.parseNodeName,
                        matched,
                        reason: `Failed at step '${step.name}'`,
                        successes: matches,
                        failureName: step.name,
                        failureReport: report,
                        extraProperties: bindingTarget,
                        computeEffects,
                    });
                }
            }
            else {
                // It's a function taking the contexts.
                // See if we should stop matching.
                if (isMatchVeto(step)) {
                    const effects = applyComputation(step.$id, step.veto, bindingTarget, [thisMatchContext, parseContext]);
                    computeEffects.push(effects);
                    // tslint:disable-next-line:no-boolean-literal-compare
                    if (effects.computeResult === false) {
                        return treeMatchReport_2.failedTreeMatchReport(this, {
                            originalOffset: initialInputState.offset,
                            parseNodeName: this.parseNodeName,
                            failingOffset: currentInputState.offset,
                            matched,
                            failureName: step.$id,
                            reason: `Match vetoed by ${step.$id}`,
                            successes: matches,
                            extraProperties: bindingTarget,
                            computeEffects,
                        });
                    }
                }
                else {
                    const effects = applyComputation(step.$id, step.compute, bindingTarget);
                    computeEffects.push(effects);
                }
            }
        }
        return treeMatchReport_1.successfulTreeMatchReport(this, {
            matched,
            parseNodeName: this.parseNodeName,
            offset: initialInputState.offset,
            children: matches,
            extraProperties: bindingTarget,
            computeEffects,
        });
    }
    matchPrefix(initialInputState, thisMatchContext, parseContext) {
        return MatchReport_1.toMatchPrefixResult(this.matchPrefixReport(initialInputState, thisMatchContext, parseContext));
    }
}
exports.Concat = Concat;
function applyComputation(stepName, compute, argument, additionalArgs = []) {
    const beforeProperties = Object.entries(argument).map(([k, v]) => {
        return {
            name: k,
            before: stringify(v),
        };
    });
    const computeResult = compute(argument, ...additionalArgs);
    if (computeResult !== undefined) {
        argument[stepName] = computeResult;
    }
    const newProperties = Object.keys(argument).filter(n => !beforeProperties.find(bp => bp.name === n));
    const alteredProperties = beforeProperties.filter(bp => {
        return bp.before !== stringify(argument[bp.name]);
    }).map(bp => bp.name);
    return {
        stepName,
        computeResult,
        newProperties,
        alteredProperties,
    };
}
function isMatcher(s) {
    return s.matchPrefixReport !== undefined;
}
function whitespaceChildMatch(skipped, matcher, offset) {
    const matchReport = terminalMatchReport_1.successfulMatchReport(matcher, {
        parseNodeName: "Whitespace",
        matched: skipped,
        offset,
    });
    return { explicit: false, matchReport };
}
/**
 * Turns a JSON element such as name: "literal" into a matcher.
 * Return undefined if the object is undefined or null
 * @param o object to attempt to make into a matcher
 * @returns {any}
 */
function toMatchingLogic(o) {
    if (!o) {
        return undefined;
    }
    if (typeof o === "string") {
        return new Primitives_1.Literal(o);
    }
    else if (o.exec) {
        return new Primitives_1.Regex(o);
    }
    else if (o.matchPrefixReport) {
        return o;
    }
    else if (o.findMatches) {
        return o.matcher;
    }
    else {
        return Concat.of(o);
    }
}
exports.toMatchingLogic = toMatchingLogic;
/**
 * Give an existing matcher a name
 */
class NamedMatcher {
    constructor(name, ml) {
        this.name = name;
        this.ml = ml;
        this.$id = this.name;
    }
    matchPrefix(is, thisMatchContext, parseContext) {
        return MatchReport_1.toMatchPrefixResult(this.matchPrefixReport(is, thisMatchContext, parseContext));
    }
    matchPrefixReport(is, thisMatchContext, parseContext) {
        return this.ml.matchPrefixReport(is, thisMatchContext, parseContext);
    }
    canStartWith(char) {
        return !this.ml.canStartWith || this.ml.canStartWith(char);
    }
    get requiredPrefix() {
        return this.ml.requiredPrefix;
    }
}
exports.NamedMatcher = NamedMatcher;
function isNamedMatcher(thing) {
    return (thing.name !== undefined) && thing.ml !== undefined;
}
exports.isNamedMatcher = isNamedMatcher;
//# sourceMappingURL=Concat.js.map