"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const sdm_1 = require("@atomist/sdm");
const ghub_1 = require("../../../util/github/ghub");
const projectVersioner_1 = require("./local/projectVersioner");
/**
 * Create and return an execute goal object that creates a Git tag,
 * suitable for use with the [[Tag]] goal.
 *
 * @param opts Options that determine the tag created
 * @return Success if successful, Failure otherwise
 */
function executeTag(opts = {}) {
    return (goalInvocation) => __awaiter(this, void 0, void 0, function* () {
        const { configuration, goalEvent, credentials, id, context, progressLog } = goalInvocation;
        return configuration.sdm.projectLoader.doWithProject({ credentials, id, context, readOnly: false }, (project) => __awaiter(this, void 0, void 0, function* () {
            try {
                let tag;
                let message;
                if (opts.message) {
                    message = opts.message;
                }
                else if (goalEvent.push.after && goalEvent.push.after.message) {
                    message = goalEvent.push.after.message.split("\n")[0];
                }
                if (opts.name) {
                    tag = opts.name;
                    message = message || `Tag ${opts.name}`;
                }
                else {
                    const version = yield projectVersioner_1.goalInvocationVersion(goalInvocation);
                    if (opts.release) {
                        tag = version.replace(/[-+].*/, "");
                        message = message || `Release ${tag}`;
                    }
                    else {
                        tag = version;
                        message = message || `Prerelease ${tag}`;
                    }
                }
                if (opts.build) {
                    tag += "+" + opts.build;
                }
                yield createGitTag({ project, tag, message, log: progressLog });
                return { code: 0, message: `Created tag '${tag}' for ${goalEvent.repo.owner}/${goalEvent.repo.name}` };
            }
            catch (e) {
                const message = `Failed to create tag for ${goalEvent.repo.owner}/${goalEvent.repo.name}: ${e.message}`;
                automation_client_1.logger.error(message);
                progressLog.write(message);
                return { code: 1, message };
            }
        }));
    });
}
exports.executeTag = executeTag;
/**
 * Create and push a Git tag with optional message.
 *
 * @param opts Options for creating a Git tag.
 */
function createGitTag(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!opts.tag) {
            throw new Error("You must provide a valid Git tag");
        }
        if (!opts.project) {
            throw new Error("You must provide a Git project");
        }
        if (!opts.log) {
            opts.log = new sdm_1.LoggingProgressLog("logger");
        }
        const remote = opts.project.remote || "origin";
        try {
            const spawnOpts = { cwd: opts.project.baseDir, log: opts.log };
            const tagArgs = ["tag", opts.tag];
            if (opts.message) {
                tagArgs.splice(1, 0, "-m", opts.message);
            }
            const tagResult = yield sdm_1.spawnLog("git", tagArgs, spawnOpts);
            if (tagResult.code) {
                throw new Error(`git tag failed: ${tagResult.message}`);
            }
            const pushResult = yield sdm_1.spawnLog("git", ["push", remote, opts.tag], spawnOpts);
            if (pushResult.code) {
                throw new Error(`git push failed: ${pushResult.message}`);
            }
        }
        catch (e) {
            e.message = `Failed to create and push git tag '${opts.tag}': ${e.message}`;
            throw e;
        }
    });
}
exports.createGitTag = createGitTag;
/**
 * Create a GitHub tag using the GitHub API.
 *
 * @param id GitHub remote repository reference
 * @param sha Commit SHA to tag
 * @param message Tag message
 * @param version Name of tag
 * @param credentials GitHub token object
 * @deprecated use createGitTag
 */
function createTagForStatus(id, sha, message, version, credentials) {
    return __awaiter(this, void 0, void 0, function* () {
        const tag = {
            tag: version,
            message,
            object: sha,
            type: "commit",
            tagger: {
                name: "Atomist",
                email: "info@atomist.com",
                date: new Date().toISOString(),
            },
        };
        yield ghub_1.createTag(credentials, id, tag);
        yield ghub_1.createTagReference(credentials, id, tag);
    });
}
exports.createTagForStatus = createTagForStatus;
//# sourceMappingURL=executeTag.js.map