"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const sdm_1 = require("@atomist/sdm");
const _ = require("lodash");
const KubernetesGoalScheduler_1 = require("./KubernetesGoalScheduler");
/**
 * GoalCompletionListener factory that puts completed goal jobs into a ttl cache for later deletion.
 */
class KubernetesJobDeletingGoalCompletionListenerFactory {
    constructor(sdm) {
        this.sdm = sdm;
        this.cache = new Map();
        this.initialize();
    }
    create() {
        return (gi) => __awaiter(this, void 0, void 0, function* () {
            const goalEvent = gi.completedGoal;
            if (goalEvent.state === sdm_1.SdmGoalState.in_process) {
                return;
            }
            if (goalEvent.state === sdm_1.SdmGoalState.canceled && !(yield sdm_1.cancelableGoal(goalEvent, gi.configuration))) {
                return;
            }
            const selector = `atomist.com/goal-set-id=${goalEvent.goalSetId},atomist.com/creator=${KubernetesGoalScheduler_1.sanitizeName(this.sdm.configuration.name)}`;
            let jobs;
            try {
                jobs = yield KubernetesGoalScheduler_1.listJobs(selector);
            }
            catch (e) {
                automation_client_1.logger.warn(`Failed to read k8s jobs: ${KubernetesGoalScheduler_1.prettyPrintError(e)}`);
                return;
            }
            automation_client_1.logger.debug(`Found k8s jobs for goal set '${goalEvent.goalSetId}': '${jobs.map(j => `${j.metadata.namespace}:${j.metadata.name}`).join(", ")}'`);
            const goalJobs = jobs.filter(j => {
                const annotations = j.metadata.annotations;
                if (!!annotations && !!annotations["atomist.com/sdm"]) {
                    const sdmAnnotation = JSON.parse(annotations["atomist.com/sdm"]);
                    return sdmAnnotation.goal.uniqueName === goalEvent.uniqueName;
                }
                return false;
            });
            automation_client_1.logger.debug(`Matching k8s job for goal '${goalEvent.uniqueName}' found: '${goalJobs.map(j => `${j.metadata.namespace}:${j.metadata.name}`).join(", ")}'`);
            const ttl = _.get(this.sdm.configuration, "sdm.k8s.job.ttl", 1000 * 60 * 2);
            for (const goalJob of goalJobs) {
                this.cache.set(goalJob.metadata.uid, {
                    ttl: Date.now() + ttl,
                    name: goalJob.metadata.name,
                    namespace: goalJob.metadata.namespace,
                });
            }
        });
    }
    initialize() {
        setInterval(() => __awaiter(this, void 0, void 0, function* () {
            const now = Date.now();
            for (const uid of this.cache.keys()) {
                const job = this.cache.get(uid);
                if (job.ttl <= now) {
                    automation_client_1.logger.debug(`Deleting k8s job '${job.namespace}:${job.name}'`);
                    yield KubernetesGoalScheduler_1.deleteJob(job);
                    automation_client_1.logger.debug(`Deleting k8s pods for job '${job.namespace}:${job.name}'`);
                    yield KubernetesGoalScheduler_1.deletePods(job);
                    this.cache.delete(uid);
                }
            }
        }), _.get(this.sdm.configuration, "sdm.k8s.job.ttlCheckInterval", 15000));
    }
}
exports.KubernetesJobDeletingGoalCompletionListenerFactory = KubernetesJobDeletingGoalCompletionListenerFactory;
//# sourceMappingURL=KubernetesJobDeletingGoalCompletionListener.js.map