/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import {
    doWithRetry,
    GitCommandGitProject,
    GitHubRepoRef,
    HandlerContext,
    ProjectOperationCredentials,
    Tagger,
    TaggerTags,
} from "@atomist/automation-client";
import { EditorOrReviewerParameters } from "@atomist/automation-client/lib/operations/common/params/BaseEditorOrReviewerParameters";
import { AddressChannels } from "@atomist/sdm";
import { listTopics } from "../../util/github/ghub";
import { GitHubTagRouter } from "../../util/github/gitHubTagRouter";

/**
 * Run a tagger and publish tags for this repo
 * @param {Tagger} tagger
 * @param {GitHubRepoRef} id
 * @param {ProjectOperationCredentials} credentials
 * @param {AddressChannels} addressChannels
 * @param {HandlerContext} ctx
 */
export async function publishTags(tagger: Tagger,
                                  id: GitHubRepoRef,
                                  credentials: ProjectOperationCredentials,
                                  addressChannels: AddressChannels,
                                  ctx: HandlerContext): Promise<any> {
    const p = await GitCommandGitProject.cloned(credentials, id);
    const tags: TaggerTags = await tagger(p, ctx, undefined);
    if (tags.tags.length > 0) {
        // Add existing tags so they're not lost
        tags.tags = tags.tags.concat(await listTopics(credentials, id));

        await addressChannels(`Tagging \`${id.owner}/${id.repo}\` with tags ${format(tags.tags)}`);
        const edp: EditorOrReviewerParameters = {
            targets: {
                owner: id.owner,
                repo: id.repo,
                sha: "master",
                branch: id.branch,
                usesRegex: false,
                credentials,
                repoRef: id,
                test: () => true,
            },
        };
        return doWithRetry(() => GitHubTagRouter(tags, edp, undefined),
            "Publish tags", {
                randomize: true,
                retries: 30,
            });
    }
}

function format(tags: string[]): string {
    return tags.map(t => "`" + t + "`").join(", ");
}
