"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const sdm_1 = require("@atomist/sdm");
const handleGitHookEvent_1 = require("../invocation/git/handleGitHookEvent");
/**
 * Process the given args (probably from process.argv) into a
 * GitHookInvocation
 * Note that this is not intended for use by humans, but only from our installed git hooks.
 * @param {string[]} argv command line args
 * @param teamContextResolver resolver to find team id
 * @return {GitHookInvocation}
 */
function argsToGitHookInvocation(argv, teamContextResolver) {
    return __awaiter(this, void 0, void 0, function* () {
        if (argv.length < 3) {
            automation_client_1.logger.error("Not enough arguments to run Git hook, command line: %j", argv);
            process.exit(1);
        }
        const args = (argv[2] === "git-hook") ? argv.slice(3) : argv.slice(2);
        const event = args[0];
        if (!event) {
            automation_client_1.logger.error("Not enough arguments to run Git hook, provided arguments: %j", args);
            process.exit(1);
        }
        const argBranch = args[2];
        const argSha = args[3];
        // can be invoked in the .git/hooks, .git, or project directory
        const baseDir = cleanBaseDir(process.cwd());
        const workspaceId = teamContextResolver.workspaceContext.workspaceId;
        if (argBranch && argSha) {
            const branch = cleanBranch(argBranch);
            const sha = argSha;
            return { event, baseDir, branch, sha, workspaceId };
        }
        else {
            if (event === handleGitHookEvent_1.HookEvent.PostReceive) {
                return new Promise((resolve, reject) => {
                    let input = "";
                    process.stdin.on("data", chunk => {
                        input += chunk.toString();
                    });
                    process.stdin.on("end", () => {
                        const inputWords = input.trim().split(/\s+/);
                        if (inputWords.length < 3) {
                            const msg = `Git post-receive hook did not receive SHA and branch on standard input`;
                            automation_client_1.logger.error(msg);
                            reject(new Error(msg));
                        }
                        // post-receive stdin: before after refname
                        const sha = inputWords[1];
                        const branch = cleanBranch(inputWords[2]);
                        resolve({ event, baseDir, branch, sha, workspaceId });
                    });
                });
            }
            else if (event === handleGitHookEvent_1.HookEvent.PostCommit || event === handleGitHookEvent_1.HookEvent.PostMerge) {
                const gitBranchResult = yield sdm_1.execPromise("git", ["rev-parse", "--abbrev-ref", "HEAD"], { cwd: baseDir });
                const branch = cleanBranch(gitBranchResult.stdout.trim());
                const gitShaResult = yield sdm_1.execPromise("git", ["rev-parse", "HEAD"], { cwd: baseDir });
                const sha = gitShaResult.stdout.trim();
                return { event, baseDir, branch, sha, workspaceId };
            }
            else {
                automation_client_1.logger.error(`Unrecognized Git hook event: ${event}`);
                process.exit(1);
                return Promise.reject(new Error("Will never get here but TypeScript doesn't know that"));
            }
        }
    });
}
exports.argsToGitHookInvocation = argsToGitHookInvocation;
/**
 * Determine project directory from directory which may include .git
 * or .git/hooks.
 *
 * @param dir directory to clean
 * @return base directory with any .git tail and trailing slash removed
 */
function cleanBaseDir(dir) {
    return dir.replace(/(?:[\/\\]\.git(?:[\/\\]hooks)?)?[\/\\]?$/, "");
}
exports.cleanBaseDir = cleanBaseDir;
/**
 * Remove leading refs/heads from Git branch.
 *
 * @param fullBranch complete Git branch specification
 * @return simple branch name
 */
function cleanBranch(fullBranch) {
    return fullBranch.replace(/^refs\/heads\//, "");
}
exports.cleanBranch = cleanBranch;
//# sourceMappingURL=argsToGitHookInvocation.js.map