"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const inquirer_1 = require("inquirer");
const consoleOutput_1 = require("../../ui/consoleOutput");
const bootstrapGenerators_1 = require("./generator/bootstrapGenerators");
const NodeProjectCreationParameters_1 = require("./generator/NodeProjectCreationParameters");
const updatePackageJsonIdentification_1 = require("./generator/updatePackageJsonIdentification");
const embeddedCommandExecution_1 = require("./support/embeddedCommandExecution");
const javaVerification_1 = require("./support/javaVerification");
const addLocalModeTransform_1 = require("./transform/addLocalModeTransform");
/**
 * Add bootstrap commands to generate a new SDM
 * and add local capability to an existing SDM
 *
 * @param {YargBuilder} yargs
 */
function addBootstrapCommands(yargs) {
    addSdmGenerator(yargs);
    addExtensionPackGenerator(yargs);
    addSuperforkGenerator(yargs);
    addEnableLocalSupport(yargs);
}
exports.addBootstrapCommands = addBootstrapCommands;
function addExtensionPackGenerator(yargs, deprecated) {
    const name = deprecated ? "extensionPackDeprecated" : "extensionPack";
    embeddedCommandExecution_1.addEmbeddedCommand(yargs, {
        name,
        cliCommand: deprecated ? "create extension pack" : "create extension-pack",
        cliDescription: "Create an Atomist extension pack",
        configurer: () => __awaiter(this, void 0, void 0, function* () {
            return sdm => sdm.addGeneratorCommand({
                name,
                startingPoint: new automation_client_1.GitHubRepoRef("atomist-seeds", "sdm-pack-seed"),
                parameters: Object.assign(Object.assign({}, NodeProjectCreationParameters_1.NodeProjectCreationParametersDefinition), { "targets.repo": {
                        pattern: /sdm-pack-[A-Za-z0-9\-_]+/,
                        description: "Repo name. Must begin 'sdm-pack'",
                        validInput: "Must begin with 'sdm-pack'",
                    } }),
                transform: [
                    updatePackageJsonIdentification_1.UpdatePackageJsonIdentification,
                ],
                tags: "sdm-pack",
            });
        }),
        listeners: [{
                after: () => __awaiter(this, void 0, void 0, function* () {
                    consoleOutput_1.infoMessage("Add your new extension pack to any SDM you please!\n");
                }),
            }],
        deprecated,
    });
}
function addSdmGenerator(yargs) {
    const choices = ["aspect", "blank", "spring"];
    const typeDescription = "Type of SDM to create";
    const name = "newSdm";
    embeddedCommandExecution_1.addEmbeddedCommand(yargs, {
        name,
        cliCommand: "create sdm",
        cliDescription: "Create an SDM",
        parameters: bootstrapGenerators_1.nodeGenerator(name, undefined).parameters,
        build: argv => {
            // Expose type parameter
            argv.option("type", {
                required: false,
                description: typeDescription,
                choices,
            });
            return argv;
        },
        configurer: () => __awaiter(this, void 0, void 0, function* () {
            consoleOutput_1.adviceDoc("docs/newSdm.md");
            // Gather type parameter
            const questions = [{
                    name: "type",
                    message: typeDescription,
                    type: "list",
                    choices,
                    default: "blank",
                    validate: (input) => choices.includes(input) ?
                        true :
                        `Please enter one of following values: ${choices}`,
                }];
            const answers = yield inquirer_1.prompt(questions);
            switch (answers.type) {
                case "aspect":
                    return sdm => {
                        sdm.addGeneratorCommand(bootstrapGenerators_1.nodeGenerator(name, new automation_client_1.GitHubRepoRef("atomist-seeds", "aspect"), [updatePackageJsonIdentification_1.UpdatePackageJsonIdentification], "atomist"));
                    };
                case "spring":
                    return sdm => {
                        sdm.addGeneratorCommand(bootstrapGenerators_1.nodeGenerator(name, new automation_client_1.GitHubRepoRef("atomist-seeds", "demo-sdm"), [updatePackageJsonIdentification_1.UpdatePackageJsonIdentification], "atomist"));
                    };
                case "blank":
                    return sdm => sdm.addGeneratorCommand(bootstrapGenerators_1.nodeGenerator(name, new automation_client_1.GitHubRepoRef("atomist-seeds", "empty-sdm"), [updatePackageJsonIdentification_1.UpdatePackageJsonIdentification], "atomist"));
                default:
                    throw new Error("Unknown SDM type " + answers.type);
            }
        }),
        listeners: [{
                before: () => __awaiter(this, void 0, void 0, function* () {
                    consoleOutput_1.infoMessage("Please follow the prompts to create a new SDM\n\n");
                }),
                after: (hr, _, chm) => __awaiter(this, void 0, void 0, function* () {
                    // TODO tags seem to be getting set wrongly somewhere, or type definition is wrong
                    if (chm.tags.includes("spring")) {
                        yield doAfterSpringSdmCreation();
                    }
                    if (hr.generatedRepositoryUrl) {
                        consoleOutput_1.infoMessage(`To run your new SDM:
cd ${removeUrlScheme(hr.generatedRepositoryUrl)};
atomist start`);
                    }
                }),
            }],
    });
}
function removeUrlScheme(url) {
    return url.replace(/^.*:\/\//, "");
}
function doAfterSpringSdmCreation() {
    return __awaiter(this, void 0, void 0, function* () {
        consoleOutput_1.adviceDoc("docs/springSdm.md");
        // await verifyJDK();
        yield javaVerification_1.verifyMaven();
    });
}
function addSuperforkGenerator(yargs) {
    const name = "superfork";
    embeddedCommandExecution_1.addEmbeddedCommand(yargs, {
        name,
        cliCommand: "superfork",
        cliDescription: "Superfork a repo",
        parameters: bootstrapGenerators_1.superforkGenerator.parameters,
        configurer: () => __awaiter(this, void 0, void 0, function* () { return sdm => sdm.addGeneratorCommand(bootstrapGenerators_1.superforkGenerator); }),
        listeners: [{
                before: () => __awaiter(this, void 0, void 0, function* () {
                    consoleOutput_1.infoMessage("Please follow the prompts to create a new repo based on a GitHub repo\n\n");
                }),
                after: () => __awaiter(this, void 0, void 0, function* () {
                    consoleOutput_1.infoMessage("Superfork complete\n");
                }),
            }],
        deprecated: true,
    });
}
/**
 * Add local support to this project
 * @param {YargBuilder} yargs
 */
function addEnableLocalSupport(yargs) {
    embeddedCommandExecution_1.addEmbeddedCommand(yargs, {
        name: "addLocalMode",
        cliCommand: "enable local",
        cliDescription: "Add local mode support to a repo",
        parameters: bootstrapGenerators_1.superforkGenerator.parameters,
        configurer: () => __awaiter(this, void 0, void 0, function* () { return sdm => sdm.addCodeTransformCommand(addLocalModeTransform_1.AddLocalMode); }),
        listeners: [{
                before: () => __awaiter(this, void 0, void 0, function* () {
                    consoleOutput_1.infoMessage("Will add local mode support to a GitHub repo\n\n");
                }),
                after: () => __awaiter(this, void 0, void 0, function* () {
                    consoleOutput_1.adviceDoc("docs/runLocally.md");
                }),
            }],
        deprecated: true,
    });
}
//# sourceMappingURL=addBootstrapCommands.js.map