"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const microgrammar_1 = require("@atomist/microgrammar");
const defaultLocalSoftwareDeliveryMachineConfiguration_1 = require("../../../sdm/configuration/defaultLocalSoftwareDeliveryMachineConfiguration");
const consoleOutput_1 = require("../../ui/consoleOutput");
const cloneAndAtomize_1 = require("./support/cloneAndAtomize");
/*
 * Args we'll simply pass through without examining
 * git clone [--template=<template_directory>]
      [-l] [-s] [--no-hardlinks] [-q] [-n] [--bare] [--mirror]
      [-o <name>] [-b <name>] [-u <upload-pack>] [--reference <repository>]
      [--dissociate] [--separate-git-dir <git dir>]
      [--depth <depth>] [--[no-]single-branch] [--no-tags]
      [--recurse-submodules[=<pathspec>]] [--[no-]shallow-submodules]
      [--jobs <n>] [--] <repository> [<directory>]
 */
const GitCloneArgs = [
    "b",
    // "bare", -- We don't support this as it's impossible to install git hooks on a bare repo
    "depth",
    "directory",
    "dissocate",
    "jobs",
    "l",
    "mirror",
    "n",
    "no-hardlinks",
    "no-shallow-submodules",
    "no-single-branch",
    "no-tags",
    "o",
    "q",
    "recurse-submodules",
    "reference",
    "repository",
    "s",
    "separate-git-dir",
    "shallow-submodules",
    "single-branch",
    "template",
    "upload-pack",
];
/**
 * Takes the same arguments as Git clone but onboards the repo with Atomist.
 * We suppress certain flags like 'bare' as they don't make sense when we need
 * to install Atomist git hooks.
 * @param {AutomationClientInfo[]} clients
 * @param {YargBuilder} yargs
 * @param workspaceContextResolver
 */
function addCloneCommand(clients, yargs, workspaceContextResolver) {
    yargs.command({
        command: "clone <args>",
        describe: "Like git clone but also onboards the repo with Atomist " +
            `under the Atomist root at ${defaultLocalSoftwareDeliveryMachineConfiguration_1.determineDefaultRepositoryOwnerParentDirectory()}`,
        builder: a => {
            GitCloneArgs.forEach(arg => {
                a.option(arg, {
                    // TODO might be nice to split these into options and flags
                    // but we are letting git clone do the validation
                    boolean: true,
                    required: false,
                });
            });
            return a;
        },
        handler: () => {
            if (process.argv.length < 3) {
                consoleOutput_1.warningMessage("Not enough arguments to git or atomist clone\n");
                return undefined;
            }
            const argsToGitClone = process.argv.slice(3).join(" ");
            return consoleOutput_1.logExceptionsToConsole(() => __awaiter(this, void 0, void 0, function* () {
                yield superclone(clients, argsToGitClone, workspaceContextResolver.workspaceContext);
            }), true);
        },
    });
}
exports.addCloneCommand = addCloneCommand;
function superclone(clients, args, workspaceContext) {
    return __awaiter(this, void 0, void 0, function* () {
        const repositoryOwnerDirectory = defaultLocalSoftwareDeliveryMachineConfiguration_1.determineDefaultRepositoryOwnerParentDirectory();
        const repoInfo = exports.GitRemoteParser.firstMatch(args);
        if (!repoInfo) {
            consoleOutput_1.warningMessage("Cannot parse repo info from %s\n", args);
            return;
        }
        const { owner, repo } = repoInfo;
        return cloneAndAtomize_1.cloneAndAtomize({
            repositoryOwnerDirectory,
            owner,
            repo,
            cloneCommand: `git clone ${args}`,
            clients,
            workspaceContext,
        });
    });
}
/**
 * Parse a command line to git clone, e.g. git clone https://github.com/owner/repo
 * @type {Microgrammar<{base: string; owner: string; repo: string}>}
 */
exports.GitRemoteParser = microgrammar_1.Microgrammar.fromString(
// tslint:disable-next-line:no-invalid-template-strings
"${base}${sep}${owner}/${repo}${dotgit}", {
    base: /(git@|https?:\/\/)[^:\/]+/,
    sep: /[:\/]/,
    owner: /[^\s\.\/]+/,
    repo: /[^\s\.]+/,
    dotgit: microgrammar_1.optional(".git"),
});
//# sourceMappingURL=addCloneCommand.js.map