"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const expandedTreeRepoFinder_1 = require("../../../sdm/binding/project/expandedTreeRepoFinder");
const expandedTreeUtils_1 = require("../../../sdm/binding/project/expandedTreeUtils");
const FileSystemRemoteRepoRef_1 = require("../../../sdm/binding/project/FileSystemRemoteRepoRef");
const addGitHooks_1 = require("../../setup/addGitHooks");
const consoleOutput_1 = require("../../ui/consoleOutput");
/**
 * Command to add git hooks to current directory or all projects
 * @param {yargs.Argv} yargs
 */
function addAddGitHooksCommand(yargs) {
    yargs.command({
        command: "add git hooks",
        describe: `Install git hooks to current project, or if 'base' parameter is supplied, under that base`,
        builder: (args) => {
            return args.option("base", {
                required: false,
                description: "Base for machine to add",
            });
        },
        handler: (argv) => {
            return consoleOutput_1.logExceptionsToConsole(() => installHookOrHooks(argv.base), true);
        },
    });
}
exports.addAddGitHooksCommand = addAddGitHooksCommand;
function addRemoveGitHooksCommand(yargs) {
    yargs.command({
        command: "remove git hooks",
        describe: `Remove git hooks from all projects, or in current directory if not connected to an SDM`,
        builder: args => {
            return args.option("base", {
                required: false,
                description: "Base for machine to remove",
            });
        },
        handler: args => {
            return consoleOutput_1.logExceptionsToConsole(() => {
                return removeHookOrHooks(args.base);
            }, true);
        },
    });
}
exports.addRemoveGitHooksCommand = addRemoveGitHooksCommand;
function installHookOrHooks(repositoryOwnerParentDirectory) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!repositoryOwnerParentDirectory) {
            return addGitHooks_1.addGitHooks(expandedTreeUtils_1.determineCwd());
        }
        // Install hooks in all directories
        return installAllGitHooks(repositoryOwnerParentDirectory);
    });
}
function removeHookOrHooks(repositoryOwnerParentDirectory) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!repositoryOwnerParentDirectory) {
            const cwd = expandedTreeUtils_1.determineCwd();
            return addGitHooks_1.removeGitHooks(cwd);
        }
        return removeAllGitHooks(repositoryOwnerParentDirectory);
    });
}
/**
 * Install git hooks in all git projects under our expanded directory structure
 * @return {Promise<void>}
 */
function installAllGitHooks(repositoryOwnerParentDirectory) {
    return __awaiter(this, void 0, void 0, function* () {
        const repoFinder = expandedTreeRepoFinder_1.expandedTreeRepoFinder({ repositoryOwnerParentDirectory });
        const allRepos = yield repoFinder(undefined);
        for (const rr of allRepos) {
            if (!FileSystemRemoteRepoRef_1.isFileSystemRemoteRepoRef(rr)) {
                throw new Error(`Unexpected return from repo ref resolver: ${JSON.stringify(rr)}`);
            }
            yield addGitHooks_1.addGitHooks(expandedTreeUtils_1.determineCwd());
        }
    });
}
function removeAllGitHooks(repositoryOwnerParentDirectory) {
    return __awaiter(this, void 0, void 0, function* () {
        const repoFinder = expandedTreeRepoFinder_1.expandedTreeRepoFinder({ repositoryOwnerParentDirectory });
        const allRepos = yield repoFinder(undefined);
        for (const rr of allRepos) {
            if (!FileSystemRemoteRepoRef_1.isFileSystemRemoteRepoRef(rr)) {
                throw new Error(`Unexpected return from repo ref resolver: ${JSON.stringify(rr)}`);
            }
            yield addGitHooks_1.removeGitHooks(rr.fileSystemLocation);
        }
    });
}
//# sourceMappingURL=addGitHooksCommands.js.map