"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const GitHubRepoRef_1 = require("@atomist/automation-client/lib/operations/common/GitHubRepoRef");
// tslint:disable-next-line:import-blacklist
const axios_1 = require("axios");
const defaultLocalSoftwareDeliveryMachineConfiguration_1 = require("../../../sdm/configuration/defaultLocalSoftwareDeliveryMachineConfiguration");
const consoleOutput_1 = require("../../ui/consoleOutput");
const cloneAndAtomize_1 = require("./support/cloneAndAtomize");
/**
 * Import an org from github
 * @param {AutomationClientInfo[]} clients
 * @param {YargBuilder} yargs
 * @param workspaceContextResolver
 */
function addGitHubImportCommand(clients, yargs, workspaceContextResolver) {
    yargs.command({
        command: "import github <org> [githubBase]",
        describe: "Clone all accessible repos under a github org " +
            `under the Atomist root at ${defaultLocalSoftwareDeliveryMachineConfiguration_1.determineDefaultRepositoryOwnerParentDirectory()}`,
        handler: argv => {
            const githubBase = argv.githubBase || GitHubRepoRef_1.GitHubDotComBase;
            return consoleOutput_1.logExceptionsToConsole(() => __awaiter(this, void 0, void 0, function* () {
                yield cloneAllInOrg(clients, githubBase, argv.org, workspaceContextResolver.workspaceContext);
            }), true);
        },
    });
}
exports.addGitHubImportCommand = addGitHubImportCommand;
function cloneAllInOrg(clients, githubBase, owner, workspaceContext) {
    return __awaiter(this, void 0, void 0, function* () {
        const repositoryOwnerDirectory = defaultLocalSoftwareDeliveryMachineConfiguration_1.determineDefaultRepositoryOwnerParentDirectory();
        // Find all accessible repos
        const repos = (yield findAllRepos(githubBase, owner))
            .map((repo) => repo.name);
        consoleOutput_1.infoMessage("Found %d repos\n", repos.length);
        return Promise.all(repos.map(repo => cloneAndAtomize_1.cloneAndAtomize({
            repositoryOwnerDirectory,
            owner,
            repo,
            cloneCommand: `git clone ${githubBase.replace("api.", "")}/${owner}/${repo}`,
            clients,
            workspaceContext,
        }).catch(err => {
            consoleOutput_1.warningMessage("Unable to clone %s/%s/ - \n\t%s\n", githubBase, owner, repo, err.messageClient);
        })));
    });
}
function findAllRepos(githubBase, owner) {
    return __awaiter(this, void 0, void 0, function* () {
        // Try user first, then org
        return findAllReposFor(githubBase, owner, true)
            .catch(err => findAllReposFor(githubBase, owner, false));
    });
}
function findAllReposFor(githubBase, owner, user) {
    return __awaiter(this, void 0, void 0, function* () {
        const url = `${githubBase}/${user ? "users" : "orgs"}/${owner}/repos?per_page=1000`;
        consoleOutput_1.infoMessage("Querying for repos at %s\n", url);
        const token = process.env.GITHUB_TOKEN;
        return (yield axios_1.default.get(url, authHeaders(token))).data;
    });
}
function authHeaders(token) {
    return token ? {
        headers: {
            Authorization: `token ${token}`,
        },
    }
        : {};
}
//# sourceMappingURL=addGitHubImportCommand.js.map