"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const sdm_1 = require("@atomist/sdm");
const embeddedMachine_1 = require("../../../embedded/embeddedMachine");
const consoleOutput_1 = require("../../../ui/consoleOutput");
const fetchMetadataFromAutomationClient_1 = require("../../http/fetchMetadataFromAutomationClient");
const runCommandOnColocatedAutomationClient_1 = require("./runCommandOnColocatedAutomationClient");
/**
 * Add a command, asking for repositoryOwnerParentDirectory, and spawn an SDM
 * to execute it. Shut the sdm.machine down afterwards.
 * Once the client connects, it will prompt for parameters required by the command.
 * These parameters can also be passed through using the initial yargs request,
 * being exposed as optional command parameters.
 * @param {yargs.Argv} yargs
 */
function addEmbeddedCommand(yargs, spec) {
    yargs.command({
        command: spec.cliCommand,
        describe: (spec.deprecated ? "DEPRECATED " : "") + spec.cliDescription,
        builder: ra => {
            // Always require the repositoryOwnerParentDirectory, as
            // we cannot create an embedded SDM otherwise
            ra.option("repositoryOwnerParentDirectory", {
                required: false,
                alias: "base",
                description: "Base of the checked out directory tree the new SDM will operate on",
                default: undefined,
            });
            // Expose optional parameters for the command's parameters if there are any
            if (!!spec.parameters) {
                const pl = sdm_1.toParametersListing(spec.parameters);
                for (const param of pl.parameters) {
                    ra.option(param.name, {
                        description: param.description,
                        required: false,
                    });
                }
            }
            if (!!spec.build) {
                spec.build(ra);
            }
            return ra;
        },
        handler: (argv) => __awaiter(this, void 0, void 0, function* () {
            if (spec.deprecated) {
                consoleOutput_1.warningMessage(`The '${spec.cliCommand}' command is DEPRECATED and will be removed in a future release.\n`);
            }
            return consoleOutput_1.logExceptionsToConsole(() => __awaiter(this, void 0, void 0, function* () {
                yield runCommandOnEmbeddedMachine(argv.repositoryOwnerParentDirectory, yield spec.configurer(argv), spec.name, argv, spec.listeners);
                consoleOutput_1.infoMessage("Execution of command %s complete", spec.name);
            }), true);
        }),
    });
}
exports.addEmbeddedCommand = addEmbeddedCommand;
function runCommandOnEmbeddedMachine(repositoryOwnerParentDirectory, configure, name, params, listeners = []) {
    return __awaiter(this, void 0, void 0, function* () {
        const aca = yield embeddedMachine_1.startEmbeddedMachine({
            repositoryOwnerParentDirectory,
            configure,
        });
        const ai = yield fetchMetadataFromAutomationClient_1.fetchMetadataFromAutomationClient(aca.location);
        if (!ai.client) {
            consoleOutput_1.errorMessage("Could not connect to the bootstrap SDM at %s\n", aca.location.baseEndpoint);
            process.exit(1);
        }
        if (!ai.localConfig) {
            consoleOutput_1.infoMessage("The bootstrap SDM is not running in local mode\n");
        }
        const hm = ai.client.commands.find(c => c.name === name);
        if (!hm) {
            consoleOutput_1.errorMessage("No command named '%s'\n", name);
            process.exit(1);
        }
        return runCommandOnColocatedAutomationClient_1.runCommandOnColocatedAutomationClient(aca.location, aca.localConfig, {
            workspaceName: "embedded",
            workspaceId: "T0",
        }, hm, params, listeners);
    });
}
//# sourceMappingURL=embeddedCommandExecution.js.map