"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const microgrammar_1 = require("@atomist/microgrammar");
const consoleOutput_1 = require("../../../ui/consoleOutput");
const verifyCommandResult_1 = require("./verifyCommandResult");
/**
 * Verify that the correct version of the JDK is present
 * @return {Promise<void>}
 */
function verifyJDK() {
    return __awaiter(this, void 0, void 0, function* () {
        yield verifyCommandResult_1.verifyCommandResult({
            command: "java -version",
            outputTest: verifyJavaTest,
            onFailure: () => consoleOutput_1.infoMessage("Please install Java\n"),
            onWrongVersion: () => consoleOutput_1.infoMessage("Java 8 or above is required\n"),
        });
    });
}
exports.verifyJDK = verifyJDK;
/**
 * Verify that the correct version of Maven is present
 * @return {Promise<void>}
 */
function verifyMaven() {
    return __awaiter(this, void 0, void 0, function* () {
        yield verifyCommandResult_1.verifyCommandResult({
            command: "mvn --version",
            outputTest: verifyMavenTest,
            onFailure: () => consoleOutput_1.infoMessage("Please install Maven\n"),
            onWrongVersion: () => consoleOutput_1.infoMessage("Please update your Maven version\n"),
        });
    });
}
exports.verifyMaven = verifyMaven;
/**
 * The Java version string format keeps changing -
 * https://www.journaldev.com/20930/check-java-version
 * @param {CommandResult} r
 * @return {boolean}
 */
function verifyJavaTest(r) {
    if (!r.stdout) {
        return false;
    }
    let parsed = ThreePointJavaVersionGrammar.firstMatch(r.stdout);
    if (!parsed) {
        parsed = OnePointJavaVersionGrammar.firstMatch(r.stdout);
    }
    if (!parsed) {
        return false;
    }
    return parsed.major >= 9 || (parsed.major === 1 && parsed.minor >= 8);
}
exports.verifyJavaTest = verifyJavaTest;
const ThreePointJavaVersionGrammar = microgrammar_1.Microgrammar.fromString(
// tslint:disable-next-line:no-invalid-template-strings
"java version \"${major}.${minor}.${b1}", {
    major: microgrammar_1.Integer,
    minor: microgrammar_1.Integer,
    b1: microgrammar_1.Integer,
});
// Java version output changed in 9, so this is valid
const OnePointJavaVersionGrammar = microgrammar_1.Microgrammar.fromString(
// tslint:disable-next-line:no-invalid-template-strings
"java version \"${major}", {
    major: microgrammar_1.Integer,
});
function verifyMavenTest(r) {
    if (!r.stdout) {
        return false;
    }
    const parsed = MavenVersionGrammar.firstMatch(r.stdout);
    return !!parsed && parsed.major >= 3;
}
exports.verifyMavenTest = verifyMavenTest;
const MavenVersionGrammar = microgrammar_1.Microgrammar.fromString(
// tslint:disable-next-line:no-invalid-template-strings
"Apache Maven ${major}.${minor}.${point}", {
    major: microgrammar_1.Integer,
    minor: microgrammar_1.Integer,
    point: microgrammar_1.Integer,
});
//# sourceMappingURL=javaVerification.js.map