"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const chalk_1 = require("chalk");
const inquirer = require("inquirer");
const _ = require("lodash");
const EnvironmentTokenCredentialsResolver_1 = require("../../../../sdm/binding/EnvironmentTokenCredentialsResolver");
const CommandLineMappedParameterResolver_1 = require("../../../../sdm/binding/mapped-parameter/CommandLineMappedParameterResolver");
const FromAnyMappedParameterResolver_1 = require("../../../../sdm/binding/mapped-parameter/FromAnyMappedParameterResolver");
const ExpandedTreeMappedParameterResolver_1 = require("../../../../sdm/binding/project/ExpandedTreeMappedParameterResolver");
const expandedTreeRepoFinder_1 = require("../../../../sdm/binding/project/expandedTreeRepoFinder");
const expandedTreeUtils_1 = require("../../../../sdm/binding/project/expandedTreeUtils");
const defaultLocalSoftwareDeliveryMachineConfiguration_1 = require("../../../../sdm/configuration/defaultLocalSoftwareDeliveryMachineConfiguration");
const HttpMessageListener_1 = require("../../../../sdm/ui/HttpMessageListener");
const consoleOutput_1 = require("../../../ui/consoleOutput");
const invokeCommandHandlerUsingHttp_1 = require("../../http/invokeCommandHandlerUsingHttp");
const newCorrelationId_1 = require("../../http/support/newCorrelationId");
const portAllocation_1 = require("../../http/support/portAllocation");
const suggestStartingAllMessagesListener_1 = require("./suggestStartingAllMessagesListener");
/**
 * All invocations from the CLI to local SDMs go through this function.
 * Validate command line arguments and prompt for missing or invalid arguments.
 * @param command command populated by yargs
 * @return {Promise<any>}
 */
function runCommandOnColocatedAutomationClient(connectionConfig, opts, target, hm, command, listeners) {
    return __awaiter(this, void 0, void 0, function* () {
        for (const l of listeners) {
            if (!!l.before) {
                yield l.before(hm);
            }
        }
        const listener = new HttpMessageListener_1.HttpMessageListener({
            port: yield portAllocation_1.portToListenOnFor(process.pid),
            transient: true,
        }).start();
        const extraArgs = Object.getOwnPropertyNames(command)
            .map(name => ({ name: convertToUsable(name), value: command[name] }))
            .filter(keep => !!keep.value);
        const args = [
            { name: "github://user_token?scopes=repo,user:email,read:user", value: EnvironmentTokenCredentialsResolver_1.credentialsFromEnvironment().token },
        ]
            .concat(extraArgs);
        yield promptForMissingParameters(hm, args);
        const mpr = new FromAnyMappedParameterResolver_1.FromAnyMappedParameterResolver(new ExpandedTreeMappedParameterResolver_1.ExpandedTreeMappedParameterResolver(opts.repositoryOwnerParentDirectory), new CommandLineMappedParameterResolver_1.ExtraParametersMappedParameterResolver(extraArgs));
        const mappedParameters = hm.mapped_parameters.map(mp => ({
            name: mp.name,
            value: mpr.resolve(mp),
        }));
        yield promptForMissingMappedParameters(hm, mappedParameters);
        const correlationId = yield newCorrelationId_1.newCliCorrelationId({
            channel: expandedTreeUtils_1.parseOwnerAndRepo(opts.repositoryOwnerParentDirectory).repo,
            encodeListenerPort: true,
        });
        const invocation = Object.assign(Object.assign({ name: hm.name, 
            // targets.repos is gathered through the mapped parameters query flow, but
            // is actually a normal parameter
            parameters: args.concat(mappedParameters.filter(mp => mp.name === "targets.repos")), mappedParameters: mappedParameters.filter(mp => !!mp.value).filter(mp => mp.name !== "targets.repos") }, target), { correlationId });
        automation_client_1.logger.debug("Sending invocation %j\n", invocation);
        yield Promise.all(listeners
            .filter(l => !!l.onDispatch)
            .map(l => l.onDispatch(hm, invocation)));
        // Use repo channel if we're in a mapped repo channel
        const r = yield invokeCommandHandlerUsingHttp_1.invokeCommandHandlerUsingHttp(connectionConfig)(invocation);
        yield Promise.all(listeners
            .filter(l => !!l.after)
            .map(l => l.after(r, invocation, hm)));
        yield suggestStartingAllMessagesListener_1.suggestStartingAllMessagesListener();
        if (listener.canTerminate) {
            process.exit(0);
        }
        return r;
    });
}
exports.runCommandOnColocatedAutomationClient = runCommandOnColocatedAutomationClient;
function valid(p, value) {
    return !p.pattern || new RegExp(p.pattern).test(value);
}
/**
 * Gather missing parameters from the command line and add them to args
 * @param {CommandHandlerMetadata} hi
 * @param args Args we've already found
 * @return {object}
 */
function promptForMissingParameters(hi, args) {
    return __awaiter(this, void 0, void 0, function* () {
        function findArg(p) {
            return args.find(a => a.name === p.name);
        }
        function mustBeSupplied(p) {
            const arg = findArg(p);
            return p.required && (!arg || arg.value === undefined ||
                !valid(p, arg.value.toString()));
        }
        function validInputDisplay(p) {
            return p.valid_input ? (": " + chalk_1.default.italic(p.valid_input)) : "";
        }
        _.sortBy(hi.parameters, p => p.name)
            .map(p => ({ parameter: p, arg: findArg(p) }))
            .filter(pair => !!pair.arg && !!pair.arg.value)
            .filter(pair => !valid(pair.parameter, pair.arg.value.toString()))
            .forEach(pair => consoleOutput_1.warningMessage("Value of '%s' for '%s' is invalid%s\n", pair.arg.value, pair.parameter.name, validInputDisplay(pair.parameter)));
        const questions = hi.parameters
            .filter(mustBeSupplied)
            .map(p => {
            const nameToUse = convertToDisplayable(p.name);
            return {
                name: nameToUse,
                default: p.default_value,
                message: p.description,
                validate: (value) => {
                    if (valid(p, value)) {
                        return true;
                    }
                    return `Please enter a valid ${nameToUse}${validInputDisplay(p)}`;
                },
            };
        });
        const fromPrompt = yield inquirer.prompt(questions);
        Object.getOwnPropertyNames(fromPrompt)
            .forEach(enteredName => {
            // Replace any existing argument with this name that yargs has
            _.remove(args, arg => arg.name === enteredName);
            args.push({
                name: convertToUsable(enteredName),
                value: fromPrompt[enteredName],
            });
        });
    });
}
/**
 * handled specially for transform targets
 */
const TransformTargetOwnerParameter = "targets.owner";
/**
 * handled specially for transform targets
 */
const TransformTargetRepositoryParameter = "targets.repo";
/**
 * handle specially for generators
 */
const GeneratorTargetOwnerParameter = "target.owner";
/**
 * Pattern to match all repos
 */
const AllReposPattern = "all repos";
/**
 * Pattern for a custom regular expression
 */
const RegexPattern = "custom regex";
/**
 * Gather missing mapped parameters from the command line and add them to args
 * @param {CommandHandlerMetadata} hi
 * @param mappedParameters mapped parameters we've already found
 * @return {object}
 */
function promptForMissingMappedParameters(hi, mappedParameters) {
    return __awaiter(this, void 0, void 0, function* () {
        const allRepos = yield determineAvailableRepos();
        // Don't ask for repo filter if we have an owner
        if (mappedParameters.some(mp => mp.name === TransformTargetOwnerParameter && !!mp.value) &&
            !mappedParameters.some(mp => mp.name === TransformTargetRepositoryParameter && !!mp.value)) {
            addOrReplaceMappedParameter(mappedParameters, TransformTargetRepositoryParameter, AllReposPattern);
        }
        const questions = mappedParameters
            .filter(mp => !mp.value)
            .map(p => {
            const nameToUse = convertToDisplayable(p.name);
            return {
                name: nameToUse,
                message: () => {
                    switch (p.name) {
                        case GeneratorTargetOwnerParameter:
                            return "Organization (or GitHub user) to hold the new repository";
                        case TransformTargetOwnerParameter:
                            return "Org to target";
                        case TransformTargetRepositoryParameter:
                            return "Repos to target";
                        default:
                            return `(mapped parameter) ${nameToUse}`;
                    }
                },
                // Use a dropdown for orgs and repos
                type: [TransformTargetOwnerParameter, TransformTargetRepositoryParameter].includes(p.name) ? "list" : "string",
                // Choices will be undefined unless it's a list type
                choices: (answer) => {
                    switch (p.name) {
                        case TransformTargetOwnerParameter:
                            return _.uniq(allRepos.map(r => r.owner));
                        case TransformTargetRepositoryParameter:
                            // Used the mapped name, which will be bound to the answer
                            const owner = answer[convertToDisplayable(TransformTargetOwnerParameter)];
                            return allRepos.filter(r => r.owner === owner).map(r => r.repo)
                                .concat(new inquirer.Separator(), chalk_1.default.italic(AllReposPattern));
                        default: return undefined;
                    }
                },
                validate: (value) => {
                    // We don't really know how to validate this,
                    // but make the user input something
                    const mpdef = hi.mapped_parameters.find(mp => mp.name === p.name);
                    if (!!value || (mpdef && (!mpdef.required))) {
                        return true;
                    }
                    return `Please enter a valid value`;
                },
            };
        });
        const fromPrompt = yield inquirer.prompt(questions);
        Object.getOwnPropertyNames(fromPrompt)
            .forEach(enteredName => {
            addOrReplaceMappedParameter(mappedParameters, convertToUsable(enteredName), fromPrompt[enteredName]);
        });
        const trReposField = fromPrompt[convertToDisplayable(TransformTargetRepositoryParameter)];
        if (!!trReposField && trReposField.includes(AllReposPattern)) {
            addOrReplaceMappedParameter(mappedParameters, "targets.repo");
            addOrReplaceMappedParameter(mappedParameters, "targets.repos", ".*");
        }
        else if (!!trReposField && trReposField.includes(RegexPattern)) {
            // Ask extra question to clarify regex if we need to
            const filter = yield inquirer.prompt([{
                    name: "regex",
                    message: "Regex to filter repos by name: Considering including anchors",
                }]);
            consoleOutput_1.infoMessage("Using regex pattern /%s/\n", filter.regex);
            addOrReplaceMappedParameter(mappedParameters, "targets.repo");
            addOrReplaceMappedParameter(mappedParameters, "targets.repos", filter.regex);
        }
        automation_client_1.logger.debug("Mapped parameters after prompt are %j", mappedParameters);
    });
}
/**
 * If value is undefined, delete it
 */
function addOrReplaceMappedParameter(mappedParameters, name, value) {
    _.remove(mappedParameters, arg => arg.name === name);
    if (!!value) {
        mappedParameters.push({ name, value });
    }
}
/**
 * Return all available repos
 * @param repositoryOwnerParentDirectory base of expanded tree
 */
function determineAvailableRepos(opts = defaultLocalSoftwareDeliveryMachineConfiguration_1.defaultLocalSoftwareDeliveryMachineConfiguration({}).local) {
    return __awaiter(this, void 0, void 0, function* () {
        return expandedTreeRepoFinder_1.expandedTreeRepoFinder(opts)(undefined);
    });
}
/**
 * Convert this command to a displayable name
 * @param {string} name
 * @return {string}
 */
function convertToDisplayable(name) {
    const display = name.replace(".", "-");
    return display;
}
exports.convertToDisplayable = convertToDisplayable;
function convertToUsable(entered) {
    return entered.replace("-", ".");
}
//# sourceMappingURL=runCommandOnColocatedAutomationClient.js.map