"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const inquirer = require("inquirer");
const _ = require("lodash");
const handleInstruction_1 = require("./handleInstruction");
const interfaces_1 = require("./interfaces");
const positional_1 = require("./positional");
const sentences_1 = require("./sentences");
function descriptionForConflictWarning(ysc) {
    return ysc.conflictResolution.commandDescription;
}
function errorToString(ve) {
    const prefix = ve.contexts.length > 0 ?
        ve.contexts.join(" -> ") + ": " : "";
    return prefix + ve.complaint;
}
/**
 * Can we combine all of these commands into one, as they are, would it be a problem?
 * @param yss
 */
function whyNotCombine(yss) {
    if (yss.length <= 1) {
        return [];
    }
    const reasons = [];
    if (yss.some(positional_1.hasPositionalArguments)) {
        reasons.push({ complaint: "Cannot combine commands with positional arguments", contexts: [] });
    }
    const completeCommands = yss.filter(ys => ys.isRunnable);
    if (completeCommands.length > 1) {
        reasons.push({
            complaint: `There are ${completeCommands.length} complete commands:\n` +
                completeCommands.map(describeConflictingCommand).join("\n"),
            contexts: [],
        });
    }
    return reasons;
}
function describeConflictingCommand(yc) {
    return `    ${yc.commandName}\t${yc.description}`;
}
function thereIsConflict(yss) {
    return whyNotCombine(yss).length > 0;
}
function dropNonessentialCommands(yss) {
    const [essential, nonessential] = _.partition(yss, yc => yc.conflictResolution.kind !== "drop with warnings");
    const nonessentialChildren = nonessential
        .map(yc => yc)
        .filter(yc => yc.nestedCommands && yc.nestedCommands.length > 0)
        .map(wrapChildren);
    const warnings = nonessential.map(ys => `Warning: ${descriptionForConflictWarning(ys)} is not available because it conflicts with another command`);
    return [[...essential, ...nonessentialChildren], warnings];
}
function wrapChildren(yc) {
    return new sentences_1.YargCommandWord({
        commandName: yc.commandName,
        description: yc.description,
        conflictResolution: yc.conflictResolution,
        nestedCommands: yc.nestedCommands,
    });
}
function combine(commandName, yss) {
    const [combineThese1, warnings1] = thereIsConflict(yss) ? dropNonessentialCommands(yss) : [yss, []];
    const [combineThese2, warnings2] = thereIsConflict(combineThese1) ? combineIntoPrompt(combineThese1) : [combineThese1, []];
    if (thereIsConflict(combineThese2)) {
        // still??
        throw new Error("Unresolvable conflict between commands: " + whyNotCombine(combineThese2).map(errorToString).join("\n"));
    }
    const combineThese = combineThese2;
    const warnings = [...warnings1, ...warnings2];
    if (combineThese.length === 0) {
        return contributeOnlyHelpMessages(commandName, warnings);
    }
    if (combineThese.length === 1) {
        // This might be positional
        const output = combineThese[0];
        output.addHelpMessages(warnings);
        return output;
    }
    // ok. If we got here, we now have multiple commands;
    // none are positional;
    // all may have children
    // but at most one is runnable by itself
    const ycws = combineThese;
    const realCommand = ycws.find(ys => ys.isRunnable);
    return new sentences_1.YargCommandWord({
        commandName,
        description: _.uniq(ycws.map(y => y.description)).join("; or, "),
        runnableCommand: realCommand ? realCommand.runnableCommand : undefined,
        nestedCommands: _.flatMap(ycws.map(ys => ys.nestedCommands)),
        conflictResolution: {
            kind: "expected to be unique",
            failEverything: true,
            commandDescription: "This is already a combined command. Don't call build() twice",
        },
        warnings,
    });
}
exports.combine = combine;
function combineIntoPrompt(ycs) {
    const [promptable, rest] = _.partition(ycs, yc => interfaces_1.isPromptForChoice(yc.conflictResolution));
    const [uniqueChoices, duplicatedChoices] = checkChoicesForUniqueness(promptable);
    const warnings = warningsFromDuplicateChoices(duplicatedChoices);
    if (uniqueChoices.length === 0) {
        return [rest, warnings];
    }
    if (uniqueChoices.length === 1) {
        return [[uniqueChoices[0], ...rest], warnings];
    }
    return [[constructPromptCommandFrom(uniqueChoices), ...rest], warnings];
}
function warningsFromDuplicateChoices(duplicatedChoices) {
    return duplicatedChoices.map(yc => `WARNING: Command '${yc.conflictResolution.commandDescription}' not available. Duplicate name: ${yc.commandName} Duplicate choice: ${yc.conflictResolution.uniqueChoice}`);
}
function checkChoicesForUniqueness(promptableCommands /* with conflictResolution of prompt for choice */) {
    return [promptableCommands, []]; // TODO: implement
}
function constructPromptCommandFrom(promptableCommands /* with conflictResolution of prompt for choice */) {
    return new sentences_1.YargCommandWord({
        commandName: promptableCommands[0].commandName,
        description: _.uniq(promptableCommands.map(y => y.description)).join(" or "),
        runnableCommand: combineChoicesIntoRunnableCommand(promptableCommands),
        nestedCommands: _.flatMap(promptableCommands.map(ys => ys.nestedCommands)),
        conflictResolution: interfaces_1.dropWithWarningsInHelp(`Choice: ${_.uniq(promptableCommands.map(c => c.conflictResolution.commandDescription)).join(" or ")}`),
    });
}
function combineChoicesIntoRunnableCommand(promptableCommands) {
    const runnableOnes = promptableCommands.filter(pc => pc.isRunnable).map(pc => pc);
    return {
        description: "Choice: " + _.uniq(runnableOnes.map(r => r.runnableCommand.description)).join(" or "),
        commandLine: runnableOnes[0].runnableCommand.commandLine,
        handleInstructions: promptAndRun(runnableOnes),
        parameters: _.flatMap(runnableOnes, rc => rc.runnableCommand.parameters),
        positional: [],
    };
}
function promptAndRun(runnableOnes) {
    return {
        fn: (args) => __awaiter(this, void 0, void 0, function* () {
            const question = {
                type: "list",
                name: "selection",
                message: "There is more than one way to do that. Choose one:",
                choices: runnableOnes.map(rc => rc.conflictResolution.uniqueChoice),
            };
            const answer = yield inquirer.prompt([question]);
            const winner = runnableOnes.find(r => r.conflictResolution.uniqueChoice === answer.selection);
            return handleInstruction_1.handleFunctionFromInstructions(winner.runnableCommand.handleInstructions)(args);
        }),
    };
}
function contributeOnlyHelpMessages(formerCommandName, ms) {
    return {
        build() {
            return {
                helpMessages: ms,
                descriptions: [],
                commandName: formerCommandName,
                save(v) {
                    return v;
                },
            };
        },
    };
}
//# sourceMappingURL=combining.js.map