"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const combining_1 = require("./combining");
const interfaces_1 = require("./interfaces");
const sentences_1 = require("./sentences");
function freshYargBuilder(opts = {}) {
    return new YargBuilderTopLevel(opts);
}
exports.freshYargBuilder = freshYargBuilder;
function isYargBuilder(ya) {
    return !!ya.build;
}
exports.isYargBuilder = isYargBuilder;
class YargBuilderTopLevel {
    constructor(opts) {
        this.nestedCommands = [];
        this.parameters = [];
        this.epilogsForHelpMessage = opts.epilogForHelpMessage ? [opts.epilogForHelpMessage] : [];
    }
    demandCommand() {
        // no-op. Only here for compatibility with yargs syntax.
        // We can figure out whether to demand a command.
        return this;
    }
    command(params) {
        const newCommands = sentences_1.imitateYargsCommandMethod(params);
        newCommands.forEach(c => this.nestedCommands.push(c));
        return this;
    }
    withParameter(p) {
        this.parameters.push(p);
        return this;
    }
    withSubcommand(c) {
        if (interfaces_1.isYargCommand(c)) {
            this.nestedCommands.push(c);
        }
        else {
            this.command(c);
        }
        return this;
    }
    option(parameterName, opts) {
        this.withParameter(Object.assign({ parameterName }, opts));
        return this;
    }
    build() {
        const self = this;
        const commandsByNames = _.groupBy(this.nestedCommands, nc => nc.commandName);
        const nestedCommandSavers = Object.entries(commandsByNames).map(([k, v]) => combining_1.combine(k, v).build());
        const helpMessages = [
            ..._.flatMap(nestedCommandSavers, nc => nc.helpMessages),
            ...this.epilogsForHelpMessage
        ];
        const result = {
            commandName: "top",
            helpMessages,
            descriptions: [],
            save(y) {
                _.sortBy(nestedCommandSavers, n => n.commandName).forEach(c => c.save(y));
                if (self.nestedCommands && self.nestedCommands.length > 0) {
                    y.demandCommand();
                    y.recommendCommands();
                }
                self.parameters.forEach(p => y.option(p.parameterName, p));
                y.showHelpOnFail(true);
                y.epilog(helpMessages.join("\n"));
                return y;
            },
        };
        result.nested = nestedCommandSavers;
        return result;
    }
}
//# sourceMappingURL=freshYargBuilder.js.map