"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const commandLine_1 = require("./commandLine");
const handleInstruction_1 = require("./handleInstruction");
const interfaces_1 = require("./interfaces");
function yargCommandWithPositionalArguments(params) {
    return new YargSaverPositionalCommand({
        commandLine: commandLine_1.parseCommandLine(params.command),
        description: params.describe,
        handleInstructions: { fn: params.handler },
        parameters: params.parameters || [],
        positional: params.positional,
        conflictResolution: params.conflictResolution || { kind: "expected to be unique", failEverything: true, commandDescription: params.command },
    });
}
exports.yargCommandWithPositionalArguments = yargCommandWithPositionalArguments;
function positionalCommand(conflictResolution, spec) {
    return new YargSaverPositionalCommand(Object.assign(Object.assign({}, spec), { conflictResolution }));
}
exports.positionalCommand = positionalCommand;
// TODO: check in multilevelCommand and call this one if it fits
class YargSaverPositionalCommand {
    constructor(spec) {
        this.parameters = [];
        this.isRunnable = true;
        this.helpMessages = [];
        commandLine_1.verifyOneWord(spec.commandLine);
        this.commandName = spec.commandLine.firstWord;
        this.commandLine = spec.commandLine;
        this.description = spec.description;
        this.handleInstructions = spec.handleInstructions;
        this.positionalArguments = spec.positional || [];
        this.conflictResolution = spec.conflictResolution;
        this.parameters = spec.parameters;
        if (interfaces_1.isPromptForChoice(this.conflictResolution)) {
            throw new Error("Can't prompt for which positional command you want");
        }
    }
    get contributorName() { return this.commandName; }
    withSubcommand() {
        throw new Error("You cannot have both subcommands and positional arguments");
    }
    withParameter(p) {
        this.parameters.push(p);
        return this;
    }
    addHelpMessages(strs) {
        strs.forEach(s => this.helpMessages.push(s));
    }
    option(parameterName, opts) {
        this.withParameter(Object.assign({ parameterName }, opts));
        return this;
    }
    demandCommand() {
        throw new Error("Commands with positional arguments may not have subcommands");
    }
    command() {
        throw new Error("Commands with positional arguments may not have subcommands");
    }
    build() {
        const ypc = this; // mutating this object will screw this up. Conceptually, should copy
        return {
            helpMessages: ypc.helpMessages,
            commandName: ypc.commandName,
            descriptions: [ypc.description],
            save(yarg) {
                yarg.command({
                    command: ypc.commandLine.toString(),
                    describe: ypc.description,
                    handler: handleInstruction_1.handleFunctionFromInstructions(ypc.handleInstructions),
                    builder: y => {
                        ypc.parameters.forEach(p => yarg.option(p.parameterName, p));
                        for (const pa of ypc.positionalArguments) {
                            y.positional(pa.key, pa.opts);
                        }
                        yarg.showHelpOnFail(true);
                        return y;
                    },
                });
                return yarg;
            },
        };
    }
}
function hasPositionalArguments(ys) {
    return !!ys.positionalArguments;
}
exports.hasPositionalArguments = hasPositionalArguments;
//# sourceMappingURL=positional.js.map