"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const combining_1 = require("./combining");
const commandLine_1 = require("./commandLine");
const handleInstruction_1 = require("./handleInstruction");
const interfaces_1 = require("./interfaces");
const positional_1 = require("./positional");
function apologize() {
    return __awaiter(this, void 0, void 0, function* () {
        // tslint:disable-next-line:no-console
        console.log("I'm sorry. I don't know how to do this. Try --help");
    });
}
/**
 * This command might be runnable by itself, and might also have subcommands.
 *
 * It can be constructed whole, or built up yargs-style.
 */
class YargCommandWord {
    constructor(spec) {
        this.runnableCommand = undefined;
        this.nestedCommands = spec.nestedCommands || [];
        if (spec.runnableCommand && handleInstruction_1.doesSomething(spec.runnableCommand.handleInstructions)) {
            this.runnableCommand = spec.runnableCommand;
            commandLine_1.verifyOneWord(this.runnableCommand.commandLine);
            verifyEmpty(this.runnableCommand.positional);
        }
        this.conflictResolution = spec.conflictResolution;
        this.description = spec.description;
        this.commandName = spec.commandName;
        this.warnings = spec.warnings || [];
        if (spec.runnableCommand && spec.runnableCommand.positional.length > 0) {
            throw new Error("Positional arguments are only ok in a PositionalCommand");
        }
    }
    withSubcommand(c) {
        if (interfaces_1.isYargCommand(c)) {
            this.nestedCommands.push(c);
        }
        else {
            this.command(c);
        }
        return this;
    }
    addHelpMessages(strs) {
        strs.forEach(s => this.warnings.push(s));
    }
    option(parameterName, opts) {
        this.withParameter(Object.assign({ parameterName }, opts));
        return this;
    }
    demandCommand() {
        // no-op. Only here for compatibility with yargs syntax.
        // We can figure out whether to demand a command.
        return this;
    }
    command(params) {
        const newCommands = imitateYargsCommandMethod(params);
        newCommands.forEach(c => this.nestedCommands.push(c));
        return this;
    }
    withParameter(p) {
        this.beRunnable();
        this.runnableCommand.parameters.push(p);
        return this;
    }
    beRunnable() {
        if (!this.runnableCommand) {
            this.runnableCommand = {
                handleInstructions: handleInstruction_1.DoNothing,
                parameters: [],
                commandLine: commandLine_1.parseCommandLine(this.commandName),
                description: this.description,
                positional: [],
            };
        }
    }
    /**
     * if true, this command has a handler function
     * if false, there are only subcommands
     */
    get isRunnable() {
        return !!this.runnableCommand;
    }
    build() {
        const self = this;
        const commandsByNames = _.groupBy(this.nestedCommands, nc => nc.commandName);
        const nestedCommandSavers = Object.entries(commandsByNames).map(([k, v]) => combining_1.combine(k, v).build());
        const descendantHelps = _.flatMap(nestedCommandSavers, nc => nc.helpMessages);
        const helpMessages = [...self.warnings, ...descendantHelps];
        const childDescriptions = _.flatMap(nestedCommandSavers, nc => nc.descriptions);
        const myDescription = self.isRunnable ? self.runnableCommand.description : undefined;
        const allDescriptions = childDescriptions.slice();
        if (myDescription) {
            allDescriptions.push(myDescription);
        }
        const result = {
            helpMessages,
            descriptions: allDescriptions,
            commandName: self.commandName,
            save(yarg) {
                yarg.command({
                    command: self.commandName,
                    describe: condenseDescriptions(childDescriptions, myDescription),
                    builder: y => {
                        y.version(false);
                        nestedCommandSavers.forEach(c => c.save(yarg));
                        if (!self.runnableCommand && self.nestedCommands.length > 0) {
                            y.demandCommand();
                            y.recommendCommands();
                        }
                        if (!!self.runnableCommand) {
                            self.runnableCommand.parameters.forEach(p => y.option(p.parameterName, p));
                        }
                        y.showHelpOnFail(true);
                        y.epilog(helpMessages.join("\n"));
                        return y;
                    },
                    handler: self.runnableCommand ?
                        handleInstruction_1.handleFunctionFromInstructions(self.runnableCommand.handleInstructions) : apologize,
                });
                return yarg;
            },
        };
        // the data here is for testing and debugging
        result.nested = nestedCommandSavers;
        return result;
    }
    get handleInstructions() {
        if (!!this.runnableCommand) {
            return this.runnableCommand.handleInstructions;
        }
        else {
            return handleInstruction_1.DoNothing;
        }
    }
}
exports.YargCommandWord = YargCommandWord;
function condenseDescriptions(allChildDescriptions, myDescription) {
    if (allChildDescriptions.length === 0) {
        // it's just me
        return myDescription || "no action";
    }
    const childDescriptions = _.uniq(allChildDescriptions); // if they happen to be duplicates, don't count them separately
    const childrenDescription = childDescriptions.length === 1 ? "... " + childDescriptions[0] :
        "... " + childDescriptions.length + " commands";
    if (myDescription === undefined) {
        return childrenDescription;
    }
    return myDescription + ", or " + childrenDescription;
}
exports.condenseDescriptions = condenseDescriptions;
function imitateYargsCommandMethod(params) {
    const conflictResolution = params.conflictResolution ||
        { kind: "expected to be unique", failEverything: true, commandDescription: params.command };
    return yargsSpecToMySpecs(params).map(spec => multilevelCommand(spec, params.describe, conflictResolution, params.builder));
}
exports.imitateYargsCommandMethod = imitateYargsCommandMethod;
function verifyEmpty(positionalArgs) {
    if (positionalArgs.length > 1) {
        throw new Error("Positional arguments are only ok in a PositionalCommand");
    }
}
function yargsSpecToMySpecs(params) {
    const commandLine = commandLine_1.parseCommandLine(params.command);
    const originalSpec = {
        commandLine,
        description: params.describe,
        handleInstructions: handleInstruction_1.handleInstructionsFromFunction(params.handler),
        parameters: params.parameters || [],
        positional: params.positional || [],
    };
    const aliasSpecs = oneOrMany(params.aliases).map(a => (Object.assign(Object.assign({}, originalSpec), { commandLine: commandLine_1.commandLineAlias(commandLine, a) })));
    return [originalSpec, ...aliasSpecs];
}
function oneOrMany(t) {
    return t === undefined ? [] : (Array.isArray(t) ? t : [t]);
}
/**
 * Recursively create a command with subcommands for all the words
 * @param params
 */
function multilevelCommand(params, description, conflictResolution, configureInner) {
    const { commandLine } = params;
    if (commandLine.words.length === 1) {
        /* This is the real thing */
        const cmd = commandLine.positionalArguments.length > 0 ?
            positional_1.positionalCommand(conflictResolution, params)
            : new YargCommandWord({
                commandName: commandLine.firstWord,
                description,
                conflictResolution,
                runnableCommand: params,
            });
        if (configureInner) {
            configureInner(cmd);
        }
        return cmd;
    }
    else {
        const nextWord = commandLine.firstWord;
        const rest = commandLine_1.dropFirstWord(commandLine);
        /* build the nested commands */
        const inner = multilevelCommand(Object.assign(Object.assign({}, params), { commandLine: rest }), description, conflictResolution, configureInner);
        /* now build the container */
        return new YargCommandWord({
            commandName: nextWord,
            description: "...",
            conflictResolution,
            nestedCommands: [inner],
        });
    }
}
//# sourceMappingURL=sentences.js.map