"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const defaultWorkspaceContextResolver_1 = require("../../../common/binding/defaultWorkspaceContextResolver");
const FileSystemProjectLoader_1 = require("../../../sdm/binding/project/FileSystemProjectLoader");
const argsToGitHookInvocation_1 = require("../../entry/argsToGitHookInvocation");
const consoleOutput_1 = require("../../ui/consoleOutput");
const renderClientInfo_1 = require("../../ui/renderClientInfo");
const suggestStartingAllMessagesListener_1 = require("../command/support/suggestStartingAllMessagesListener");
const defaultAutomationClientFinder_1 = require("../http/support/defaultAutomationClientFinder");
const handleGitHookEvent_1 = require("./handleGitHookEvent");
const loggingConfiguration = {
    console: Object.assign(Object.assign({}, automation_client_1.PlainLogging.console), { level: "warn" }),
    file: {
        enabled: false,
    },
};
const verbose = process.env.ATOMIST_GITHOOK_VERBOSE === "true";
/**
 * Main entry point for processing git hooks
 * Usage command <git hook name> <directory> <branch> <sha>
 */
function runOnGitHook(argv, clientFinder = defaultAutomationClientFinder_1.defaultAutomationClientFinder()) {
    return __awaiter(this, void 0, void 0, function* () {
        automation_client_1.configureLogging(loggingConfiguration);
        const invocation = yield argsToGitHookInvocation_1.argsToGitHookInvocation(argv, defaultWorkspaceContextResolver_1.DefaultWorkspaceContextResolver);
        if (FileSystemProjectLoader_1.isAtomistTemporaryBranch(invocation.branch)) {
            automation_client_1.logger.info("Ignoring Atomist temporary branch in '%j': Atomist will eventually surface these changes to let hook react", invocation);
            return undefined;
        }
        const clients = yield clientFinder.findAutomationClients();
        if (clients.length > 0) {
            yield suggestStartingAllMessagesListener_1.suggestStartingAllMessagesListener();
        }
        if (clients.length === 0 && verbose) {
            consoleOutput_1.infoMessage("No Atomist connected clients found");
            process.exit(0);
        }
        yield Promise.all(clients.map(client => sendTo(client, invocation)));
    });
}
exports.runOnGitHook = runOnGitHook;
/**
 * Send to a single automation client
 * @param {AutomationClientInfo} automationClientInfo
 * @param {GitHookInvocation} invocation
 * @return {Promise<void>}
 */
function sendTo(automationClientInfo, invocation) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!automationClientInfo.localConfig) {
            if (verbose) {
                consoleOutput_1.infoMessage("Not a local machine; not delivering push event.\n");
            }
            // process.exit(0); // This is a lot faster than just returning. I don't want to make your commit slow.
        }
        else {
            automation_client_1.logger.debug("Executing git hook against project %j", invocation);
            if (verbose) {
                consoleOutput_1.infoMessage(renderClientInfo_1.renderEventDispatch(automationClientInfo, invocation));
            }
            return consoleOutput_1.logExceptionsToConsole(() => handleGitHookEvent_1.handleGitHookEvent(automationClientInfo.location, automationClientInfo.localConfig, invocation), true);
        }
    });
}
//# sourceMappingURL=runOnGitHook.js.map