"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const expandedTreeUtils_1 = require("../../../sdm/binding/project/expandedTreeUtils");
const FileSystemRemoteRepoRef_1 = require("../../../sdm/binding/project/FileSystemRemoteRepoRef");
const git_1 = require("../../../sdm/util/git");
const consoleOutput_1 = require("../../ui/consoleOutput");
const renderClientInfo_1 = require("../../ui/renderClientInfo");
const handleGitHookEvent_1 = require("./handleGitHookEvent");
/**
 * Trigger git events to the given depth in the current project repo,
 * as determined by the current working directory.
 * @param clients clients to onDispatch to
 * @param {string} event
 * @param {number} depth
 * @return {Promise<void>}
 */
function triggerGitEvents(clients, event, depth, teamContextResolver) {
    return __awaiter(this, void 0, void 0, function* () {
        const relevantClients = clients.filter(client => !!client.localConfig && expandedTreeUtils_1.withinExpandedTree(client.localConfig.repositoryOwnerParentDirectory));
        yield Promise.all(relevantClients.map(client => triggerGitEventsOn(client, event, depth, teamContextResolver)));
    });
}
exports.triggerGitEvents = triggerGitEvents;
function triggerGitEventsOn(ai, event, depth, teamContextResolver) {
    return __awaiter(this, void 0, void 0, function* () {
        const currentDir = expandedTreeUtils_1.determineCwd();
        const workspaceId = teamContextResolver.workspaceContext.workspaceId;
        if (expandedTreeUtils_1.withinExpandedTree(ai.localConfig.repositoryOwnerParentDirectory, currentDir)) {
            const p = automation_client_1.GitCommandGitProject.fromBaseDir(FileSystemRemoteRepoRef_1.FileSystemRemoteRepoRef.fromDirectory({
                repositoryOwnerParentDirectory: ai.localConfig.repositoryOwnerParentDirectory,
                mergePullRequests: ai.localConfig.mergePullRequests,
                baseDir: currentDir,
            }), currentDir, undefined, () => undefined);
            automation_client_1.logger.debug("Trigger %s within [%s], depth=%d", event, currentDir, depth);
            const { branch } = yield p.gitStatus();
            // Go back on the current branch
            const shas = (yield git_1.shaHistory(p))
                .slice(0, depth)
                .reverse();
            for (const sha of shas) {
                if (depth > 1) {
                    consoleOutput_1.infoMessage("Sha [%s]\n", sha);
                }
                const invocation = { event, baseDir: currentDir, branch, sha, workspaceId };
                automation_client_1.logger.debug("Trigger %j", invocation);
                consoleOutput_1.infoMessage(renderClientInfo_1.renderEventDispatch(ai, invocation));
                yield handleGitHookEvent_1.handleGitHookEvent(ai.location, ai.localConfig, invocation);
            }
        }
        else {
            consoleOutput_1.errorMessage("Working directory %s is not within expanded working tree under %s for machine %s\n", currentDir, ai.localConfig.repositoryOwnerParentDirectory, ai.client.name);
        }
    });
}
//# sourceMappingURL=triggerGitEvents.js.map