"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
// tslint:disable-next-line:import-blacklist
const axios_1 = require("axios");
const _ = require("lodash");
/**
 * Make a post to the SDM
 * @param {AutomationClientConnectionConfig} config
 * @param {string} relativePath
 * @param data
 * @return {AxiosPromise}d
 */
function postToSdm(config, relativePath, data) {
    let url = `${config.baseEndpoint}/${relativePath}`;
    if (relativePath.startsWith("/")) {
        url = `${config.baseEndpoint}${relativePath}`;
    }
    automation_client_1.logger.debug("POST to %s with payload %j", url, data);
    return axios_1.default.post(url, data, automationServerAuthHeaders(config))
        .then(logResponse(url), interpretSdmResponse(config, url));
}
exports.postToSdm = postToSdm;
function logResponse(url) {
    return (resp) => {
        automation_client_1.logger.debug(`Response from %s was %d, data %j`, url, resp.status, resp.data);
        return resp.data;
    };
}
function interpretSdmResponse(config, url) {
    return (err) => {
        automation_client_1.logger.error("Error accessing %s: %s", url, err.message);
        if (_.get(err, "message", "").includes("ECONNREFUSED")) {
            const linkThatDemonstratesWhyTheSdmMightNotBeListening = "https://github.com/atomist/github-sdm/blob/acd5f89cb2c3e96fa47ef85b32b2028ea2e045fb/src/atomist.config.ts#L62";
            automation_client_1.logger.error("The SDM is not running or is not accepting connections.\n" +
                "If it's running, check its environment variables. See: " + linkThatDemonstratesWhyTheSdmMightNotBeListening);
            return {
                code: 1,
                message: "Unable to connect to the SDM at " + config.baseEndpoint,
            };
        }
        if (_.get(err, "response.status") === 401) {
            return {
                code: 1,
                message: `Status 401 trying to contact the SDM. You are connecting as: [ ${config.user}:${config.password} ]`,
            };
        }
        return { code: 1, message: err.message };
    };
}
function automationServerAuthHeaders(config) {
    return {
        headers: {
            "content-type": "application/json",
            "Cache-Control": "no-cache",
        },
        auth: {
            username: config.user,
            password: config.password,
        },
    };
}
//# sourceMappingURL=httpInvoker.js.map