"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const fs = require("fs");
const os = require("os");
const path = require("path");
const handleGitHookEvent_1 = require("../invocation/git/handleGitHookEvent");
const consoleOutput_1 = require("../ui/consoleOutput");
/**
 * Add Git hooks to the given repo
 * @param {string} projectBaseDir
 * @return {Promise<void>}
 */
function addGitHooks(projectBaseDir) {
    return __awaiter(this, void 0, void 0, function* () {
        if (fs.existsSync(path.join(projectBaseDir, ".git"))) {
            const p = yield automation_client_1.NodeFsLocalProject.fromExistingDirectory(undefined, projectBaseDir);
            return addGitHooksToProject(p);
        }
        else {
            consoleOutput_1.infoMessage("Ignoring directory at %s as it is not a git project\n", projectBaseDir);
        }
    });
}
exports.addGitHooks = addGitHooks;
function addGitHooksToProject(p) {
    return __awaiter(this, void 0, void 0, function* () {
        for (const event of Object.values(handleGitHookEvent_1.HookEvent)) {
            const atomistContent = scriptFragments()[event];
            if (!atomistContent) {
                consoleOutput_1.errorMessage("Unable to create git script content for event '%s'", event);
                process.exit(1);
            }
            yield reatomizeScript(p, path.join(".git", "hooks", event), markAsAtomistContent(atomistContent));
            yield p.makeExecutable(path.join(".git", "hooks", event));
        }
    });
}
exports.addGitHooksToProject = addGitHooksToProject;
function removeGitHooks(baseDir) {
    return __awaiter(this, void 0, void 0, function* () {
        if (fs.existsSync(path.join(baseDir, ".git"))) {
            const p = yield automation_client_1.NodeFsLocalProject.fromExistingDirectory({ owner: "doesn't", repo: "matter", url: undefined }, baseDir);
            yield removeGitHooksFromProject(p);
        }
        else {
            consoleOutput_1.infoMessage("Ignoring directory at %s as it is not a git project", baseDir);
        }
    });
}
exports.removeGitHooks = removeGitHooks;
function removeGitHooksFromProject(p) {
    return __awaiter(this, void 0, void 0, function* () {
        for (const hookFile of Object.values(handleGitHookEvent_1.HookEvent)) {
            yield deatomizeScript(p, path.join(".git", "hooks", hookFile));
        }
    });
}
exports.removeGitHooksFromProject = removeGitHooksFromProject;
const atomistContentRegExp = /\n#+\s+Atomist start\s+#+\n[\S\s]*?\n#+\s+Atomist end\s+#+\n/;
/**
 * Update the Atomist script element if found
 */
function reatomizeScript(p, scriptPath, newContent) {
    return __awaiter(this, void 0, void 0, function* () {
        const scriptFile = yield p.getFile(scriptPath);
        if (!scriptFile) {
            yield p.addFile(scriptPath, `#!/bin/sh\n${newContent}`);
            consoleOutput_1.infoMessage("Added git hook %s in project at %s\n", scriptPath, p.baseDir);
        }
        else {
            const content = yield scriptFile.getContent();
            if (atomistContentRegExp.test(content)) {
                const updatedAtomist = content.replace(atomistContentRegExp, newContent);
                yield scriptFile.setContent(updatedAtomist);
                consoleOutput_1.infoMessage("Updated Atomist content in git hook %s in project at %s\n", scriptPath, p.baseDir);
            }
            else {
                yield scriptFile.setContent(content + newContent);
                consoleOutput_1.infoMessage("Added Atomist content to git hook %s in project at %s\n", scriptPath, p.baseDir);
            }
        }
        return p;
    });
}
exports.reatomizeScript = reatomizeScript;
/**
 * Remove Atomist hook from this script
 * @param {LocalProject} p
 * @param {string} scriptPath
 * @return {Promise<void>}
 */
function deatomizeScript(p, scriptPath) {
    return __awaiter(this, void 0, void 0, function* () {
        const script = yield p.getFile(scriptPath);
        if (!script) {
            consoleOutput_1.infoMessage("No git hook %s in project at %s\n", scriptPath, p.baseDir);
        }
        else {
            const content = yield script.getContent();
            const nonAtomist = content.replace(atomistContentRegExp, "");
            if (nonAtomist === content) {
                consoleOutput_1.infoMessage("No Atomist content found in git hook %s in project at %s, ignoring\n", scriptPath, p.baseDir);
            }
            else if (/^(?:#!\/bin\/sh)?\s*$/.test(nonAtomist)) {
                yield p.deleteFile(scriptPath);
                consoleOutput_1.infoMessage("Deleted empty git hook %s in project at %s\n", scriptPath, p.baseDir);
            }
            else {
                yield script.setContent(nonAtomist);
                consoleOutput_1.infoMessage("Removed Atomist content from git hook %s in project at %s\n", scriptPath, p.baseDir);
            }
        }
        return p;
    });
}
exports.deatomizeScript = deatomizeScript;
/**
 * Indexed templates fragments for use in git hooks
 * @return object whose keys are HookEvents and whose values are the shell script snippet for the hook
 */
function scriptFragments() {
    const bg = (os.platform() === "win32") ? "" : " &";
    return {
        [handleGitHookEvent_1.HookEvent.PostReceive]: `
ATOMIST_GITHOOK_VERBOSE=true
export ATOMIST_GITHOOK_VERBOSE
while read before after ref; do
    atomist git-hook ${handleGitHookEvent_1.HookEvent.PostReceive} "$PWD" "$ref" "$after"${bg}
done
`,
        [handleGitHookEvent_1.HookEvent.PostCommit]: `
atomist git-hook ${handleGitHookEvent_1.HookEvent.PostCommit}${bg}
`,
        [handleGitHookEvent_1.HookEvent.PostMerge]: `
atomist git-hook ${handleGitHookEvent_1.HookEvent.PostMerge}${bg}
`,
    };
}
const AtomistStartComment = "\n######## Atomist start ########\n";
const AtomistEndComment = "\n######### Atomist end #########\n";
/**
 * Make it clear this is Atomist content. Makes it easy to remove later.
 * @param {string} toAppend
 * @return {string}
 */
function markAsAtomistContent(toAppend) {
    return `${AtomistStartComment}${toAppend}${AtomistEndComment}`;
}
exports.markAsAtomistContent = markAsAtomistContent;
//# sourceMappingURL=addGitHooks.js.map