"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const boxen = require("boxen");
const chalk_1 = require("chalk");
const sprintf_js_1 = require("sprintf-js");
const docChunk_1 = require("./docChunk");
/**
 * Perform the given action, logging exceptions to the console
 * @param {() => Promise<any>} what
 * @param showStack whether or not to show the stack
 * @return {Promise<void>}
 */
function logExceptionsToConsole(what, showStack) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            yield what();
            // TODO this prevented waiting, but causes feed command to die immediately
            // process.exit(0);
        }
        catch (err) {
            const msg = (showStack ? err.stack : `Error: ${err.message}`) + "\n";
            errorMessage(msg);
            automation_client_1.logger.error(`Error: ${err.message} - ${err.stack}`);
            process.exit(1);
        }
    });
}
exports.logExceptionsToConsole = logExceptionsToConsole;
/**
 * Display an error message to the console, supporting sprintf style
 * @param msg message
 * @param args arguments
 */
function errorMessage(msg, ...args) {
    process.stdout.write(chalk_1.default.red(sprintf_js_1.sprintf("✘ " + msg, ...args)));
}
exports.errorMessage = errorMessage;
/**
 * Display a warning message to the console, supporting sprintf style
 * @param msg message
 * @param args arguments
 */
function warningMessage(msg, ...args) {
    process.stdout.write(chalk_1.default.yellowBright(sprintf_js_1.sprintf("⚠︎ " + msg, ...args)));
}
exports.warningMessage = warningMessage;
/**
 * Get the user's attention without implying there's a problem,
 * supporting sprintf style
 * @param {string} msg
 * @param args
 */
function adviceMessage(msg, ...args) {
    process.stdout.write(chalk_1.default.cyan(sprintf_js_1.sprintf("⚠︎ " + msg, ...args)));
}
exports.adviceMessage = adviceMessage;
/**
 * Display an info message to the console, supporting sprintf style
 * @param msg message
 * @param args arguments
 */
function infoMessage(msg, ...args) {
    process.stdout.write(chalk_1.default.cyan(sprintf_js_1.sprintf(msg, ...args)));
}
exports.infoMessage = infoMessage;
/**
 * Dynamically build an advice block from one or more documentation chunks
 * @param {string} relativePaths paths relative to the base of the current project
 */
function adviceDoc(...relativePaths) {
    const docChunk = relativePaths.map(docChunk_1.renderProjectDocChunk).join("\n\n");
    if (docChunk) {
        // Some times you want to draw boxes manually (eg emoji use)
        // box characters is at pos 4; before is the terminal markdown
        if (docChunk.charAt(4) === "┌") {
            process.stdout.write("\n" + docChunk + "\n\n");
        }
        else {
            process.stdout.write("\n" + boxen(docChunk, { padding: 1 }) + "\n\n");
        }
    }
    else if (docChunk === "") {
        // this is fine, it's empty
    }
    else {
        warningMessage("Warning: unable to display advice: Document(s) at '%s' not found", relativePaths.join(":"));
    }
}
exports.adviceDoc = adviceDoc;
//# sourceMappingURL=consoleOutput.js.map