"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const consoleOutput_1 = require("../../cli/ui/consoleOutput");
const expandedTreeUtils_1 = require("../../sdm/binding/project/expandedTreeUtils");
const FileSystemProjectLoader_1 = require("../../sdm/binding/project/FileSystemProjectLoader");
const FileSystemRemoteRepoRef_1 = require("../../sdm/binding/project/FileSystemRemoteRepoRef");
const pushFromLastCommit_1 = require("./pushFromLastCommit");
function validateEventOnRepo(payload) {
    if (!payload) {
        // TODO return or throw exception
        consoleOutput_1.errorMessage("Payload must be supplied");
        return false;
    }
    if (!payload.branch || !payload.sha || !payload.baseDir) {
        consoleOutput_1.errorMessage("Invalid git hook invocation payload - branch and sha and baseDir are required: %j", payload);
        return false;
    }
    if (FileSystemProjectLoader_1.isAtomistTemporaryBranch(payload.branch)) {
        consoleOutput_1.errorMessage("Atomist internal branch should not have triggered: %j", payload);
        return false;
    }
    if (isValidSHA1(payload.branch)) {
        consoleOutput_1.errorMessage("Looks like branch has been confused with sha: %j", payload);
        return false;
    }
    if (!isValidSHA1(payload.sha)) {
        consoleOutput_1.errorMessage("Invalid sha: %j", payload);
        return false;
    }
    return true;
}
function isValidSHA1(s) {
    return /^[a-fA-F0-9]{40}$/.test(s);
}
exports.isValidSHA1 = isValidSHA1;
/**
 * Perform push-based event handling on this repo
 */
function handlePushBasedEventOnRepo(workspaceId, sender, opts, payload, eventHandlerName, pushToPayload = p => ({ Push: [p] })) {
    return __awaiter(this, void 0, void 0, function* () {
        // This git hook may be invoked from another git hook. This will cause these values to
        // be incorrect, so we need to delete them to have git work them out again from the directory we're passing via cwd
        // https://stackoverflow.com/questions/3542854/calling-git-pull-from-a-git-post-update-hook
        delete process.env.GIT_DIR;
        delete process.env.GIT_WORK_TREE;
        if (!validateEventOnRepo(payload)) {
            return undefined;
        }
        if (!expandedTreeUtils_1.isWithin(opts.repositoryOwnerParentDirectory, payload.baseDir)) {
            // I would rather send this to the atomist feed
            return consoleOutput_1.infoMessage(`SDM: skipping push because '${payload.baseDir}' is not inside '${opts.repositoryOwnerParentDirectory}'\n`);
        }
        const push = yield createPush(workspaceId, opts, payload);
        return sender({
            name: eventHandlerName,
            payload: pushToPayload(push),
        });
    });
}
exports.handlePushBasedEventOnRepo = handlePushBasedEventOnRepo;
function createPush(workspaceId, opts, payload) {
    return __awaiter(this, void 0, void 0, function* () {
        const { baseDir, branch, sha } = payload;
        return doWithProjectUnderExpandedDirectoryTree(baseDir, branch, sha, opts, (p) => __awaiter(this, void 0, void 0, function* () {
            return pushFromLastCommit_1.pushFromLastCommit(workspaceId, branch, sha, p);
        }));
    });
}
function doWithProjectUnderExpandedDirectoryTree(baseDir, branch, sha, opts, action) {
    return __awaiter(this, void 0, void 0, function* () {
        const p = automation_client_1.GitCommandGitProject.fromBaseDir(FileSystemRemoteRepoRef_1.FileSystemRemoteRepoRef.fromDirectory({
            repositoryOwnerParentDirectory: opts.repositoryOwnerParentDirectory,
            mergePullRequests: opts.mergePullRequests,
            baseDir,
            branch,
            sha,
        }), baseDir, {}, () => undefined);
        return action(p);
    });
}
//# sourceMappingURL=handlePushBasedEventOnRepo.js.map