"use strict";
/*
 * Copyright © 2018 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const git_1 = require("../../sdm/util/git");
function repoFieldsFromProject(workspaceId, id) {
    return {
        owner: id.owner,
        name: id.repo,
        org: {
            owner: id.owner,
            ownerType: "organization",
            provider: {
                providerType: "github_com",
                apiUrl: "just.not.there",
                url: "and.nor.is.this",
                providerId: "foo",
            },
        },
        channels: [
            {
                name: id.repo,
                id: id.repo,
                team: { id: workspaceId },
            },
        ],
        /**
         * Default branch is a GitHub concept, so it will always be master here.
         */
        defaultBranch: "master",
    };
}
exports.repoFieldsFromProject = repoFieldsFromProject;
function authorFromCommit(sha, project) {
    return __awaiter(this, void 0, void 0, function* () {
        const result = yield git_1.retrieveLogDataForSha(sha, project.baseDir, "an");
        return {
            name: result.an,
            login: result.an,
        };
    });
}
/**
 * Make a push to send to the Atomist event endpoint from the last commit to this local git project.
 * @param workspaceId id of current team
 * @param branch of the commit to process
 * @param sha sha of the commit to process
 * @param {GitProject} project
 * @return {OnPushToAnyBranch.Push}
 */
function pushFromLastCommit(workspaceId, branch, sha, project) {
    return __awaiter(this, void 0, void 0, function* () {
        const repo = repoFieldsFromProject(workspaceId, project.id);
        const lastCommit = yield buildCommitFromSha(sha, project);
        const lastShas = yield git_1.shaHistory(project);
        const penultimateCommit = (!!lastShas && lastShas.length >= 2) ?
            yield buildCommitFromSha(lastShas[1], project) :
            undefined;
        return {
            id: Date.now() + "_",
            branch,
            repo,
            commits: [
                lastCommit,
            ],
            after: lastCommit,
            before: penultimateCommit,
            timestamp: lastCommit.timestamp,
        };
    });
}
exports.pushFromLastCommit = pushFromLastCommit;
function committerFromCommit(sha, project) {
    return __awaiter(this, void 0, void 0, function* () {
        const result = yield git_1.retrieveLogDataForSha(sha, project.baseDir, "cn", "cN", "ce");
        return {
            login: result.cn,
            person: {
                chatId: {
                    screenName: result.cN,
                },
                gitHubId: {
                    login: result.cn,
                },
                forename: "",
                surname: result.cn,
                name: result.cn,
                emails: [{
                        address: result.ce,
                    }],
            },
        };
    });
}
/**
 * Build an Atomist commit structure from the given sha
 * @param {string} sha
 * @param {GitProject} project
 * @return {Promise<PushFields.Commits & PushForSdmGoal.Before & PushFields.After>}
 */
function buildCommitFromSha(sha, project) {
    return __awaiter(this, void 0, void 0, function* () {
        return {
            message: yield git_1.commitMessageForSha(sha, project),
            sha,
            committer: yield committerFromCommit(sha, project),
            author: yield authorFromCommit(sha, project),
            timestamp: yield git_1.timestampFromCommit(sha, project),
        };
    });
}
//# sourceMappingURL=pushFromLastCommit.js.map