"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const sdm_1 = require("@atomist/sdm");
const child_process_1 = require("child_process");
const fetchMetadataFromAutomationClient_1 = require("../../../cli/invocation/http/fetchMetadataFromAutomationClient");
const renderClientInfo_1 = require("../../../cli/ui/renderClientInfo");
const FileSystemRemoteRepoRef_1 = require("../../../sdm/binding/project/FileSystemRemoteRepoRef");
const runAndLog_1 = require("../../../sdm/util/runAndLog");
exports.LocalSdmDelivery = new sdm_1.GoalWithFulfillment({ uniqueName: "sdmDelivery", name: "Deliver SDM" });
/**
 * Deliver this SDM
 */
function executeLocalSdmDelivery(options) {
    const deliveryManager = new DeliveryManager();
    return (goalInvocation) => __awaiter(this, void 0, void 0, function* () {
        const { id } = goalInvocation;
        if (!FileSystemRemoteRepoRef_1.isFileSystemRemoteRepoRef(id)) {
            // Should not have been called
            throw new Error("Not a local repo ref: " + JSON.stringify(id));
        }
        try {
            yield goalInvocation.addressChannels(`Beginning SDM delivery for SDM at ${id.fileSystemLocation}`);
            const client = yield deliveryManager.deliver(id.fileSystemLocation, options, goalInvocation);
            if (!!client.client) {
                yield goalInvocation.addressChannels(`SDM updated: ${renderClientInfo_1.renderClientInfo(client)}: pid=${client.pid}`);
            }
            return { code: 0 };
        }
        catch (err) {
            automation_client_1.logger.error(err);
            return { code: 1, message: err.stack };
        }
    });
}
exports.executeLocalSdmDelivery = executeLocalSdmDelivery;
// Patterns that show success
const successPatterns = [
    /Starting Atomist automation client/,
];
/**
 * Hold state to manage one SDM
 */
class DeliveryManager {
    constructor() {
        this.childProcesses = {};
    }
    deliver(baseDir, options, goalInvocation) {
        return __awaiter(this, void 0, void 0, function* () {
            const baseEndpoint = `http://${options.host}:${options.port}`;
            let childProcess = this.childProcesses[baseDir];
            if (!!childProcess) {
                // We started a process for this
                yield killPrevious(baseDir);
                yield goalInvocation.addressChannels(`Terminating process with pid \`${childProcess.pid}\` for SDM at ${baseDir}`);
            }
            else {
                // We didn't start it. If it's connected, leave it be
                const existing = yield fetchMetadataFromAutomationClient_1.fetchMetadataFromAutomationClient({ baseEndpoint });
                if (!!existing.client) {
                    // It's running but we didn't start it
                    yield goalInvocation.addressChannels(`Externally managed SDM already running at ${baseDir}: Doing nothing`);
                    return Object.assign(Object.assign({}, existing), { pid: undefined });
                }
                yield goalInvocation.addressChannels(`No previous process found for SDM at ${baseDir}`);
            }
            childProcess = child_process_1.spawn("atomist", ["start", "--install=false", "--local=true", "--compile=false"], { cwd: baseDir });
            this.childProcesses[baseDir] = childProcess;
            // Record output from child process
            const newLineDelimitedLog = new sdm_1.DelimitedWriteProgressLogDecorator(goalInvocation.progressLog, "\n");
            childProcess.stdout.on("data", what => newLineDelimitedLog.write(what.toString()));
            childProcess.stderr.on("data", what => newLineDelimitedLog.write(what.toString()));
            return new Promise((resolve, reject) => {
                let stdout = "";
                childProcess.stdout.addListener("data", (what) => __awaiter(this, void 0, void 0, function* () {
                    if (!!what) {
                        stdout += what;
                    }
                    if (successPatterns.some(successPattern => successPattern.test(stdout))) {
                        const ccr = {
                            baseEndpoint,
                        };
                        yield fetchMetadataFromAutomationClient_1.fetchMetadataFromAutomationClient(ccr)
                            .then(aca => resolve(Object.assign({ pid: childProcess.pid }, aca)))
                            .catch(reject);
                    }
                }));
                childProcess.addListener("exit", () => {
                    this.childProcesses[baseDir] = undefined;
                    reject(new Error("Error starting managed SDM. We should have found success message pattern by now! Please check logs"));
                });
                childProcess.addListener("error", err => {
                    this.childProcesses[baseDir] = undefined;
                    return reject(err);
                });
            });
        });
    }
}
// This is rather heavy-handed: Kills all processes that have started in this directory
function killPrevious(baseDir) {
    return __awaiter(this, void 0, void 0, function* () {
        const key = `${baseDir}/node_modules/@atomist/automation-client/start.client.js`;
        const cmd = `for pid in $(ps -ef | grep "${key}" | awk '{print $2}'); do kill -9 $pid; done`;
        let result;
        try {
            result = yield runAndLog_1.runAndLog(cmd, {});
        }
        catch (err) {
            automation_client_1.logger.warn("%j", result);
        }
    });
}
//# sourceMappingURL=LocalSdmDelivery.js.map