"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const fs = require("fs");
const _ = require("lodash");
const httpMessaging_1 = require("../../../common/ui/httpMessaging");
const runAndLog_1 = require("../../util/runAndLog");
const expandedTreeUtils_1 = require("./expandedTreeUtils");
const FileSystemRemoteRepoRef_1 = require("./FileSystemRemoteRepoRef");
/**
 * Local project loader backed by expanded directory tree.
 * Modifies push behavior before acting on project,
 * and prefers local directories if found on disk
 */
class FileSystemProjectLoader {
    constructor(delegate, opts, preprocess = changeToPushToAtomistBranch(opts)) {
        this.delegate = delegate;
        this.opts = opts;
        this.preprocess = preprocess;
    }
    doWithProject(params, action) {
        // Use local seed as preference if possible
        const localDir = expandedTreeUtils_1.dirFor(this.opts.repositoryOwnerParentDirectory, params.id.owner, params.id.repo);
        const foundLocally = fs.existsSync(localDir);
        if (this.opts.preferLocalSeeds && !FileSystemRemoteRepoRef_1.isFileSystemRemoteRepoRef(params.id) && foundLocally) {
            params.id = FileSystemRemoteRepoRef_1.FileSystemRemoteRepoRef.implied(this.opts.repositoryOwnerParentDirectory, this.opts.mergePullRequests, params.id.owner, params.id.repo);
        }
        const decoratedAction = (p) => __awaiter(this, void 0, void 0, function* () {
            const p2 = yield this.preprocess(p);
            return action(p2);
        });
        // It's trickier to clone a file:// repo with --depth, and that optimization is less impactful
        // on the local filesystem anyway, so don't try to use it. Clone deeply.
        _.set(params, "cloneOptions.alwaysDeep", true);
        return this.delegate.doWithProject(params, decoratedAction);
    }
}
exports.FileSystemProjectLoader = FileSystemProjectLoader;
const AtomistTemporaryBranch = "atomist-internal";
/**
 * Create a branch that will be used internally only
 * @param {string} branch
 * @return {string}
 */
function atomistTemporaryBranchFor(branch) {
    return `${AtomistTemporaryBranch}/${branch}`;
}
function isAtomistTemporaryBranch(branch) {
    return branch.includes(AtomistTemporaryBranch);
}
exports.isAtomistTemporaryBranch = isAtomistTemporaryBranch;
/**
 * Change the behavior of our project to push to an Atomist branch and merge if it cannot
 * push to the checked out branch.
 */
function changeToPushToAtomistBranch(localConfig) {
    return (p) => __awaiter(this, void 0, void 0, function* () {
        p.push = (opts) => __awaiter(this, void 0, void 0, function* () {
            try {
                // First try to push this branch. If it's the checked out branch
                // we'll get an error
                yield httpMessaging_1.logAndSend(`Trying to push to branch ${p.branch} on ${p.id.owner}:${p.id.repo} in ${p.baseDir}`);
                yield runAndLog_1.runAndLog(`git push --force --set-upstream origin ${p.branch}`, {
                    cwd: p.baseDir,
                });
            }
            catch (err) {
                // If this throws an exception it's because we can't push to the checked out branch.
                // Autofix will attempt to do this.
                // So we create a new branch, push that, and then go to the original directory and merge it.
                const newBranch = atomistTemporaryBranchFor(p.branch);
                yield httpMessaging_1.logAndSend(`Pushing to new local branch ${newBranch} after error`);
                yield p.createBranch(newBranch);
                // We disable verify to stop our git hooks running on this
                yield runAndLog_1.runAndLog(`git push --force --no-verify --set-upstream origin ${p.branch}`, { cwd: p.baseDir });
                if (localConfig.mergeAutofixes) {
                    const originalRepoDir = expandedTreeUtils_1.dirFor(localConfig.repositoryOwnerParentDirectory, p.id.owner, p.id.repo);
                    automation_client_1.logger.info("Trying merge in %s", originalRepoDir);
                    // Automerge it
                    yield runAndLog_1.runAndLog(`git merge ${newBranch}`, { cwd: originalRepoDir });
                }
            }
            return p;
        });
        return p;
    });
}
//# sourceMappingURL=FileSystemProjectLoader.js.map