"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const ActionResult_1 = require("@atomist/automation-client/lib/action/ActionResult");
const AbstractRemoteRepoRef_1 = require("@atomist/automation-client/lib/operations/common/AbstractRemoteRepoRef");
const path = require("path");
const runAndLog_1 = require("../../util/runAndLog");
const expandedTreeUtils_1 = require("./expandedTreeUtils");
/**
 * RemoteRepoRef working against our expanded directory structure.
 * Supports cloning and pushing.
 */
class FileSystemRemoteRepoRef extends AbstractRemoteRepoRef_1.AbstractRemoteRepoRef {
    constructor(opts) {
        super(undefined, "http://not.a.real.remote", "http://not.a.real.apiBase", opts.owner, opts.repo, opts.sha, undefined, opts.branch);
        this.kind = "file";
        this.repositoryOwnerParentDirectory = opts.repositoryOwnerParentDirectory;
        this.mergePullRequests = opts.mergePullRequests;
    }
    static fromDirectory(opts) {
        const { owner, repo } = expandedTreeUtils_1.parseOwnerAndRepo(opts.repositoryOwnerParentDirectory, opts.baseDir);
        if (!owner || !repo) {
            throw new Error(`Cannot parse ${opts.repositoryOwnerParentDirectory}/owner/repo from '${opts.baseDir}'`);
        }
        return new FileSystemRemoteRepoRef(Object.assign(Object.assign({}, opts), { owner, repo }));
    }
    /**
     * Remote RepoRef for the given owner and directory. Not guaranteed
     * to exist.
     * @param {string} repositoryOwnerParentDirectory
     * @param {string} owner
     * @param {string} repo
     * @return {RemoteRepoRef}
     */
    static implied(repositoryOwnerParentDirectory, mergePullRequests, owner, repo) {
        const baseDir = expandedTreeUtils_1.dirFor(repositoryOwnerParentDirectory, owner, repo);
        return this.fromDirectory({ repositoryOwnerParentDirectory, mergePullRequests, baseDir });
    }
    createRemote(creds, description, visibility) {
        throw new Error("FileSystemRemoteRepoRef does not implement createRemote");
    }
    deleteRemote(creds) {
        throw new Error("FileSystemRemoteRepoRef does not implement deleteRemote");
    }
    raisePullRequest(creds, title, body, head, base) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.mergePullRequests) {
                const cwd = path.join(this.repositoryOwnerParentDirectory, this.owner, this.repo);
                // 1. Checkout branch
                yield runAndLog_1.runAndLog(`git checkout ${head}`, { cwd });
                // 2. Rebase from master
                yield runAndLog_1.runAndLog(`git rebase ${base}`, { cwd });
                // 3. Checkout master
                yield runAndLog_1.runAndLog(`git checkout ${base}`, { cwd });
                // 4. Merge branch into master
                yield runAndLog_1.runAndLog(`git merge --ff-only ${head}`, { cwd });
            }
            else {
                automation_client_1.logger.info(`Not merging local branch. Please set 'local.mergePullRequests' to enable merging of PR branches`);
            }
            return ActionResult_1.successOn(this);
        });
    }
    setUserConfig(credentials, project) {
        return __awaiter(this, void 0, void 0, function* () {
            return undefined;
        });
    }
    cloneUrl() {
        return `file://${this.repositoryOwnerParentDirectory}/${this.owner}/${this.repo}`;
    }
    get url() {
        return `file://${this.repositoryOwnerParentDirectory}/${this.owner}/${this.repo}`;
    }
    get fileSystemLocation() {
        return `${this.repositoryOwnerParentDirectory}/${this.owner}/${this.repo}`;
    }
}
exports.FileSystemRemoteRepoRef = FileSystemRemoteRepoRef;
function isFileSystemRemoteRepoRef(rr) {
    const maybe = rr;
    return !!maybe.fileSystemLocation;
}
exports.isFileSystemRemoteRepoRef = isFileSystemRemoteRepoRef;
//# sourceMappingURL=FileSystemRemoteRepoRef.js.map