"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const os = require("os");
const path = require("path");
/**
 * Return the directory in our expanded structure for the given directory
 * @param {string} repositoryOwnerParentDirectory
 * @param {string} owner
 * @param {string} repo
 * @return {string}
 */
function dirFor(repositoryOwnerParentDirectory, owner, repo) {
    return `${repositoryOwnerParentDirectory}/${owner}/${repo}`;
}
exports.dirFor = dirFor;
const sep = (path.sep === "\\") ? "\\\\" : path.sep;
const OwnerAndRepoPattern = new RegExp(["^", "([^", "]+)", "([^", "]+)$"].join(sep));
const OwnerOnlyPattern = new RegExp(["^", "([^", "]+)$"].join(sep));
function trimTrailingSlash(dir) {
    return dir.replace(/[\/\\]$/, "");
}
/**
 * Transform native win32 path to a cygwin path.  All backslashes in
 * the path are replaced forward slashes and any leading drive letter
 * is incorprated into the path as a lowercased first directory of the
 * path.
 *
 * @param dir original path
 * @return cygwinized path
 */
function cygwinizePath(dir) {
    return dir.replace(/\\/g, "/").replace(/^([A-Za-z]):/, (f, m) => `/${m.toLocaleLowerCase()}`);
}
exports.cygwinizePath = cygwinizePath;
/**
 * Determine whether baseDir is under repositoryOwnerparentdirectory.
 * On win32, the comparison is done in a case-insensitive way and also
 * tries to deal with Cygwin-style paths.
 *
 * @param repositoryOwnerParentDirectory parent directory
 * @param baseDir putative child directory
 * @return true if baseDir is under repositoryownerparentdirectory
 */
function isWithin(repositoryOwnerParentDirectory, baseDir) {
    if (os.platform() === "win32") {
        const owner = repositoryOwnerParentDirectory.toLocaleLowerCase();
        const repo = baseDir.toLocaleLowerCase();
        return repo.startsWith(owner) || cygwinizePath(repo).startsWith(cygwinizePath(owner));
    }
    return baseDir.startsWith(repositoryOwnerParentDirectory);
}
exports.isWithin = isWithin;
/**
 * Find the owner and repo from the given directory, returning the empty
 * object if it isn't within our expanded directory structure
 * @param {string} repositoryOwnerParentDirectory
 * @param {string} baseDir directory to test
 * @return {{owner?: string; repo?: string}}
 */
function parseOwnerAndRepo(repositoryOwnerParentDirectory, baseDir = determineCwd()) {
    if (!repositoryOwnerParentDirectory || !isWithin(repositoryOwnerParentDirectory, baseDir)) {
        return {};
    }
    const pathUnder = trimTrailingSlash(baseDir
        .replace(trimTrailingSlash(repositoryOwnerParentDirectory), ""));
    const ownerAndRepoMatch = pathUnder.match(OwnerAndRepoPattern);
    if (!!ownerAndRepoMatch && ownerAndRepoMatch.length >= 2) {
        return { owner: ownerAndRepoMatch[1], repo: ownerAndRepoMatch[2] };
    }
    const ownerMatch = pathUnder.match(OwnerOnlyPattern);
    if (!!ownerMatch) {
        return { owner: ownerMatch[1], repo: undefined };
    }
    return {};
}
exports.parseOwnerAndRepo = parseOwnerAndRepo;
/**
 * Is the given directory within the expanded directory tree?
 * @param repositoryOwnerParentDirectory
 * @param {string} baseDir directory to test
 * @return {boolean}
 */
function withinExpandedTree(repositoryOwnerParentDirectory, baseDir = determineCwd()) {
    const { owner, repo } = parseOwnerAndRepo(repositoryOwnerParentDirectory, baseDir);
    return !!owner && !!repo;
}
exports.withinExpandedTree = withinExpandedTree;
/**
 * Where are we running? Respecting symlinks
 * @return {string}
 */
function determineCwd() {
    // Be sure to respect symlinks
    return process.cwd();
}
exports.determineCwd = determineCwd;
//# sourceMappingURL=expandedTreeUtils.js.map