"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const ActionResult_1 = require("@atomist/automation-client/lib/action/ActionResult");
const fs = require("fs-extra");
const path = require("path");
const addGitHooks_1 = require("../../../cli/setup/addGitHooks");
const handlePushBasedEventOnRepo_1 = require("../../../common/git/handlePushBasedEventOnRepo");
const invokeEventHandlerInProcess_1 = require("../../invocation/invokeEventHandlerInProcess");
const git_1 = require("../../util/git");
const runAndLog_1 = require("../../util/runAndLog");
const repoOnboardingEvents_1 = require("../event/repoOnboardingEvents");
const FileSystemRemoteRepoRef_1 = require("./FileSystemRemoteRepoRef");
const InitialCommitMessage = "Initial commit from Atomist";
/**
 * Persist the project to the given local directory given expanded directory
 * conventions. Perform a git init and other after actions, such as installing
 * our git hooks. Fires off the necessary new repo events using HTTP to
 * all local automation clients.
 * @return {ProjectPersister}
 */
function fileSystemProjectPersister(teamContext, opts, automationClientFinder) {
    return (p, _, id, params) => __awaiter(this, void 0, void 0, function* () {
        const baseDir = `${opts.repositoryOwnerParentDirectory}${path.sep}${id.owner}${path.sep}${id.repo}`;
        const frr = FileSystemRemoteRepoRef_1.FileSystemRemoteRepoRef.fromDirectory({
            repositoryOwnerParentDirectory: opts.repositoryOwnerParentDirectory,
            mergePullRequests: opts.mergePullRequests,
            baseDir,
        });
        // Override target repo to get file url
        Object.defineProperty(params.target, "repoRef", {
            get() {
                return frr;
            },
        });
        automation_client_1.logger.info("Persisting to [%s]", baseDir);
        if (yield fs.pathExists(baseDir)) {
            throw new Error(`Cannot write new project to [${baseDir}] as this directory already exists`);
        }
        const createdProject = yield automation_client_1.NodeFsLocalProject.copy(p, baseDir);
        yield runAndLog_1.runAndLog("git init", { cwd: baseDir });
        yield runAndLog_1.runAndLog("git add .", { cwd: baseDir });
        yield runAndLog_1.runAndLog(`git commit -a -m "${InitialCommitMessage}"`, { cwd: baseDir });
        yield addGitHooks_1.addGitHooksToProject(createdProject);
        yield emitEventsForNewProject(teamContext, opts, createdProject, id, automationClientFinder);
        return ActionResult_1.successOn(createdProject);
    });
}
exports.fileSystemProjectPersister = fileSystemProjectPersister;
/**
 * Send events that should apply to a new project
 */
function emitEventsForNewProject(workspaceContext, lc, createdProject, id, automationClientFinder) {
    return __awaiter(this, void 0, void 0, function* () {
        const eventSender = invokeEventHandlerInProcess_1.invokeEventHandlerInProcess(workspaceContext);
        yield repoOnboardingEvents_1.sendRepoCreationEvent(workspaceContext, id, eventSender);
        const sha = yield git_1.lastSha(createdProject);
        const branch = "master";
        yield handlePushBasedEventOnRepo_1.handlePushBasedEventOnRepo(workspaceContext.workspaceId, eventSender, lc, {
            baseDir: createdProject.baseDir,
            sha,
            branch,
        }, "OnFirstPushToRepo");
        // This is the first push
        yield handlePushBasedEventOnRepo_1.handlePushBasedEventOnRepo(workspaceContext.workspaceId, eventSender, lc, {
            baseDir: createdProject.baseDir,
            sha,
            branch,
        }, "SetGoalsOnPush");
        yield repoOnboardingEvents_1.sendRepoOnboardingEvent(workspaceContext, id, eventSender);
        yield repoOnboardingEvents_1.sendChannelLinkEvent(workspaceContext, id, eventSender);
    });
}
//# sourceMappingURL=fileSystemProjectPersister.js.map