"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const globals_1 = require("@atomist/automation-client/lib/globals");
const port_1 = require("@atomist/automation-client/lib/util/port");
const sdm_core_1 = require("@atomist/sdm-core");
const assert = require("assert");
const exphbs = require("express-handlebars");
const stringify = require("json-stringify-safe");
const _ = require("lodash");
const path = require("path");
const newCorrelationId_1 = require("../../cli/invocation/http/support/newCorrelationId");
const EnvConfigWorkspaceContextResolver_1 = require("../../common/binding/EnvConfigWorkspaceContextResolver");
const parseCorrelationId_1 = require("../../common/invocation/parseCorrelationId");
const httpMessaging_1 = require("../../common/ui/httpMessaging");
const LocalGraphClient_1 = require("../binding/graph/LocalGraphClient");
const ActionStore_1 = require("../binding/message/ActionStore");
const BroadcastingMessageClient_1 = require("../binding/message/BroadcastingMessageClient");
const GoalEventForwardingMessageClient_1 = require("../binding/message/GoalEventForwardingMessageClient");
const HttpClientMessageClient_1 = require("../binding/message/HttpClientMessageClient");
const invokeCommandHandlerInProcess_1 = require("../invocation/invokeCommandHandlerInProcess");
const invokeEventHandlerInProcess_1 = require("../invocation/invokeEventHandlerInProcess");
const defaultLocalSoftwareDeliveryMachineConfiguration_1 = require("./defaultLocalSoftwareDeliveryMachineConfiguration");
const NotifyOnCompletionAutomationEventListener_1 = require("./support/NotifyOnCompletionAutomationEventListener");
const NotifyOnStartupAutomationEventListener_1 = require("./support/NotifyOnStartupAutomationEventListener");
/**
 * Configures an automation client in local mode
 * @param {LocalModeConfiguration} localModeConfiguration
 * @return {ConfigurationPostProcessor}
 */
function configureLocal(options = { forceLocal: false }) {
    return (config) => __awaiter(this, void 0, void 0, function* () {
        if (_.isEmpty(config.groups) && _.isEmpty(config.workspaceIds) && !sdm_core_1.isInLocalMode()) {
            throw new Error("No 'workspaceIds' provided in configuration. To start this SDM in local mode, run 'atomist start --local'. " +
                "To connect to the Atomist API, please configure your 'workspaceIds' by running 'atomist config'");
        }
        // Don't mess with a non local SDM
        if (!(options.forceLocal || sdm_core_1.isInLocalMode())) {
            return config;
        }
        const workspaceContext = new EnvConfigWorkspaceContextResolver_1.EnvConfigWorkspaceContextResolver().workspaceContext;
        const defaultSdmConfiguration = defaultLocalSoftwareDeliveryMachineConfiguration_1.defaultLocalSoftwareDeliveryMachineConfiguration(config, workspaceContext);
        const mergedConfig = _.merge(defaultSdmConfiguration, config);
        // Set up workspaceIds and apiKey
        if (_.isEmpty(mergedConfig.workspaceIds)) {
            mergedConfig.workspaceIds = [workspaceContext.workspaceId];
        }
        if (_.isEmpty(config.apiKey)) {
            mergedConfig.apiKey = automation_client_1.guid();
        }
        mergedConfig.ws.enabled = false;
        mergedConfig.cluster.enabled = false;
        const globalActionStore = ActionStore_1.freshActionStore();
        yield configureWebEndpoints(mergedConfig, workspaceContext, globalActionStore);
        configureMessageClientFactory(mergedConfig, workspaceContext, globalActionStore);
        configureGraphClient(mergedConfig);
        configureListeners(mergedConfig);
        return mergedConfig;
    });
}
exports.configureLocal = configureLocal;
function configureWebEndpoints(configuration, teamContext, actionStore) {
    return __awaiter(this, void 0, void 0, function* () {
        _.set(configuration, "http.enabled", true);
        // Binding to localhost will prevent Express to be accessible from a different computer on the network
        // which is good as we are disabling all security further down
        _.set(configuration, "http.host", configuration.local.hostname);
        if (!_.get(configuration, "http.port")) {
            _.set(configuration, "http.port", yield port_1.scanFreePort());
        }
        // Disable auth as we're only expecting local clients
        _.set(configuration, "http.auth.basic.enabled", false);
        _.set(configuration, "http.auth.token.enabled", false);
        _.set(configuration, "http.auth.bearer.enabled", false);
        process.env.ATOMIST_WEBHOOK_BASEURL = `http://${configuration.local.hostname}:${configuration.http.port}`;
        configuration.http.customizers = [
            ...(configuration.http.customizers || []),
            (app) => {
                // TODO could use this to set local mode for a server - e.g. the name to send to
                app.get("/local/configuration", (req, res) => __awaiter(this, void 0, void 0, function* () {
                    res.json(configuration.local);
                }));
                const bodyParser = require("body-parser");
                app.use(bodyParser.urlencoded({ extended: false }));
                app.use(bodyParser.json());
                // Handlebars setup
                app.set("view engine", "handlebars");
                app.set("views", path.join(__dirname, "..", "..", "views"));
                app.engine("handlebars", exphbs({
                    defaultLayout: "main",
                    layoutsDir: path.join(__dirname, "..", "..", "views", "layouts"),
                }));
                // Add a GET route for convenient links to command handler invocation, as a normal automation client doesn't expose one
                app.get("/command/:name", (req, res) => __awaiter(this, void 0, void 0, function* () {
                    const payload = req.query;
                    const command = automation_client_1.automationClientInstance().automations.automations.commands.find(c => c.name === req.params.name);
                    if (!command) {
                        return res.status(404).send(`Command '${req.params.name}' not found`);
                    }
                    return res.render("command", {
                        payload,
                        command,
                        configuration: automation_client_1.automationClientInstance().configuration,
                    });
                }));
                app.post("/command/:name", (req, res) => __awaiter(this, void 0, void 0, function* () {
                    const command = automation_client_1.automationClientInstance().automations.automations.commands.find(c => c.name === req.params.name);
                    if (!command) {
                        return res.status(404).send(`Command '${req.params.name}' not found`);
                    }
                    const payload = req.body;
                    const invocation = {
                        name: req.params.name,
                        parameters: payload,
                        mappedParameters: payload,
                        workspaceName: teamContext.workspaceName,
                        workspaceId: teamContext.workspaceId,
                        correlationId: yield newCorrelationId_1.newCliCorrelationId(),
                    };
                    return invokeCommandHandlerInProcess_1.invokeCommandHandlerInProcess((result) => __awaiter(this, void 0, void 0, function* () {
                        const r = yield result;
                        return res.render("result", {
                            result: JSON.stringify(decircle(r), undefined, 2),
                            command,
                            configuration: automation_client_1.automationClientInstance().configuration,
                        });
                    }))(invocation);
                }));
                app.post("/atomist/link-image/teams/:team", (req, res) => __awaiter(this, void 0, void 0, function* () {
                    const payload = req.body;
                    // TODO Hack to get image into the Push
                    globals_1.eventStore().messages().filter(m => m.value.sha === payload.git.sha && m.value.goalSet && m.value.goalSetId)
                        .forEach(m => _.set(m.value, "push.after.image.imageName", payload.docker.image));
                    const event = automation_client_1.automationClientInstance().automations.automations.events.find(e => e.name === "FindArtifactOnImageLinked");
                    if (!event) {
                        return res.status(200).send(`Event 'FindArtifactOnImageLinked' not found`);
                    }
                    const invocation = {
                        name: "FindArtifactOnImageLinked",
                        payload: {
                            ImageLinked: [{
                                    commit: {
                                        sha: payload.git.sha,
                                        repo: {
                                            owner: payload.git.owner,
                                            name: payload.git.repo,
                                            org: {
                                                provider: {
                                                    providerId: "n/a",
                                                },
                                            },
                                        },
                                    },
                                    image: {
                                        image: payload.docker.image,
                                        imageName: payload.docker.image,
                                    },
                                }],
                        },
                    };
                    return invokeEventHandlerInProcess_1.invokeEventHandlerInProcess({ workspaceId: req.params.team, workspaceName: req.params.team })(invocation)
                        .then(resp => res.json(decircle(resp)), boo => res.status(500).send(boo.message));
                }));
                app.post("/atomist/build/teams/:team", (req, res) => __awaiter(this, void 0, void 0, function* () {
                    const event = automation_client_1.automationClientInstance().automations.automations.events.find(e => e.name === "InvokeListenersOnBuildComplete");
                    if (!event) {
                        return res.status(200).send(`Event 'InvokeListenersOnBuildComplete' not found`);
                    }
                    const body = req.body;
                    const build = {
                        Build: [{
                                buildId: body.number,
                                status: body.status,
                                commit: {
                                    sha: body.commit,
                                    message: "",
                                    repo: {
                                        name: body.repository.name,
                                        owner: body.repository.owner_name,
                                        channels: [{
                                                name: body.repository.name,
                                                id: body.repository.name,
                                                team: {
                                                    id: req.params.team,
                                                },
                                            }],
                                    },
                                    statuses: [],
                                },
                            }],
                    };
                    const push = globals_1.eventStore().messages().find(m => m.value.sha === body.commit).value.push;
                    build.Build[0].push = push;
                    const invocation = {
                        name: "InvokeListenersOnBuildComplete",
                        payload: build,
                    };
                    return invokeEventHandlerInProcess_1.invokeEventHandlerInProcess({ workspaceId: req.params.team, workspaceName: req.params.team })(invocation)
                        .then(resp => res.json(decircle(resp)), boo => res.status(500).send(boo.message));
                }));
                app.get(ActionStore_1.ActionRoute + "/:description", (req, res) => __awaiter(this, void 0, void 0, function* () {
                    automation_client_1.logger.debug("Action clicked: params=%j; query=%j", req.params, req.query);
                    const actionKey = req.query.key;
                    if (!actionKey) {
                        automation_client_1.logger.error("No action key provided. Please include ?key=< actionKey that has been stored by this sdm >");
                        return res.status(404).send("Required query parameter: key");
                    }
                    const storedAction = yield actionStore.getAction(actionKey);
                    if (!storedAction) {
                        automation_client_1.logger.error("Action key %s not found", actionKey);
                        return res.status(404).send("Action not stored. Did you restart the SDM?");
                    }
                    const storedCommand = storedAction.command;
                    if (!storedCommand) {
                        automation_client_1.logger.error("No command stored on action object: %j", storedAction);
                        return res.status(500).send("This will never work");
                    }
                    const command = automation_client_1.automationClientInstance().automations.automations.commands.find(c => c.name === storedCommand.name);
                    if (!command) {
                        return res.status(404).send(`Command '${req.params.name}' not found`);
                    }
                    automation_client_1.logger.debug("The command name is: %s", storedCommand.name);
                    automation_client_1.logger.debug("The parameters are: %j", command.parameters);
                    automation_client_1.logger.debug("The stored parameters are: %j", storedCommand.parameters);
                    automation_client_1.logger.debug("The command description is: %s", storedCommand.description);
                    return res.render("command", {
                        payload: storedCommand.parameters,
                        command,
                        configuration: automation_client_1.automationClientInstance().configuration,
                    });
                }));
                app.post("/atomist/fingerprints/teams/:team", (req, res) => __awaiter(this, void 0, void 0, function* () {
                    automation_client_1.logger.warn("Received fingerprint: " + JSON.stringify(req.body));
                    return res.status(200).send(`Ignoring fingerprint. Thanks though`);
                }));
            },
        ];
    });
}
function configureListeners(configuration) {
    if (!configuration.listeners) {
        configuration.listeners = [];
    }
    configuration.listeners.push(new NotifyOnCompletionAutomationEventListener_1.NotifyOnCompletionAutomationEventListener());
    configuration.listeners.push(new NotifyOnStartupAutomationEventListener_1.NotifyOnStartupAutomationEventListener());
}
function decircle(result) {
    let noncircular = result;
    try {
        JSON.stringify(noncircular);
    }
    catch (err) {
        automation_client_1.logger.error("Circular object returned from event handler: %s", stringify(result));
        noncircular = { code: result.code, message: stringify(result.message) };
        automation_client_1.logger.error("Substituting for circular object: %j", noncircular);
    }
    return noncircular;
}
/**
 * Use custom message client to update HTTP listeners and forward goal events back to the SDM via HTTP
 * @param {Configuration} configuration
 * @param {LocalModeConfiguration} localMachineConfig
 */
function configureMessageClientFactory(configuration, teamContext, actionStore) {
    configuration.http.messageClientFactory =
        aca => {
            assert(!!aca.context.correlationId);
            const channel = parseCorrelationId_1.parseChannel(aca.context.correlationId);
            const portToCommunicateToListenerOn = parseCorrelationId_1.parsePort(aca.context.correlationId);
            return new BroadcastingMessageClient_1.BroadcastingMessageClient(new HttpClientMessageClient_1.HttpClientMessageClient({
                workspaceId: teamContext.workspaceId,
                channel,
                port: httpMessaging_1.AllMessagesPort,
                actionStore,
                transient: false,
            }), new GoalEventForwardingMessageClient_1.GoalEventForwardingMessageClient(), 
            // Communicate back to client if possible
            !!portToCommunicateToListenerOn ? new HttpClientMessageClient_1.HttpClientMessageClient({
                workspaceId: teamContext.workspaceId,
                channel,
                port: portToCommunicateToListenerOn,
                actionStore,
                transient: true,
            }) : undefined);
        };
}
function configureGraphClient(configuration) {
    configuration.http.graphClientFactory =
        () => new LocalGraphClient_1.LocalGraphClient(false);
}
//# sourceMappingURL=configureLocal.js.map