"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
// tslint:disable:deprecation
const automation_client_1 = require("@atomist/automation-client");
const configuration_1 = require("@atomist/automation-client/lib/configuration");
const sdm_1 = require("@atomist/sdm");
const sdm_core_1 = require("@atomist/sdm-core");
const fs = require("fs-extra");
const _ = require("lodash");
const os = require("os");
const path = require("path");
const defaultAutomationClientFinder_1 = require("../../cli/invocation/http/support/defaultAutomationClientFinder");
const defaultWorkspaceContextResolver_1 = require("../../common/binding/defaultWorkspaceContextResolver");
const EnvironmentTokenCredentialsResolver_1 = require("../binding/EnvironmentTokenCredentialsResolver");
const SimpleNodeLoggerProgressLog_1 = require("../binding/log/SimpleNodeLoggerProgressLog");
const expandedTreeRepoFinder_1 = require("../binding/project/expandedTreeRepoFinder");
const ExpandedTreeRepoRefResolver_1 = require("../binding/project/ExpandedTreeRepoRefResolver");
const FileSystemProjectLoader_1 = require("../binding/project/FileSystemProjectLoader");
const fileSystemProjectPersister_1 = require("../binding/project/fileSystemProjectPersister");
const LocalRepoTargets_1 = require("../binding/project/LocalRepoTargets");
const DefaultAtomistRoot = path.join("atomist", "projects");
/**
 * Defaults for local-SDM configuration
 */
function defaultLocalSoftwareDeliveryMachineConfiguration(configuration, workspaceContext = defaultWorkspaceContextResolver_1.DefaultWorkspaceContextResolver.workspaceContext) {
    const automationClientFinder = defaultAutomationClientFinder_1.defaultAutomationClientFinder();
    const defaultLocalSdmConfiguration = {
        preferLocalSeeds: true,
        mergeAutofixes: true,
        mergePullRequests: false,
        repositoryOwnerParentDirectory: determineDefaultRepositoryOwnerParentDirectory(),
        hostname: determineDefaultHostUrl(),
    };
    const localSdmConfiguration = _.merge(defaultLocalSdmConfiguration, configuration.local);
    const sdmConfiguration = {
        artifactStore: new sdm_core_1.EphemeralLocalArtifactStore(),
        projectLoader: new FileSystemProjectLoader_1.FileSystemProjectLoader(new sdm_1.CachingProjectLoader(), localSdmConfiguration),
        logFactory: (context, goal) => __awaiter(this, void 0, void 0, function* () { return new SimpleNodeLoggerProgressLog_1.SimpleNodeLoggerProgressLog(configuration.name, goal.name, path.join(os.homedir(), ".atomist", "log")); }),
        credentialsResolver: new EnvironmentTokenCredentialsResolver_1.EnvironmentTokenCredentialsResolver(),
        preferenceStoreFactory: sdm_core_1.FilePreferenceStoreFactory,
        repoRefResolver: new ExpandedTreeRepoRefResolver_1.ExpandedTreeRepoRefResolver(localSdmConfiguration),
        repoFinder: expandedTreeRepoFinder_1.expandedTreeRepoFinder(localSdmConfiguration),
        projectPersister: fileSystemProjectPersister_1.fileSystemProjectPersister(workspaceContext, localSdmConfiguration, automationClientFinder),
        targets: () => new LocalRepoTargets_1.LocalRepoTargets(localSdmConfiguration),
    };
    const result = {
        sdm: sdmConfiguration,
        local: localSdmConfiguration,
    };
    return result;
}
exports.defaultLocalSoftwareDeliveryMachineConfiguration = defaultLocalSoftwareDeliveryMachineConfiguration;
const HostnamePath = "local.hostname";
const RepositoryOwnerParentDirectoryPath = "local.repositoryOwnerParentDirectory";
/**
 * Resolve the repositoryOwnerParentDirectory.
 * This will search in the client configuration, user config in .atomist, at ATOMIST_ROOT env var and
 * lastly default to ~/atomist/projects.
 *
 * The path will get created if it doesn't exist.
 */
function determineDefaultRepositoryOwnerParentDirectory() {
    let root;
    try {
        root = automation_client_1.configurationValue(RepositoryOwnerParentDirectoryPath);
    }
    catch (err) {
        const config = configuration_1.getUserConfig();
        if (config) {
            root = _.get(config, RepositoryOwnerParentDirectoryPath);
        }
    }
    if (!root) {
        root = process.env.ATOMIST_ROOT || path.join(os.homedir(), DefaultAtomistRoot);
    }
    if (!fs.existsSync(root)) {
        automation_client_1.logger.info(`Creating Atomist repository owner parent directory at '${root}'`);
        fs.mkdirsSync(root);
    }
    return root;
}
exports.determineDefaultRepositoryOwnerParentDirectory = determineDefaultRepositoryOwnerParentDirectory;
/**
 * Resolve the hostname to use for calling local urls.
 * This will search in the client configuration, user config in .atomist and finally
 * default to 127.0.0.1.
 */
function determineDefaultHostUrl() {
    try {
        return automation_client_1.configurationValue(HostnamePath);
    }
    catch (err) {
        const config = configuration_1.getUserConfig();
        if (config) {
            const hostname = _.get(config, HostnamePath);
            if (hostname) {
                return hostname;
            }
        }
    }
    return "127.0.0.1";
}
exports.determineDefaultHostUrl = determineDefaultHostUrl;
//# sourceMappingURL=defaultLocalSoftwareDeliveryMachineConfiguration.js.map