"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const sdm_1 = require("@atomist/sdm");
const chalk_1 = require("chalk");
const formatDate = require("format-date");
const _ = require("lodash");
const marked = require("marked");
const TerminalRenderer = require("marked-terminal");
const strip_ansi_1 = require("strip-ansi");
const progressBar_1 = require("../../../bin/progressBar");
const consoleOutput_1 = require("../../cli/ui/consoleOutput");
marked.setOptions({
    // Define custom renderer
    renderer: new TerminalRenderer(),
});
// tslint:disable-next-line:no-var-requires
const term = require("terminal-kit").terminal;
term.options.crlf = true;
const SchemaRequested = ":spinner :icon :name :bar :description :link";
const SchemaSuccess = "  :icon :name :bar :description :link";
const Frames = [
    "⠋",
    "⠙",
    "⠹",
    "⠸",
    "⠼",
    "⠴",
    "⠦",
    "⠧",
    "⠇",
    "⠏",
];
class ConsoleGoalRendering {
    constructor() {
        this.goalSets = [];
        this.unkownGoals = [];
        term.windowTitle("Atomist - SDM Goals");
        [`exit`, `SIGINT`, `SIGUSR1`, `SIGUSR2`, `uncaughtException`, `SIGTERM`].forEach(eventType => {
            process.on(eventType, () => {
                term.hideCursor(false);
                process.exit(0);
            });
        });
        term.getCursorLocation((cb, x, y) => {
            consoleOutput_1.infoMessage("Listening for SDM goals...");
            term.moveTo(x, y);
            progressBar_1.init({ row: y, col: x });
            term.hideCursor();
            setInterval(() => {
                this.goalSets.forEach(gs => gs.goals.forEach(g => {
                    const bar = g.bar;
                    if (bar.completed) {
                        bar.setSchema(SchemaSuccess, {
                            name: mapStateToColor(g.displayName, g.goal.state),
                            description: formatDescription(g.goal.description),
                            link: formatLink(g.goal.url),
                            icon: mapStateToIcon(g.goal.state),
                            spinner: " ",
                        });
                        bar.archived = true;
                    }
                    else if (g.goal.state === sdm_1.SdmGoalState.in_process) {
                        let tick = g.tick;
                        if (tick + 1 === Frames.length) {
                            tick = 0;
                        }
                        else {
                            tick++;
                        }
                        bar.setSchema(SchemaRequested, {
                            name: mapStateToColor(g.displayName, g.goal.state),
                            description: formatDescription(g.goal.description),
                            link: formatLink(g.goal.url),
                            icon: mapStateToIcon(g.goal.state),
                            spinner: chalk_1.default.grey(Frames[g.tick]),
                        });
                        g.tick = tick;
                    }
                }));
                this.goalSets.filter(gs => !gs.goals.some(g => !g.bar.completed))
                    .forEach(gs => {
                    this.goalSets.splice(this.goalSets.indexOf(gs), 1);
                });
            }, 100);
        });
    }
    addGoals(id, goals, p) {
        process.stdout.write(push(p) + "\n");
        const gss = goals.map(g => ({
            name: g,
            displayName: strip_ansi_1.default(marked(g.split("#")[0]).trim()),
        }));
        const ml = _.maxBy(gss, "displayName.length");
        const gs = gss.map(g => ({
            displayName: _.padEnd(g.displayName, ml.name.length, " "),
            name: g.name,
        }));
        const ugs = this.unkownGoals.filter(g => g.goalSetId === id);
        const bars = gs.map(g => {
            const ug = ugs.find(tug => tug.name === g.name);
            if (ug) {
                this.unkownGoals.slice(this.unkownGoals.indexOf(ug), 1);
            }
            return {
                name: g.name,
                displayName: g.displayName,
                goal: {
                    description: (ug ? ug.description : "") || "",
                    url: (ug && ug.externalUrls ? ug.externalUrls.map(e => e.url).join(" ") : "") || "",
                    stage: (ug ? ug.phase : "") || "",
                    state: ug ? ug.state : sdm_1.SdmGoalState.planned,
                },
                bar: this.createBar(g.displayName, ug ? ug.description : `Planned: ${g.displayName}`, (ug ? ug.url : "") || "", ug ? ug.state : sdm_1.SdmGoalState.planned),
                tick: 0,
            };
        });
        this.goalSets.push({
            goals: bars,
            goalSetId: id,
            push: p,
        });
    }
    updateGoal(goal) {
        const goalSet = this.goalSets.find(gs => gs.goalSetId === goal.goalSetId);
        if (goalSet) {
            const gtu = goalSet.goals.find(g => g.name === goal.name);
            if (gtu) {
                gtu.goal.url = goal.externalUrls ? goal.externalUrls.map(e => e.url).join(" ") : "";
                if (goal.state === sdm_1.SdmGoalState.failure || goal.state === sdm_1.SdmGoalState.in_process) {
                    gtu.goal.url = goal.url || "";
                }
                gtu.goal.description = goal.description || "";
                if (goal.phase && goal.state !== sdm_1.SdmGoalState.success) {
                    gtu.goal.description += chalk_1.default.gray(` ${goal.phase}`);
                }
                gtu.goal.state = goal.state;
                // Update the bar
                gtu.bar.update(mapStateToRatio(goal.state), {
                    name: mapStateToColor(gtu.displayName, gtu.goal.state),
                    description: formatDescription(gtu.goal.description),
                    link: formatLink(gtu.goal.url),
                    icon: mapStateToIcon(gtu.goal.state),
                    spinner: " ",
                }, false);
            }
        }
        else {
            const ix = this.unkownGoals.findIndex(g => g.name === goal.name && g.goalSetId === goal.goalSetId);
            if (ix >= 0) {
                this.unkownGoals[ix] = goal;
            }
            else {
                this.unkownGoals.push(goal);
            }
        }
    }
    createBar(name, description, link, state) {
        const bar = new progressBar_1.ProgressBar({
            schema: SchemaRequested,
            total: 5,
            clean: false,
            width: 5,
            filled: ".",
            blank: " ",
            spinner: " ",
            fixedWidth: true,
        });
        bar.update(mapStateToRatio(state), {
            name: mapStateToColor(name, state),
            description: formatDescription(description),
            link: formatLink(link),
            icon: mapStateToIcon(state),
            spinner: " ",
        }, true);
        return bar;
    }
}
exports.ConsoleGoalRendering = ConsoleGoalRendering;
function formatDescription(description) {
    return marked(description, { gfm: true, breaks: false }).trim();
}
function formatLink(url) {
    if (url && url.length > 0) {
        return `∞ ${chalk_1.default.grey(url)}`;
    }
    return "";
}
function mapStateToRatio(state) {
    switch (state) {
        case sdm_1.SdmGoalState.planned:
        case sdm_1.SdmGoalState.waiting_for_pre_approval:
        case sdm_1.SdmGoalState.pre_approved:
            return 0;
        case sdm_1.SdmGoalState.requested:
            return 0.25;
        case sdm_1.SdmGoalState.in_process:
            return 0.5;
        case sdm_1.SdmGoalState.waiting_for_approval:
        case sdm_1.SdmGoalState.approved:
            return 0.75;
        case sdm_1.SdmGoalState.failure:
        case sdm_1.SdmGoalState.success:
        case sdm_1.SdmGoalState.skipped:
        case sdm_1.SdmGoalState.stopped:
        case sdm_1.SdmGoalState.canceled:
            return 1;
    }
    return 0;
}
function mapStateToIcon(state) {
    switch (state) {
        case sdm_1.SdmGoalState.planned:
            return chalk_1.default.gray("▫");
        case sdm_1.SdmGoalState.waiting_for_pre_approval:
        case sdm_1.SdmGoalState.pre_approved:
            return chalk_1.default.gray("॥");
        case sdm_1.SdmGoalState.requested:
            return chalk_1.default.gray("▹");
        case sdm_1.SdmGoalState.in_process:
            return chalk_1.default.yellow("▸");
        case sdm_1.SdmGoalState.waiting_for_approval:
        case sdm_1.SdmGoalState.approved:
            return chalk_1.default.yellow("॥");
        case sdm_1.SdmGoalState.failure:
            return chalk_1.default.red("⨯");
        case sdm_1.SdmGoalState.success:
            return chalk_1.default.green("✓");
        case sdm_1.SdmGoalState.stopped:
            return chalk_1.default.yellow("▪");
        case sdm_1.SdmGoalState.skipped:
            return chalk_1.default.gray("⬩");
        case sdm_1.SdmGoalState.canceled:
            return chalk_1.default.gray("⁃");
    }
    return "";
}
function mapStateToColor(label, state) {
    switch (state) {
        case sdm_1.SdmGoalState.planned:
        case sdm_1.SdmGoalState.waiting_for_pre_approval:
        case sdm_1.SdmGoalState.pre_approved:
        case sdm_1.SdmGoalState.requested:
            return chalk_1.default.gray(label);
        case sdm_1.SdmGoalState.in_process:
            return chalk_1.default.yellow(label);
        case sdm_1.SdmGoalState.waiting_for_approval:
        case sdm_1.SdmGoalState.approved:
            return chalk_1.default.yellow(label);
        case sdm_1.SdmGoalState.failure:
            return chalk_1.default.red(label);
        case sdm_1.SdmGoalState.success:
            return chalk_1.default.green(label);
        case sdm_1.SdmGoalState.stopped:
            return chalk_1.default.yellow(label);
        case sdm_1.SdmGoalState.skipped:
            return chalk_1.default.gray(label);
        case sdm_1.SdmGoalState.canceled:
            return chalk_1.default.gray(label);
    }
    return "";
}
function date() {
    return chalk_1.default.dim(formatDate("{year}-{month}-{day} {hours}:{minutes}:{seconds}", new Date()));
}
function push(p) {
    return `${chalk_1.default.grey("#")} ${chalk_1.default.bold(p.repo)} ${date()} ${chalk_1.default.yellow(`${p.owner}/${p.repo}/${p.branch}`)} - ${chalk_1.default.yellow(p.sha.slice(0, 7))} ${p.message.split("\n")[0]}`;
}
// tslint:disable
/*
const c = new ConsoleGoalRendering();
let counter = 0;
setInterval(() => {
    counter++;
    if (counter > 3) {
        return;
    }
    const id = Date.now().toString();
    c.addGoals(id, ["autofix#Test.ts:14", "code review", "code reaction", "build", "deploy locally"], {
        branch: "master",
        repo: "cli",
        owner: "atomist",
        sha: id,
        message: "Update dependencies",
    });

    setTimeout(() => {
        c.updateGoal({
            goalSetId: id,
            name: "autofix#Test.ts:14",
            description: "Ready to autofix",
            state: SdmGoalState.requested,
        } as SdmGoalEvent);
    }, 500);

    setTimeout(() => {
        c.updateGoal({
            goalSetId: id,
            name: "autofix#Test.ts:14",
            description: "Running autofix",
            state: SdmGoalState.in_process,
            url: "https://app.atomist.com/workspace/T29E48P34/logs/atomist/sdm-local/adf25354cdbbde523e39c8c318821dca6a9e7674/0-code/review/d19bcdd0-4e12-4d65-9b9e-5096703ddbc1/5b64acdf-cac4-42a4-a7a7-1adeb549bf82",
        } as SdmGoalEvent);
    }, 2500);

    setTimeout(() => {
        c.updateGoal({
            goalSetId: id,
            name: "autofix#Test.ts:14",
            description: "Autofixed `bla`",
            state: SdmGoalState.success,
        } as SdmGoalEvent);
    }, 4500);

    setTimeout(() => {
        c.updateGoal({
            goalSetId: id,
            name: "build",
            description: "Ready to build",
            state: SdmGoalState.requested,
        } as SdmGoalEvent);
    }, 4500);

    setTimeout(() => {
        c.updateGoal({
            goalSetId: id,
            name: "build",
            description: "Building",
            state: SdmGoalState.in_process,
        } as SdmGoalEvent);
    }, 5000);

    setTimeout(() => {
        c.updateGoal({
            goalSetId: id,
            name: "build",
            description: "Build successful",
            state: SdmGoalState.success,
            phase: "tsc",
        } as SdmGoalEvent);
    }, 8000);

    setTimeout(() => {
        c.updateGoal({
            goalSetId: id,
            name: "deploy locally",
            description: "Ready to deploy",
            state: SdmGoalState.requested,
        } as SdmGoalEvent);
    }, 8500);

    setTimeout(() => {
        c.updateGoal({
            goalSetId: id,
            name: "deploy locally",
            description: "Deploying",
            state: SdmGoalState.in_process,
        } as SdmGoalEvent);
    }, 10000);

    setTimeout(() => {
        c.updateGoal({
            goalSetId: id,
            name: "deploy locally",
            description: "Deploy failed",
            state: SdmGoalState.failure,
            phase: "tsc",
            url: "https://app.atomist.com/workspace/T29E48P34/logs/atomist/sdm-local/adf25354cdbbde523e39c8c318821dca6a9e7674/0-code/review/d19bcdd0-4e12-4d65-9b9e-5096703ddbc1/5b64acdf-cac4-42a4-a7a7-1adeb549bf82",
        } as SdmGoalEvent);
    }, 11000);

}, 2000);
*/
//# sourceMappingURL=ConsoleGoalRendering.js.map