"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const chalk_1 = require("chalk");
const formatDate = require("format-date");
const _ = require("lodash");
const marked = require("marked");
const TerminalRenderer = require("marked-terminal");
const ActionStore_1 = require("../binding/message/ActionStore");
const GoalEventForwardingMessageClient_1 = require("../binding/message/GoalEventForwardingMessageClient");
marked.setOptions({
    // Define custom renderer
    renderer: new TerminalRenderer(),
});
/**
 * Sender that writes to console using process.stdout.
 * This works even if console out is rerouted to logging.
 * @param msg message
 */
exports.ProcessStdoutSender = msg => Promise.resolve(process.stdout.write(msg));
/**
 * Message client logging to the console. Uses color and renders markdown
 */
class ConsoleMessageClient {
    /**
     * Construct a new ConsoleMessageClient, supporting actionable messages.
     * @param {string} linkedChannel
     * @param {Sender} sender
     * @param {AutomationClientConnectionRequest} connectionConfig config to use to
     * format URLs to call back the source SDM
     * @param {marked.MarkedOptions} markedOptions
     */
    constructor(linkedChannel, sender, connectionConfig, markedOptions = {
        breaks: false,
    }) {
        this.linkedChannel = linkedChannel;
        this.sender = sender;
        this.connectionConfig = connectionConfig;
        this.markedOptions = markedOptions;
    }
    respond(msg, options) {
        return __awaiter(this, void 0, void 0, function* () {
            automation_client_1.logger.debug("MessageClient.respond: Raw mesg=\n%j\n", msg);
            return this.addressChannels(msg, this.linkedChannel, options);
        });
    }
    send(msg, destinations, options) {
        return __awaiter(this, void 0, void 0, function* () {
            if (GoalEventForwardingMessageClient_1.isSdmGoalStoreOrUpdate(msg)) {
                return;
            }
            const dests = (Array.isArray(destinations) ? destinations : [destinations])
                .filter((a) => a.userAgent !== "ingester");
            return this.addressChannels(msg, _.flatten(dests.map(d => d.channels)), options);
        });
    }
    addressChannels(msg, channels, options) {
        return __awaiter(this, void 0, void 0, function* () {
            automation_client_1.logger.debug("MessageClient.addressChannels: Raw mesg=\n%j\nChannels=%s\n", msg, channels);
            const chans = automation_client_1.toStringArray(channels);
            chans.forEach((channel) => __awaiter(this, void 0, void 0, function* () {
                // TODO isSlackMessage doesn't return right
                if (automation_client_1.isSlackMessage(msg)) {
                    if (!!msg.text) {
                        yield this.writeToChannel(channel, msg.text);
                    }
                    const blocks = [];
                    (msg.attachments || []).forEach(att => {
                        let text = "";
                        if (!!att.author_name) {
                            text += `__${att.author_name}__\n`;
                        }
                        if (!!att.title) {
                            text += `**${att.title}**\n`;
                        }
                        if (!!att.text) {
                            text += att.text;
                        }
                        if (!!text) {
                            blocks.push(text);
                        }
                        (att.actions || []).forEach((action, index) => {
                            blocks.push(this.renderAction(channel, action, ActionStore_1.actionKeyFor(msg, index)));
                        });
                        (att.fields || []).forEach(field => {
                            blocks.push(`${field.title}: ${field.value}`);
                        });
                    });
                    yield this.writeToChannel(channel, blocks.join("\n"));
                }
                else if (typeof msg === "string") {
                    yield this.writeToChannel(channel, msg);
                }
                else {
                    const m = msg;
                    if (!!m.content) {
                        yield this.writeToChannel(channel, m.content);
                    }
                    else {
                        yield this.writeToChannel(channel, "???? What is " + JSON.stringify(msg));
                    }
                }
            }));
        });
    }
    addressUsers(msg, users, options) {
        return __awaiter(this, void 0, void 0, function* () {
            automation_client_1.logger.debug("MessageClient.addressUsers: Raw mesg=\n%j\nUsers=%s", msg, users);
            return this.sender(`#${users} ${this.dateString()} ${msg}\n`);
        });
    }
    renderAction(channel, action, actionKey) {
        if (action.type === "button") {
            const url = `${this.connectionConfig.baseEndpoint}${ActionStore_1.ActionRoute}/${ActionStore_1.actionDescription(action)}?key=${actionKey}`;
            return `${action.text} - ${url}`;
        }
        else {
            return JSON.stringify(action);
        }
    }
    /**
     * Apply consistent formatting to mimic writing to a Slack channel
     * @param {string[] | string} channels
     * @param {string} markdown
     */
    writeToChannel(channels, markdown) {
        const outputText = ` ${marked(`**${channels}**`, this.markedOptions).trim()} ${this.dateString()} ${marked(markdown, this.markedOptions).trim()}`;
        return this.sender(chalk_1.default.gray("#") + outputText.split("\n").join("\n  ") + "\n");
    }
    dateString() {
        return chalk_1.default.dim(formatDate("{year}-{month}-{day} {hours}:{minutes}:{seconds}", new Date()));
    }
}
exports.ConsoleMessageClient = ConsoleMessageClient;
//# sourceMappingURL=ConsoleMessageClient.js.map