"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const bodyParser = require("body-parser");
const express = require("express");
const consoleOutput_1 = require("../../cli/ui/consoleOutput");
const CommandCompletionDestination_1 = require("../../common/ui/CommandCompletionDestination");
const httpMessaging_1 = require("../../common/ui/httpMessaging");
const canConnectTo_1 = require("../../common/util/http/canConnectTo");
const GoalEventForwardingMessageClient_1 = require("../binding/message/GoalEventForwardingMessageClient");
const defaultLocalSoftwareDeliveryMachineConfiguration_1 = require("../configuration/defaultLocalSoftwareDeliveryMachineConfiguration");
const NotifyOnCompletionAutomationEventListener_1 = require("../configuration/support/NotifyOnCompletionAutomationEventListener");
const ConsoleGoalRendering_1 = require("./ConsoleGoalRendering");
const ConsoleMessageClient_1 = require("./ConsoleMessageClient");
/**
 * Construction arguments to HttpMessageListener
 */
class HttpMessageListenerParameters {
}
exports.HttpMessageListenerParameters = HttpMessageListenerParameters;
/**
 * Start process to listen to HTTP POSTs from HttpClientMessageClient
 * and display them to the console. Supports both both structured
 * messages coming from a MessageClient and raw output sent
 * for diagnostic or other purposes.
 * The /message endpoint takes StreamedMessage
 * The /write endpoint takes a simple { message }
 */
class HttpMessageListener {
    /**
     * Start process to listen to HTTP messages from HttpClientMessageClient
     * and display them to the console
     */
    constructor(parameters) {
        this.parameters = parameters;
        this.channels = automation_client_1.toStringArray(parameters.channels || []);
        if (parameters.goals) {
            this.goalRenderer = new ConsoleGoalRendering_1.ConsoleGoalRendering();
        }
    }
    /**
     * Could this process be terminated, based on its configuration
     * and the events it has seen?
     * @return {boolean}
     */
    get canTerminate() {
        return this.parameters.transient && this.seenCompletion;
    }
    /**
     * Start the server process
     * @return {this}
     */
    start() {
        const app = express();
        app.use(bodyParser.json());
        app.get("/", (req, res) => res.send("Atomist Listener Daemon\n"));
        this.addMessageRoute(app);
        this.addGoalRoute(app);
        this.addWriteRoute(app);
        return this;
    }
    addMessageRoute(app) {
        app.post(httpMessaging_1.MessageRoute, (req, res, next) => {
            const destinations = req.body.destinations;
            // Shut down the listener
            if (this.parameters.transient && destinations.some((d) => d.rootType === CommandCompletionDestination_1.CommandCompletionDestination.rootType)) {
                if (NotifyOnCompletionAutomationEventListener_1.isFailureMessage(req.body.message)) {
                    consoleOutput_1.errorMessage("Command failure\n%s\n", req.body.message.error.message);
                }
                this.seenCompletion = true;
                res.send({ terminating: true });
                // This can be slow, so we terminate the process elsewhere
                return this.server.close();
            }
            if (this.channels.length > 0) {
                // Filter
                if (!this.channels.some(c => destinations.some(d => !!d.channels && d.channels.includes(c)))) {
                    return res.send({ ignored: true });
                }
            }
            if (!this.parameters.goals) {
                const messageClient = new ConsoleMessageClient_1.ConsoleMessageClient("general", ConsoleMessageClient_1.ProcessStdoutSender, req.body.machineAddress);
                return messageClient.send(req.body.message, req.body.destinations)
                    .then(() => res.send("Read message " + JSON.stringify(req.body) + "\n"))
                    .catch(next);
            }
            else {
                return next();
            }
        });
    }
    addGoalRoute(app) {
        app.post(httpMessaging_1.GoalRoute, (req, res, next) => {
            const destinations = req.body.destinations;
            if (this.channels.length > 0) {
                // Filter
                if (!this.channels.some(c => destinations.some(d => !!d.channels && d.channels.includes(c)))) {
                    return res.send({ ignored: true });
                }
            }
            if (!this.parameters.goals) {
                return next();
            }
            const body = req.body.message;
            if (GoalEventForwardingMessageClient_1.isSdmGoalStoreOrUpdate(body)) {
                this.goalRenderer.updateGoal(body);
            }
            else if (body.goals && body.push && body.goalSetId) {
                const push = body.push;
                this.goalRenderer.addGoals(body.goalSetId, body.goals, {
                    owner: push.repo.owner,
                    repo: push.repo.name,
                    branch: push.branch,
                    message: push.after.message,
                    sha: push.after.sha,
                });
            }
            return next();
        });
    }
    /**
     * Add raw write route
     */
    addWriteRoute(app) {
        // Raw message point. Ignore if not in verbose mode.
        const verbose = this.parameters.verbose;
        app.post("/write", (req, res) => {
            if (verbose) {
                consoleOutput_1.infoMessage(req.body.message);
            }
            res.send({ received: true, shown: verbose });
        });
        this.server = app.listen(this.parameters.port);
    }
}
exports.HttpMessageListener = HttpMessageListener;
/**
 * What will the URL of this listener be?
 * @param {number} demonPort
 * @return {string}
 */
function messageListenerEndpoint(demonPort) {
    return `http://${defaultLocalSoftwareDeliveryMachineConfiguration_1.determineDefaultHostUrl()}:${demonPort}${httpMessaging_1.MessageRoute}`;
}
exports.messageListenerEndpoint = messageListenerEndpoint;
function goalListenerEndpoint(demonPort) {
    return `http://${defaultLocalSoftwareDeliveryMachineConfiguration_1.determineDefaultHostUrl()}:${demonPort}${httpMessaging_1.GoalRoute}`;
}
exports.goalListenerEndpoint = goalListenerEndpoint;
function messageListenerRoot(demonPort) {
    return `http://${defaultLocalSoftwareDeliveryMachineConfiguration_1.determineDefaultHostUrl()}:${demonPort}/`;
}
exports.messageListenerRoot = messageListenerRoot;
/**
 * Is the all messages message listener running?
 * @param demonPort port to check on
 */
function isFeedListenerRunning(demonPort = httpMessaging_1.AllMessagesPort) {
    return __awaiter(this, void 0, void 0, function* () {
        return canConnectTo_1.canConnectTo(messageListenerRoot(demonPort));
    });
}
exports.isFeedListenerRunning = isFeedListenerRunning;
//# sourceMappingURL=HttpMessageListener.js.map