"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const sdm_1 = require("@atomist/sdm");
const sdm_core_1 = require("@atomist/sdm-core");
const chalk_1 = require("chalk");
const _ = require("lodash");
const FileSystemRemoteRepoRef_1 = require("../binding/project/FileSystemRemoteRepoRef");
/**
 * Add Local IO to the given SDM.
 * Analogous to the Slack lifecycle provided by the Atomist bot.
 * Messages sent are formatted for console output.
 */
exports.LocalLifecycle = Object.assign(Object.assign({}, sdm_1.metadata("local-lifecycle")), { configure: sdm => {
        if (sdm_core_1.isInLocalMode()) {
            addLocalLifecycle(sdm);
            addShowCreatedLocalRepo(sdm);
        }
    } });
function addShowCreatedLocalRepo(sdm) {
    sdm.addChannelLinkListener((i) => __awaiter(this, void 0, void 0, function* () {
        if (FileSystemRemoteRepoRef_1.isFileSystemRemoteRepoRef(i.id)) {
            return i.addressChannels(`🏛  Your new local repo is available at **${i.id.fileSystemLocation}**`);
        }
    }));
}
function pushIdentification(pu) {
    let msg = pu.commits[0].message.split("\n")[0];
    if (msg.length > 50) {
        msg = msg.slice(0, 47) + "...";
    }
    return `\`${pu.repo.owner}/${pu.repo.name}/${pu.branch}\` - \`${pu.commits[0].sha.slice(0, 7)}\` _${msg}_`;
}
function goalIndentification(g) {
    return _.lowerFirst(g.name);
}
function linkIndicator() {
    return chalk_1.default.grey(" >");
}
/**
 * Formatted for the console
 * @param {SoftwareDeliveryMachine} sdm
 */
function addLocalLifecycle(sdm) {
    sdm.addGoalCompletionListener((gcl) => __awaiter(this, void 0, void 0, function* () {
        switch (gcl.completedGoal.state) {
            case sdm_1.SdmGoalState.success:
                return gcl.addressChannels(`${pushIdentification(gcl.completedGoal.push)}
${chalk_1.default.green(`✓ ${goalIndentification(gcl.completedGoal)}`)} ${gcl.completedGoal.description}${links(gcl.completedGoal)}`);
            case sdm_1.SdmGoalState.stopped:
                return gcl.addressChannels(`${pushIdentification(gcl.completedGoal.push)}
${chalk_1.default.yellow(`▪ ${goalIndentification(gcl.completedGoal)}`)} ${gcl.completedGoal.description}${links(gcl.completedGoal)}`);
            case sdm_1.SdmGoalState.canceled:
                return gcl.addressChannels(`${pushIdentification(gcl.completedGoal.push)}
${chalk_1.default.gray(`⁃ ${goalIndentification(gcl.completedGoal)}`)} ${gcl.completedGoal.description}${links(gcl.completedGoal)}`);
            case sdm_1.SdmGoalState.skipped:
                return gcl.addressChannels(`${pushIdentification(gcl.completedGoal.push)}
${chalk_1.default.gray(`⬩ ${goalIndentification(gcl.completedGoal)}`)} ${gcl.completedGoal.description}${links(gcl.completedGoal)}`);
            case sdm_1.SdmGoalState.failure:
                return gcl.addressChannels(`${pushIdentification(gcl.completedGoal.push)}
${chalk_1.default.red(`⨯ ${goalIndentification(gcl.completedGoal)}`)} ${gcl.completedGoal.description}${gcl.completedGoal.url ? `${linkIndicator()} ${chalk_1.default.grey(gcl.completedGoal.url)}` : ""}${links(gcl.completedGoal)}`);
        }
    }));
    sdm.addGoalsSetListener((gsi) => __awaiter(this, void 0, void 0, function* () {
        const msg = pushIdentification(gsi.push) + (!gsi.goalSet ? "\nNo goals" : `
▸ Goals
${gsi.goalSet.goals.map(g => `⏦ ${chalk_1.default.italic(goalIndentification(g))}`).join("\n")}`);
        yield gsi.addressChannels(msg);
        return gsi.addressChannels({
            goalSetId: gsi.goalSetId,
            goals: !gsi.goalSet ? [] : gsi.goalSet.goals.map(g => g.name),
            push: gsi.push,
        });
    }));
    sdm.addGoalExecutionListener((gci) => __awaiter(this, void 0, void 0, function* () {
        switch (gci.goalEvent.state) {
            case sdm_1.SdmGoalState.requested:
                return gci.addressChannels(`${pushIdentification(gci.goalEvent.push)}
${chalk_1.default.yellow(`︎▹ ${goalIndentification(gci.goalEvent)}`)} ${gci.goalEvent.description}`);
            case sdm_1.SdmGoalState.in_process:
                return gci.addressChannels(`${pushIdentification(gci.goalEvent.push)}
${chalk_1.default.yellow(`︎▸ ${goalIndentification(gci.goalEvent)}`)} ${gci.goalEvent.description}`);
            default:
                break;
        }
    }));
}
function links(sdmGoal) {
    return sdmGoal.externalUrls && sdmGoal.externalUrls.length > 0 ? `${linkIndicator()} ${chalk_1.default.grey(sdmGoal.externalUrls.map(u => u.url).join(" "))}` : "";
}
//# sourceMappingURL=localLifecycle.js.map