"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const sdm_1 = require("@atomist/sdm");
const sdm_core_1 = require("@atomist/sdm-core");
/**
 * Return proper SDM goal externalUrls structure or `undefined` if
 * there is no externally accessible endpoint.
 */
function appExternalUrls(ka, ge) {
    return __awaiter(this, void 0, void 0, function* () {
        const url = yield endpointBaseUrl(ka);
        if (!url) {
            return undefined;
        }
        const label = url.split(":")[0];
        return [{ label, url }];
    });
}
exports.appExternalUrls = appExternalUrls;
/**
 * Create the URL for a deployment using the protocol, host, and path
 * from the [[KubernetesApplication]] object.  If `ka.path` is not
 * truthy, no ingress was created so return `undefined`.  If the path
 * does not begin and end with a forward slash, /, add them.  If
 * protocol is not provided, use "https" if tlsSecret is provided,
 * otherwise "http".  If host is not provided and the SDM is running
 * in local mode, it attempts to find the IP address of any locally
 * running Kubernetes cluster and falls back to "127.0.0.1".  If the
 * host is not given and the SDM is _not_ in local mode, `undefined`
 * is returned.
 *
 * @param ka Kubernetes application
 * @return endpoint URL for deployment service
 */
function endpointBaseUrl(ka) {
    return __awaiter(this, void 0, void 0, function* () {
        const path = endpointPath(ka);
        if (!path) {
            return undefined;
        }
        const host = yield kubeClusterHost(ka);
        if (!host) {
            return undefined;
        }
        const defaultProtocol = (ka.tlsSecret) ? "https" : "http";
        const protocol = (ka.protocol) ? ka.protocol : defaultProtocol;
        return `${protocol}://${host}${path}`;
    });
}
exports.endpointBaseUrl = endpointBaseUrl;
/**
 * Determine path for endpoint URL.  If `ka.path` is truthy, return it
 * after ensuring it starts and ends with single forward slash, "/".
 * (Typically an ingress path begins with a forward slash but does not
 * end with one.)  Otherwise, return `undefined`.
 *
 * @param ka Kubernetes application information
 * @return Standardized URL path that begins and ends with a single forward slash or `undefined`
 */
function endpointPath(ka) {
    if (!ka.path) {
        return undefined;
    }
    const path = (ka.path.startsWith("/")) ? ka.path : "/" + ka.path;
    const tail = (path.endsWith("/")) ? path : path + "/";
    return tail;
}
/**
 * Determine host for endpoint.  If `ka.host` is truthy, return it.
 * Otherwise if the SDM is running in local mode, try to find IP
 * address of locally running Kubernetes cluster.  Currently only
 * minikube is supported.  If it is unable to determine an IP address
 * of a local Kubernetes cluster, it returns "127.0.0.1".  If
 * `ka.host` is falsey and the SDM is _not_ running in local mode,
 * return `undefined`.
 *
 * @param ka Kubernetes application information
 * @return Hostname or, if local, IP address for application endpoint, or `undefined`
 */
function kubeClusterHost(ka) {
    return __awaiter(this, void 0, void 0, function* () {
        if (ka.host) {
            return ka.host;
        }
        if (!sdm_core_1.isInLocalMode()) {
            return undefined;
        }
        try {
            const minikubeIpResult = yield sdm_1.execPromise("minikube", ["ip"]);
            const host = minikubeIpResult.stdout.trim();
            if (host) {
                return host;
            }
        }
        catch (e) {
            automation_client_1.logger.debug(`Failed to run 'minikube ip': ${e.message}`);
        }
        return "127.0.0.1";
    });
}
//# sourceMappingURL=externalUrls.js.map