"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const error_1 = require("../support/error");
const retry_1 = require("../support/retry");
const api_1 = require("./api");
const config_1 = require("./config");
const resource_1 = require("./resource");
const spec_1 = require("./spec");
/**
 * Create or replace a Kubernetes resource using the provided spec.
 * This implementation uses read, patch, and create, but may switch to
 * [server-side
 * apply](https://github.com/kubernetes/enhancements/issues/555) when
 * it is available.
 *
 * @param spec Kubernetes resource spec sufficient to identify and create the resource
 * @return response from the Kubernetes API.
 */
function applySpec(spec) {
    return __awaiter(this, void 0, void 0, function* () {
        const slug = spec_1.specSlug(spec);
        let client;
        try {
            const kc = config_1.loadKubeConfig();
            client = kc.makeApiClient(api_1.K8sObjectApi);
        }
        catch (e) {
            e.message = `Failed to create Kubernetes client: ${error_1.errMsg(e)}`;
            automation_client_1.logger.error(e.message);
            throw e;
        }
        try {
            yield client.read(spec);
        }
        catch (e) {
            automation_client_1.logger.debug(`Failed to read resource ${slug}: ${error_1.errMsg(e)}`);
            automation_client_1.logger.info(`Creating resource ${slug} using '${resource_1.logObject(spec)}'`);
            return retry_1.logRetry(() => client.create(spec), `create resource ${slug}`);
        }
        automation_client_1.logger.info(`Patching resource ${slug} using '${resource_1.logObject(spec)}'`);
        const options = {
            headers: {
                "Content-Type": "application/merge-patch+json",
            },
        };
        return retry_1.logRetry(() => client.patch(spec, options), `patch resource ${slug}`);
    });
}
exports.applySpec = applySpec;
//# sourceMappingURL=apply.js.map