"use strict";
/*
 * Copyright © 2018 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const k8s = require("@kubernetes/client-node");
/**
 * Get Kubernetes configuration.  It first tries to
 * `loadFromDefault()`, which looks in the standard locations for a
 * Kubernetes config file.  If that fails, it attempts to load the
 * in-cluster client credentials.
 */
function loadKubeConfig() {
    const kc = new k8s.KubeConfig();
    try {
        kc.loadFromDefault();
    }
    catch (e) {
        automation_client_1.logger.debug(`Failed to to load default config, attempting in-cluster`);
        kc.loadFromCluster();
    }
    return kc;
}
exports.loadKubeConfig = loadKubeConfig;
/**
 * If the SDM configuration contains a Kubernetes config context,
 * validate it exists in the default Kubernetes config.  If the SDM
 * context is not available in the Kubernetes config, an error is
 * thrown.
 *
 * If there is no context in the SDM configuration, read the current
 * context from the default Kubernetes config and set it in the SDM
 * configuration.
 *
 * If this function is unable to read the default Kubernetes config,
 * it throws an error.
 *
 * @param sdm Running Software Delivery Machine.
 * @return Valid Kubernetes config context.
 */
function kubeConfigContext(sdm) {
    const kc = new k8s.KubeConfig();
    try {
        kc.loadFromDefault();
    }
    catch (e) {
        e.message = `Failed to to load default Kubernetes config: ${e.message}`;
        automation_client_1.logger.error(e.message);
        throw e;
    }
    if (sdm.configuration.sdm.k8s.options.context) {
        if (!kc.contexts.some(c => c.name === sdm.configuration.sdm.k8s.options.context)) {
            const msg = `Kubernetes config context in SDM configuration does not exist in default Kubernetes config`;
            automation_client_1.logger.error(msg);
            automation_client_1.logger.error(`Available Kubernetes config contexts: ${kc.contexts.map(c => c.name).join(",")}`);
            throw new Error(msg);
        }
    }
    else if (kc.currentContext) {
        sdm.configuration.sdm.k8s.options.context = kc.currentContext;
    }
    return sdm.configuration.sdm.k8s.options.context;
}
exports.kubeConfigContext = kubeConfigContext;
//# sourceMappingURL=config.js.map