"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const error_1 = require("../support/error");
const retry_1 = require("../support/retry");
const api_1 = require("./api");
const config_1 = require("./config");
const labels_1 = require("./labels");
const request_1 = require("./request");
const resource_1 = require("./resource");
const spec_1 = require("./spec");
/**
 * Delete a resource if it exists.  If the resource does not exist,
 * do nothing.
 *
 * @param spec Kuberenetes spec of resource to delete
 * @return DeleteResponse if object existed and was deleted, undefined if it did not exist
 */
function deleteSpec(spec) {
    return __awaiter(this, void 0, void 0, function* () {
        const slug = spec_1.specSlug(spec);
        let client;
        try {
            const kc = config_1.loadKubeConfig();
            client = kc.makeApiClient(api_1.K8sObjectApi);
        }
        catch (e) {
            e.message = `Failed to create Kubernetes client: ${error_1.errMsg(e)}`;
            automation_client_1.logger.error(e.message);
            throw e;
        }
        try {
            yield client.read(spec);
        }
        catch (e) {
            automation_client_1.logger.debug(`Kubernetes resource ${slug} does not exist: ${error_1.errMsg(e)}`);
            return undefined;
        }
        automation_client_1.logger.info(`Deleting resource ${slug} using '${resource_1.logObject(spec)}'`);
        return retry_1.logRetry(() => client.delete(spec), `delete resource ${slug}`);
    });
}
exports.deleteSpec = deleteSpec;
/**
 * Delete resources associated with application described by `arg.req`, if
 * any exists.  If no matching resources exist, do nothing.  Return
 * ann array of deleted resources, which may be empty.
 *
 * @param arg Specification of what and how to delete for what application
 * @return Array of deleted resources
 */
function deleteAppResources(arg) {
    return __awaiter(this, void 0, void 0, function* () {
        const slug = request_1.appName(arg.req);
        const selector = labels_1.labelSelector(arg.req);
        const toDelete = [];
        try {
            const limit = 500;
            let continu;
            do {
                let listResp;
                const args = [undefined, undefined, continu, undefined, selector, limit];
                if (arg.namespaced) {
                    listResp = yield arg.lister.call(arg.api, arg.req.ns, ...args);
                }
                else if (arg.namespaced === false) {
                    listResp = yield arg.lister.apply(arg.api, args);
                }
                toDelete.push(...listResp.body.items.map(r => {
                    r.kind = r.kind || arg.kind; // list response does not include kind
                    return r;
                }));
                continu = listResp.body.metadata._continue;
            } while (!!continu);
        }
        catch (e) {
            e.message = `Failed to list ${arg.kind} for ${slug}: ${error_1.errMsg(e)}`;
            automation_client_1.logger.error(e.message);
            throw e;
        }
        const deleted = [];
        const errs = [];
        for (const resource of toDelete) {
            const resourceSlug = arg.namespaced ? `${arg.kind}/${resource.metadata.namespace}/${resource.metadata.name}` :
                `${arg.kind}/${resource.metadata.name}`;
            automation_client_1.logger.info(`Deleting ${resourceSlug} for ${slug}`);
            try {
                const args = [undefined, undefined, undefined, undefined, "Background"];
                if (arg.namespaced) {
                    yield arg.deleter.call(arg.api, resource.metadata.name, resource.metadata.namespace, ...args);
                }
                else if (arg.namespaced === false) {
                    yield arg.deleter.call(arg.api, resource.metadata.name, ...args);
                }
                deleted.push(resource);
            }
            catch (e) {
                e.message = `Failed to delete ${resourceSlug} for ${slug}: ${error_1.errMsg(e)}`;
                errs.push(e);
            }
        }
        if (errs.length > 0) {
            const msg = `Failed to delete ${arg.kind} resources for ${slug}: ${errs.map(e => e.message).join("; ")}`;
            automation_client_1.logger.error(msg);
            throw new Error(msg);
        }
        return deleted;
    });
}
exports.deleteAppResources = deleteAppResources;
//# sourceMappingURL=delete.js.map