"use strict";
/*
 * Copyright © 2020 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const _ = require("lodash");
const error_1 = require("../support/error");
const retry_1 = require("../support/retry");
const labels_1 = require("./labels");
const metadata_1 = require("./metadata");
const patch_1 = require("./patch");
const request_1 = require("./request");
const resource_1 = require("./resource");
/**
 * If `req.port` and `req.path` are truthy, create or patch an ingress
 * for a Kubernetes application.  Any provided `req.ingressSpec` is
 * merged using [[ingressTemplate]] before creating/patching.
 *
 * @param req Kuberenetes resource request
 * @return Kubernetes spec used to create/patch resource
 */
function upsertIngress(req) {
    return __awaiter(this, void 0, void 0, function* () {
        const slug = request_1.appName(req);
        if (!req.port) {
            automation_client_1.logger.debug(`Port not provided, will not create ingress ${slug}`);
            return undefined;
        }
        if (!req.path) {
            automation_client_1.logger.debug(`Path not provided, will not upsert ingress ${slug}`);
            return undefined;
        }
        const spec = yield ingressTemplate(req);
        try {
            yield req.clients.ext.readNamespacedIngress(spec.metadata.name, spec.metadata.namespace);
        }
        catch (e) {
            automation_client_1.logger.debug(`Failed to read ingress ${slug}, creating: ${error_1.errMsg(e)}`);
            automation_client_1.logger.info(`Creating ingress ${slug} using '${resource_1.logObject(spec)}'`);
            yield retry_1.logRetry(() => req.clients.ext.createNamespacedIngress(spec.metadata.namespace, spec), `create ingress ${slug}`);
            return spec;
        }
        automation_client_1.logger.info(`Ingress ${slug} exists, patching using '${resource_1.logObject(spec)}'`);
        yield retry_1.logRetry(() => req.clients.ext.patchNamespacedIngress(spec.metadata.name, spec.metadata.namespace, spec, undefined, undefined, undefined, undefined, patch_1.patchHeaders(req)), `patch ingress ${slug}`);
        return spec;
    });
}
exports.upsertIngress = upsertIngress;
/**
 * Create an ingress HTTP path.
 *
 * @param req ingress request
 * @return ingress HTTP path
 */
function httpIngressPath(req) {
    const httpPath = {
        path: req.path,
        backend: {
            serviceName: req.name,
            servicePort: "http",
        },
    };
    return httpPath;
}
/**
 * Create the ingress for a deployment namespace.  If the
 * request has an `ingressSpec`, it is merged into the spec created
 * by this function using `lodash.merge(default, req.ingressSpec)`.
 *
 * It is possible to override the ingress name using the
 * [[KubernetesApplication.ingressSpec]].  If you do this, make sure
 * you know what you are doing.
 *
 * @param req Kubernestes application
 * @return ingress spec with single rule
 */
function ingressTemplate(req) {
    return __awaiter(this, void 0, void 0, function* () {
        const labels = labels_1.applicationLabels(req);
        const metadata = metadata_1.metadataTemplate({
            name: req.name,
            namespace: req.ns,
            labels,
        });
        const httpPath = httpIngressPath(req);
        const rule = {
            http: {
                paths: [httpPath],
            },
        };
        if (req.host) {
            rule.host = req.host;
        }
        const apiVersion = "extensions/v1beta1";
        const kind = "Ingress";
        const i = {
            apiVersion,
            kind,
            metadata,
            spec: {
                rules: [rule],
            },
        };
        if (req.tlsSecret) {
            i.spec.tls = [
                {
                    secretName: req.tlsSecret,
                },
            ];
            if (req.host) {
                i.spec.tls[0].hosts = [req.host];
            }
        }
        if (req.ingressSpec) {
            _.merge(i, req.ingressSpec, { apiVersion, kind });
            i.metadata.namespace = req.ns;
        }
        return i;
    });
}
exports.ingressTemplate = ingressTemplate;
//# sourceMappingURL=ingress.js.map