"use strict";
/*
 * Copyright © 2020 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const error_1 = require("../support/error");
const retry_1 = require("../support/retry");
const labels_1 = require("./labels");
const metadata_1 = require("./metadata");
const patch_1 = require("./patch");
const resource_1 = require("./resource");
exports.defaultNamespace = "default";
/**
 * Create or update a namespace.
 *
 * @param req Kuberenetes application request
 * @return Kubernetes resource spec used to create/patch the resource
 */
function upsertNamespace(req) {
    return __awaiter(this, void 0, void 0, function* () {
        const slug = req.ns;
        const spec = yield namespaceTemplate(req);
        try {
            yield req.clients.core.readNamespace(spec.metadata.name);
            automation_client_1.logger.debug(`Namespace ${slug} exists`);
        }
        catch (e) {
            automation_client_1.logger.debug(`Failed to get namespace ${slug}, creating: ${error_1.errMsg(e)}`);
            automation_client_1.logger.info(`Creating namespace ${slug} using '${resource_1.logObject(spec)}'`);
            yield retry_1.logRetry(() => req.clients.core.createNamespace(spec), `create namespace ${slug}`);
            return spec;
        }
        automation_client_1.logger.info(`Namespace ${slug} exists, patching using '${resource_1.logObject(spec)}'`);
        try {
            yield retry_1.logRetry(() => req.clients.core.patchNamespace(spec.metadata.name, spec, undefined, undefined, undefined, undefined, patch_1.patchHeaders(req)), `patch namespace ${slug}`);
        }
        catch (e) {
            automation_client_1.logger.warn(`Failed to patch existing namespace ${slug}, ignoring: ${error_1.errMsg(e)}`);
        }
        return spec;
    });
}
exports.upsertNamespace = upsertNamespace;
/**
 * Create namespace resource.
 *
 * @param req Kubernetes application
 * @return Kubernetes namespace resource
 */
function namespaceTemplate(req) {
    return __awaiter(this, void 0, void 0, function* () {
        const allLabels = labels_1.applicationLabels(req);
        const retain = ["atomist.com/workspaceId", "app.kubernetes.io/managed-by"];
        const labels = Object.assign({}, ...Object.keys(allLabels).filter(k => retain.includes(k)).map(k => ({ [k]: allLabels[k] })));
        const metadata = metadata_1.metadataTemplate({ labels, name: req.ns });
        const ns = {
            apiVersion: "v1",
            kind: "Namespace",
            metadata,
        };
        return ns;
    });
}
exports.namespaceTemplate = namespaceTemplate;
//# sourceMappingURL=namespace.js.map