"use strict";
/*
 * Copyright © 2020 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const _ = require("lodash");
const error_1 = require("../support/error");
const retry_1 = require("../support/retry");
const labels_1 = require("./labels");
const metadata_1 = require("./metadata");
const patch_1 = require("./patch");
const request_1 = require("./request");
const resource_1 = require("./resource");
/**
 * Create or patch service account.
 *
 * @param req Kuberenetes application request
 * @return Kubernetes resource spec used to create/patch the resource
 */
function upsertServiceAccount(req) {
    return __awaiter(this, void 0, void 0, function* () {
        const slug = request_1.appName(req);
        const spec = yield serviceAccountTemplate(req);
        try {
            yield req.clients.core.readNamespacedServiceAccount(spec.metadata.name, spec.metadata.namespace);
        }
        catch (e) {
            automation_client_1.logger.debug(`Failed to read service account ${slug}, creating: ${error_1.errMsg(e)}`);
            automation_client_1.logger.info(`Creating service account ${slug} using '${resource_1.logObject(spec)}'`);
            yield retry_1.logRetry(() => req.clients.core.createNamespacedServiceAccount(spec.metadata.namespace, spec), `create service account ${slug}`);
            return spec;
        }
        automation_client_1.logger.info(`Service account ${slug} exists, patching using '${resource_1.logObject(spec)}'`);
        yield retry_1.logRetry(() => req.clients.core.patchNamespacedServiceAccount(spec.metadata.name, spec.metadata.namespace, spec, undefined, undefined, undefined, undefined, patch_1.patchHeaders(req)), `patch service account ${slug}`);
        return spec;
    });
}
exports.upsertServiceAccount = upsertServiceAccount;
/**
 * Create service account spec for a Kubernetes application.  The
 * `req.rbac.serviceAccountSpec`, if it not false, is merged into the
 * spec created by this function using `lodash.merge(default,
 * req.rbac.serviceAccountSpec)`.
 *
 * It is possible to override the service account name using the
 * [[KubernetesApplication.serviceAccountSpec]].  If you do this, make
 * sure you know what you are doing and also override it in the
 * [[KubernetesApplication.roleBindingSpec]].
 *
 * @param req application request
 * @return service account resource specification
 */
function serviceAccountTemplate(req) {
    return __awaiter(this, void 0, void 0, function* () {
        const labels = labels_1.applicationLabels(req);
        const metadata = metadata_1.metadataTemplate({
            name: req.name,
            namespace: req.ns,
            labels,
        });
        const apiVersion = "v1";
        const kind = "ServiceAccount";
        const sa = {
            apiVersion,
            kind,
            metadata,
        };
        if (req.serviceAccountSpec) {
            _.merge(sa, req.serviceAccountSpec, { apiVersion, kind });
            sa.metadata.namespace = req.ns;
        }
        return sa;
    });
}
exports.serviceAccountTemplate = serviceAccountTemplate;
//# sourceMappingURL=serviceAccount.js.map