"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const sdm_core_1 = require("@atomist/sdm-core");
const cluster = require("cluster");
const _ = require("lodash");
const types_1 = require("../typings/types");
/**
 * If the SDM is the cluster master, not running in local mode, and
 * registered with one or more workspaces, ensure that this SDM is
 * available as a KubernetesClusterProvider in those workspaces.
 */
exports.providerStartupListener = (context) => __awaiter(void 0, void 0, void 0, function* () {
    if (sdm_core_1.isInLocalMode() || !cluster.isMaster) {
        return;
    }
    const sdm = context.sdm;
    if (!_.get(sdm, "configuration.sdm.k8s.options.registerCluster", false)) {
        return;
    }
    if (!sdm || !sdm.configuration || !sdm.configuration.workspaceIds) {
        automation_client_1.logger.debug(`SDM configuration contains no workspace IDs, not creating KubernetesClusterProvider`);
    }
    const name = context.sdm.configuration.name;
    const url = _.get(sdm.configuration, "sdm.kubernetes.provider.url");
    yield Promise.all(sdm.configuration.workspaceIds.map((workspaceId) => __awaiter(void 0, void 0, void 0, function* () {
        const graphClient = sdm.configuration.graphql.client.factory.create(workspaceId, sdm.configuration);
        automation_client_1.logger.debug(`Checking for KubernetesClusterProvider ${name} in workspace ${workspaceId}`);
        const providers = yield graphClient.query({
            name: "KubernetesClusterProvider",
            variables: { name },
            options: automation_client_1.QueryNoCacheOptions,
        });
        if (providers && providers.KubernetesClusterProvider && providers.KubernetesClusterProvider.length === 1) {
            automation_client_1.logger.info(`KubernetesClusterProvider ${name} already exists in ${workspaceId}`);
            const provider = providers.KubernetesClusterProvider[0];
            if (url !== provider.url) {
                automation_client_1.logger.info(`Deleting existing KubernetesClusterProvider because url doesn't match`);
                yield graphClient.mutate({
                    name: "DeleteResourceProvider",
                    variables: {
                        id: provider.id,
                    },
                });
            }
            else if (!provider.state || provider.state.name !== types_1.ResourceProviderStateName.converged) {
                yield setProviderState(graphClient, providers.KubernetesClusterProvider[0].id, types_1.ResourceProviderStateName.converged);
                return;
            }
        }
        if (providers && providers.KubernetesClusterProvider && providers.KubernetesClusterProvider.length > 1) {
            automation_client_1.logger.warn(`More than one KubernetesClusterProvider with the name ${name} exists in ${workspaceId}`);
            return;
        }
        automation_client_1.logger.info(`Creating KubernetesClusterProvider ${name} in ${workspaceId}`);
        const result = yield graphClient.mutate({
            name: "CreateKubernetesClusterProvider",
            variables: { name, url },
        });
        yield setProviderState(graphClient, result.createKubernetesClusterProvider.id, types_1.ResourceProviderStateName.converged);
        return;
    })));
    return;
});
/**
 * Set the state of the resource provider identified by provided id.
 */
function setProviderState(gc, id, state, error = "") {
    return __awaiter(this, void 0, void 0, function* () {
        yield gc.mutate({
            name: "SetResourceProviderState",
            variables: {
                id,
                state,
                error,
            },
        });
    });
}
//# sourceMappingURL=kubernetesCluster.js.map