"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const sdm_1 = require("@atomist/sdm");
const sdm_core_1 = require("@atomist/sdm-core");
const stringify = require("json-stringify-safe");
const _ = require("lodash");
const error_1 = require("../support/error");
const change_1 = require("./change");
const diff_1 = require("./diff");
const repo_1 = require("./repo");
const tag_1 = require("./tag");
function isSyncRepoCommit(sdm) {
    const syncOptions = _.get(sdm, "configuration.sdm.k8s.options.sync");
    if (!syncOptions || !syncOptions.repo) {
        automation_client_1.logger.debug(`SDM configuration contains no sync repo, will not create sync repo push test`);
        return undefined;
    }
    return sdm_1.pushTest("IsSyncRepoCommit", (p) => __awaiter(this, void 0, void 0, function* () {
        const repo = _.get(sdm, "configuration.sdm.k8s.options.sync.repo");
        if (!repo_1.isRemoteRepo(repo) && !sdm_core_1.isInLocalMode()) {
            throw new Error(`SyncRepoRef did not get converted to proper RemoteRepoRef at startup: ${stringify(repo)}`);
        }
        if (sdm_core_1.isInLocalMode() || (p.id.providerType === repo.providerType) &&
            p.id.owner === repo.owner &&
            p.id.repo === repo.repo && p.id.branch === repo.branch) {
            const tag = tag_1.commitTag(sdm.configuration);
            return p.push.commits.some(c => !c.message.includes(tag));
        }
        return false;
    }));
}
exports.isSyncRepoCommit = isSyncRepoCommit;
/**
 * Add goals for pushes on the sync repo.
 */
function syncGoals(sdm) {
    const syncRepoPushTest = isSyncRepoCommit(sdm);
    if (!syncRepoPushTest) {
        automation_client_1.logger.debug(`No sync repo push test, not adding sync repo goals`);
        return sdm;
    }
    const sync = new sdm_1.GoalWithFulfillment({
        uniqueName: "sync",
        environment: sdm_1.IndependentOfEnvironment,
        displayName: "sync",
        workingDescription: "Syncing",
        completedDescription: "Synced",
        failedDescription: "Sync failed",
        isolated: true,
    }).with({
        name: "K8sSyncRepo",
        goalExecutor: exports.K8sSync,
        logInterpreter: sdm_1.LogSuppressor,
    });
    const syncGoalSet = sdm_1.goals("Sync Kubernetes Resources").plan(sync);
    sdm.addGoalContributions(sdm_1.whenPushSatisfies(syncRepoPushTest).setGoals(syncGoalSet));
    return sdm;
}
exports.syncGoals = syncGoals;
/**
 * Create resources of added specs, update resources of changed specs,
 * and remove resources of deleted specs.
 */
exports.K8sSync = (gi) => __awaiter(void 0, void 0, void 0, function* () {
    const push = gi.goalEvent.push;
    const log = gi.progressLog;
    const params = {
        cloneOptions: sdm_1.minimalClone(push),
        context: gi.context,
        credentials: gi.credentials,
        id: gi.id,
        log,
        readOnly: true,
    };
    const tag = tag_1.commitTag(gi.configuration);
    return gi.configuration.sdm.projectLoader.doWithProject(params, (p) => __awaiter(void 0, void 0, void 0, function* () {
        if (params.cloneOptions.depth > 1) {
            // work around https://github.com/atomist/sdm/issues/729
            try {
                yield sdm_1.execPromise("git", ["fetch", `--depth=${params.cloneOptions.depth}`], { cwd: p.baseDir });
            }
            catch (e) {
                automation_client_1.logger.warn(`Failed to undo shallow clone, proceeding anyway: ${e.message}`);
            }
        }
        const changes = yield diff_1.diffPush(p, push, tag, log);
        const errs = [];
        for (const change of changes) {
            const verb = (change.change === "delete") ? "Deleting" : "Applying";
            log.write(`${verb} '${change.path}' from commit ${change.sha}`);
            try {
                yield change_1.changeResource(p, change);
            }
            catch (e) {
                e.message = `Failed to ${change.change} '${change.path}' resource for commit ${change.sha}: ${error_1.errMsg(e)}`;
                automation_client_1.logger.error(e.message);
                log.write(e.message);
                errs.push(e);
            }
        }
        if (errs.length > 0) {
            return { code: errs.length, message: errs.map(e => e.message).join("; ") };
        }
        const message = `Changed ${changes.length} resources`;
        log.write(message);
        return { code: 0, message };
    }));
});
//# sourceMappingURL=goals.js.map