"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const RepoId_1 = require("@atomist/automation-client/lib/operations/common/RepoId");
const stringify = require("json-stringify-safe");
const _ = require("lodash");
const clone_1 = require("./clone");
const defaultDefaultBranch = "master";
/**
 * If called and no sync repo is provided in the SDM configuration, a
 * warning is emitted and `undefined` is returned.
 *
 * If the SDM configuration contains a RemoteRepoRef as the value of
 * the `sdm.configuration.sdm.k8s.options.sync.repo` option and truthy
 * credentials, those are returned.
 *
 * Otherwise, the code cycles through all workspaces, querying cortex
 * for the information it needs.  If the value of the `sync.repo`
 * option is a [[SyncRepoRef]], each workspace is queried for a repo
 * matching the provided sync repo.  If none is found, it cycles
 * through all the workspaces again querying for all SCM providers and
 * try to clone the repo with the provided credentials or, if no
 * credentials are provided, the SCM provider credentials from cortex.
 * Once a repo is found using either method, an object is returned
 * with its remote repo ref and the credentials able to clone it.  In
 * addition, the `sdm` passed in will have its
 * `sdm.configuration.sdm.k8s.options.sync.repo`
 * and`sdm.configuration.sdm.k8s.options.sync.credentials` updated
 * with the objects appropriate objects.
 *
 * @param sdm this SDM object (modified if repo credentials found)
 * @param repoRef repository to look for
 * @return true if sync options set and repo found or false and sync options deleted
 */
function queryForScmProvider(sdm) {
    return __awaiter(this, void 0, void 0, function* () {
        const syncOptions = _.get(sdm, "configuration.sdm.k8s.options.sync");
        if (!syncOptions) {
            automation_client_1.logger.debug(`SDM configuration contains no sync repo`);
            return false;
        }
        const repoRef = syncOptions.repo;
        if (!repoRef || !repoRef.owner || !repoRef.repo) {
            automation_client_1.logger.error(`Provided sync repo does not contain all required properties, deleting sync options: ${stringify(repoRef)}`);
            delete sdm.configuration.sdm.k8s.options.sync;
            return false;
        }
        const repoProvided = isRemoteRepo(repoRef);
        const credsProvided = !!syncOptions.credentials;
        if (repoProvided && credsProvided) {
            automation_client_1.logger.info(`Using provided remote repo ref and credentials for sync repo`);
            return true;
        }
        const repoCreds = (yield queryRepo(sdm)) || (yield queryScm(sdm));
        if (repoCreds) {
            if (!repoProvided) {
                sdm.configuration.sdm.k8s.options.sync.repo = repoCreds.repo;
            }
            if (!credsProvided) {
                sdm.configuration.sdm.k8s.options.sync.credentials = repoCreds.credentials;
            }
            return true;
        }
        automation_client_1.logger.warn(`Failed to find sync repo, deleting sync options: ${stringify(repoRef)}`);
        delete sdm.configuration.sdm.k8s.options.sync;
        return false;
    });
}
exports.queryForScmProvider = queryForScmProvider;
/**
 * See if provided sync repo is a RemoteRepoRef.
 */
function isRemoteRepo(repo) {
    return !!repo && RepoId_1.isRemoteRepoRef(repo);
}
exports.isRemoteRepo = isRemoteRepo;
/**
 * Query cortex across all available workspaces for repo.
 */
function queryRepo(sdm) {
    return __awaiter(this, void 0, void 0, function* () {
        const repoRef = syncRepoRef(sdm);
        const slug = repoSlug(repoRef);
        const repoProviderId = repoRef.providerId;
        for (const workspaceId of sdm.configuration.workspaceIds) {
            const graphClient = sdm.configuration.graphql.client.factory.create(workspaceId, sdm.configuration);
            automation_client_1.logger.debug(`Querying workspace ${workspaceId} for repo ${slug}`);
            const repos = yield graphClient.query({
                name: "RepoScmProvider",
                variables: { repo: repoRef.repo, owner: repoRef.owner },
                options: automation_client_1.QueryNoCacheOptions,
            });
            if (!repos || !repos.Repo || repos.Repo.length < 1) {
                automation_client_1.logger.debug(`Repo ${slug} not found in workspace ${workspaceId}`);
                continue;
            }
            let searchRepos = repos.Repo;
            if (repoProviderId) {
                automation_client_1.logger.debug(`Filtering repos from workspace ${workspaceId} on providerId '${repoProviderId}'`);
                searchRepos = searchRepos.filter(r => r.org.scmProvider.providerId === repoProviderId);
            }
            if (searchRepos.length > 1) {
                automation_client_1.logger.warn(`More than one repo found in workspace ${workspaceId} with owner/repo ${slug}`);
            }
            for (const repo of searchRepos) {
                const rc = repoCredentials(sdm, repo);
                if (rc) {
                    rc.repo.branch = rc.repo.branch || repo.defaultBranch || defaultDefaultBranch;
                    automation_client_1.logger.info(`Returning first ${slug} repo with valid SCM provider`);
                    return rc;
                }
            }
        }
        return undefined;
    });
}
/**
 * For each SDM provider in cortex in each workspace, try to clone the
 * sync repo.  Return the information for the first successful clone.
 */
function queryScm(sdm) {
    return __awaiter(this, void 0, void 0, function* () {
        const repoRef = syncRepoRef(sdm);
        const slug = repoSlug(repoRef);
        const repoProviderId = repoRef.providerId;
        for (const workspaceId of sdm.configuration.workspaceIds) {
            const graphClient = sdm.configuration.graphql.client.factory.create(workspaceId, sdm.configuration);
            automation_client_1.logger.debug(`Querying workspace ${workspaceId} for SCM providers`);
            const providers = yield graphClient.query({
                name: "ScmProviders",
                options: automation_client_1.QueryNoCacheOptions,
            });
            if (!providers || !providers.SCMProvider || providers.SCMProvider.length < 1) {
                automation_client_1.logger.debug(`Found no SCM providers in workspace ${workspaceId}`);
                continue;
            }
            for (const provider of providers.SCMProvider) {
                if (repoProviderId && provider.providerId !== repoProviderId) {
                    automation_client_1.logger.debug(`SCM provider '${provider.providerId}' does not match '${repoProviderId}'`);
                    continue;
                }
                const rc = scmCredentials(sdm, provider);
                if (rc) {
                    automation_client_1.logger.debug(`Attempting to clone ${slug} using ${rc.repo.cloneUrl}`);
                    try {
                        const p = yield automation_client_1.GitCommandGitProject.cloned(rc.credentials, rc.repo, clone_1.defaultCloneOptions);
                        if (p) {
                            rc.repo.branch = rc.repo.branch || p.branch || defaultDefaultBranch;
                            return rc;
                        }
                    }
                    catch (e) {
                        automation_client_1.logger.debug(`Failed to clone ${slug} from ${rc.repo.cloneUrl}: ${e.message}`);
                    }
                }
            }
        }
        return undefined;
    });
}
/**
 * Create RemoteRepoRef and Credentials object from SDM and repo from
 * cortex.  If the provided repo does not contain an org with a
 * provider, it returns `undefined`.  Otherwise it uses the SCM
 * provider to call [[scmCredentials]] and return its value.
 */
function repoCredentials(sdm, repo) {
    if (repo.org && repo.org.scmProvider) {
        return scmCredentials(sdm, repo.org.scmProvider);
    }
    return undefined;
}
exports.repoCredentials = repoCredentials;
/**
 * Given the SDM and an SCM, use the configured repo ref resolver to
 * create a RemoteRepoRef.  Use the SDM Kubernetes option sync
 * credentials or SCM `credential.secret` to create the credentials,
 * giving the SDM sync credentials preference.  Return `undefined` if
 * there is not enough information to created the repo credential
 * object.
 */
function scmCredentials(sdm, scm) {
    const repoRef = syncRepoRef(sdm);
    const credentials = _.get(sdm, "configuration.sdm.k8s.options.sync.credentials");
    const secret = _.get(scm, "credential.secret");
    if (repoRef && repoRef.owner && repoRef.repo && scm.apiUrl && (credentials || secret)) {
        const repoResolver = sdm.configuration.sdm.repoRefResolver;
        const repoFrag = {
            owner: repoRef.owner,
            name: repoRef.repo,
            org: {
                owner: repoRef.owner,
                provider: {
                    providerId: scm.providerId,
                    providerType: scm.providerType,
                    apiUrl: scm.apiUrl,
                    url: scm.url,
                },
            },
        };
        const options = {
            branch: repoRef.branch,
        };
        try {
            const repo = repoResolver.toRemoteRepoRef(repoFrag, options);
            return {
                credentials: credentials || { token: secret },
                repo,
            };
        }
        catch (e) {
            automation_client_1.logger.warn(`Failed to resolve remote repo ref for ${repoFrag.owner}/${repoFrag.name}: ${e.message}`);
        }
    }
    return undefined;
}
exports.scmCredentials = scmCredentials;
/** Create repo slug string. */
function repoSlug(repo) {
    return `${repo.owner}/${repo.repo}`;
}
exports.repoSlug = repoSlug;
/**
 * Extract the Kubernetes option sync repo from the SDM configuration.
 * This function should only be called if the sync repo object is
 * defined.
 */
function syncRepoRef(sdm) {
    const repo = _.get(sdm, "configuration.sdm.k8s.options.sync.repo");
    if (!repo) {
        throw new Error(`Failed to get sync repo from SDM configuration`);
    }
    return repo;
}
//# sourceMappingURL=repo.js.map