"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const sdm_1 = require("@atomist/sdm");
const _ = require("lodash");
const apply_1 = require("../kubernetes/apply");
const secret_1 = require("../kubernetes/secret");
const spec_1 = require("../kubernetes/spec");
const error_1 = require("../support/error");
const name_1 = require("../support/name");
const clone_1 = require("./clone");
const diff_1 = require("./diff");
const repo_1 = require("./repo");
exports.KubernetesSync = "KubernetesSync";
/**
 * Command to synchronize the resources in a Kubernetes cluster with
 * the resource specs in the configured sync repo.  The sync repo will
 * be cloned and the resources applied against the Kubernetes API in
 * lexical order sorted by file name.  If no sync repo is configured
 * in the SDM, the command errors.  This command is typically executed
 * on an interval timer by setting the `intervalMinutes`
 * [[KubernetesSyncOptions]].
 *
 * If the SDM configuration says this packs commands should be added,
 * i.e., `sdm.configuration.sdm.k8s.options.addCommands` is `true`,
 * the command will have the intent `kube sync SDM_NAME`.  Otherwise
 * the command will be registered without an intent.
 */
function kubernetesSync(sdm) {
    const cmd = {
        name: exports.KubernetesSync,
        listener: repoSync,
    };
    if (_.get(sdm, "configuration.sdm.k8s.options.addCommands", false)) {
        cmd.intent = `kube sync ${name_1.cleanName(sdm.configuration.name)}`;
    }
    return cmd;
}
exports.kubernetesSync = kubernetesSync;
/**
 * Clone the sync repo and apply the specs to the Kubernetes cluster.
 */
function repoSync(cli) {
    return __awaiter(this, void 0, void 0, function* () {
        const opts = _.get(cli.configuration, "sdm.k8s.options.sync");
        if (!opts) {
            const message = `SDM has no sync options defined`;
            automation_client_1.logger.error(message);
            yield cli.context.messageClient.respond(message);
            return { code: 2, message };
        }
        if (!repo_1.isRemoteRepo(opts.repo)) {
            const message = `SDM sync option repo is not a valid remote repo`;
            automation_client_1.logger.error(message);
            yield cli.context.messageClient.respond(message);
            return { code: 2, message };
        }
        const projectLoadingParameters = {
            credentials: opts.credentials,
            cloneOptions: clone_1.defaultCloneOptions,
            context: cli.context,
            id: opts.repo,
            readOnly: true,
        };
        const slug = `${opts.repo.owner}/${opts.repo.repo}`;
        try {
            automation_client_1.logger.info(`Starting sync of repo ${slug}`);
            yield cli.configuration.sdm.projectLoader.doWithProject(projectLoadingParameters, syncApply(opts));
        }
        catch (e) {
            const message = `Failed to sync repo ${slug}: ${e.message}`;
            automation_client_1.logger.error(message);
            yield cli.context.messageClient.respond(message);
            return { code: 1, message };
        }
        const result = {
            code: 0,
            message: `Successfully completed sync of repo ${slug}`,
        };
        automation_client_1.logger.info(result.message);
        try {
            yield cli.context.messageClient.respond(sdm_1.slackSuccessMessage("Kubernetes Sync", result.message));
        }
        catch (e) {
            result.code++;
            result.message = `${result.message}; Failed to send response message: ${e.message}`;
            automation_client_1.logger.error(result.message);
        }
        return result;
    });
}
exports.repoSync = repoSync;
/**
 * Return a function that ensures all specs in `syncRepo` have
 * corresponding resources in the Kubernetes cluster.  If the resource
 * does not exist, it is created using the spec.  If it does exist, it
 * is patched using the spec.  Errors are collected and thrown after
 * processing all specs so one bad spec does not stop processing.
 *
 * @param opts Kubernetes sync options
 */
function syncApply(opts) {
    return (syncRepo) => __awaiter(this, void 0, void 0, function* () {
        const specFiles = yield sortSpecs(syncRepo);
        const errors = [];
        for (const specFile of specFiles) {
            automation_client_1.logger.debug(`Processing spec ${specFile.path}`);
            try {
                const specString = yield specFile.getContent();
                const specs = spec_1.parseKubernetesSpecs(specString);
                for (let spec of specs) {
                    try {
                        if (spec.kind === "Secret" && opts && opts.secretKey) {
                            spec = yield secret_1.decryptSecret(spec, opts.secretKey);
                        }
                        yield apply_1.applySpec(spec);
                    }
                    catch (e) {
                        const slug = spec_1.specSlug(spec);
                        e.message = `Failed to apply spec '${slug}' from '${specFile.path}': ${error_1.errMsg(e)}`;
                        automation_client_1.logger.error(e.message);
                        errors.push(e);
                    }
                }
            }
            catch (e) {
                e.message = `Failed to apply '${specFile.path}': ${error_1.errMsg(e)}`;
                automation_client_1.logger.error(e.message);
                errors.push(e);
            }
        }
        if (errors.length > 0) {
            errors[0].message = `There were errors during repo sync: ${errors.map(e => e.message).join("; ")}`;
            throw errors[0];
        }
        return;
    });
}
/**
 * Consume stream of files from project and sort them by their `path`
 * property using `localeCompare`.  Any file at the root of the
 * project, i.e., not in a subdirectory, having the extensions
 * ".json", ".yaml", or ".yml` are considered specs.
 *
 * Essentially, this function converts a FileStream into a Promise of
 * sorted ProjectFiles.
 *
 * @param syncRepo Repository of specs to sort
 * @return Sorted array of specs in project
 */
function sortSpecs(syncRepo) {
    return new Promise((resolve, reject) => {
        const specsStream = syncRepo.streamFiles(diff_1.k8sSpecGlob);
        const specs = [];
        specsStream.on("data", f => specs.push(f));
        specsStream.on("error", reject);
        specsStream.on("end", () => resolve(specs.sort((a, b) => a.path.localeCompare(b.path))));
    });
}
exports.sortSpecs = sortSpecs;
//# sourceMappingURL=sync.js.map