"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const projectEditor_1 = require("@atomist/automation-client/lib/operations/edit/projectEditor");
const slack_messages_1 = require("@atomist/slack-messages");
const handlerRegistrations_1 = require("../../machine/handlerRegistrations");
const child_process_1 = require("../../misc/child_process");
const messages_1 = require("../../misc/slack/messages");
const confirmEditedness_1 = require("./confirmEditedness");
/**
 * Wrap this editor to make it chatty, so it responds to Slack if there's nothing to do.
 * It also honors the dryRun parameter flag to just capture the git diff and send it back to Slack instead
 * of pushing changes to Git.
 */
function chattyDryRunAwareEditor(ctr, underlyingEditor) {
    return (project, context, params) => __awaiter(this, void 0, void 0, function* () {
        const id = project.id;
        const editorName = ctr.name;
        try {
            yield sendDryRunUpdateMessage(editorName, id, params, context, ctr);
            const tentativeEditResult = yield projectEditor_1.toEditor(underlyingEditor)(project, context, params);
            const editResult = yield confirmEditedness_1.confirmEditedness(tentativeEditResult);
            // Figure out if this CodeTransform is running in dryRun mode; if so capture git diff and don't push changes
            if (!editResult.edited) {
                if (!editResult.success) {
                    yield sendFailureMessage(editorName, id, params, editResult, context, ctr);
                }
                else {
                    yield sendNoUpdateMessage(editorName, id, params, context, ctr);
                }
                return { target: project, edited: false, success: false };
            }
            else if (isDryRun(params)) {
                if (!automation_client_1.isLocalProject(project)) {
                    const message = `Project is not a local project, cannot diff`;
                    automation_client_1.logger.warn(message);
                    return { target: project, edited: false, success: true };
                }
                let diff = "";
                try {
                    const gitDiffResult = yield child_process_1.execPromise("git", ["diff"], { cwd: project.baseDir });
                    diff = gitDiffResult.stdout;
                }
                catch (err) {
                    automation_client_1.logger.error(`Error diffing project: %s`, err.message);
                    diff = `Error obtaining \`git diff\`:\n\n${slack_messages_1.codeBlock(err.message)}`;
                }
                yield sendDryRunSummaryMessage(editorName, id, diff, params, context, ctr);
                return { target: project, edited: false, success: true };
            }
            else {
                yield sendSuccessMessage(editorName, id, params, context, ctr);
            }
            return editResult;
        }
        catch (err) {
            yield context.messageClient.respond(messages_1.slackErrorMessage(`Code Transform${isDryRun(params) ? " (dry run)" : ""}`, `Code transform ${slack_messages_1.italic(editorName)} failed while changing ${slack_messages_1.bold(slug(id))}:\n\n${slack_messages_1.codeBlock(err.message)}`, context), { id: params[handlerRegistrations_1.MsgIdParameter.name] });
            automation_client_1.logger.warn("Code Transform error acting on %j: %s", project.id, err);
            return { target: project, edited: false, success: false };
        }
    });
}
exports.chattyDryRunAwareEditor = chattyDryRunAwareEditor;
/**
 * @deprecated use chattyDryRunAwareEditor
 */
exports.chattyEditor = chattyDryRunAwareEditor;
function isDryRun(params) {
    return !!params && params[handlerRegistrations_1.DryRunParameter.name] === true;
}
function slug(id) {
    return `${id.owner}/${id.repo}/${id.branch}`;
}
function isChatty(ctr) {
    if (ctr.chatty !== undefined) {
        return ctr.chatty;
    }
    else {
        return true;
    }
}
function sendDryRunUpdateMessage(codeTransformName, id, params, ctx, ctr) {
    return __awaiter(this, void 0, void 0, function* () {
        if (isChatty(ctr) && !!params[handlerRegistrations_1.MsgIdParameter.name]) {
            yield ctx.messageClient.respond(messages_1.slackInfoMessage("Code Transform", `Applying code transform ${slack_messages_1.italic(codeTransformName)} to ${slack_messages_1.bold(slug(id))}`), { id: params[handlerRegistrations_1.MsgIdParameter.name] });
        }
    });
}
function sendFailureMessage(codeTransformName, id, params, editResult, ctx, ctr) {
    return __awaiter(this, void 0, void 0, function* () {
        if (isChatty(ctr)) {
            yield ctx.messageClient.respond(messages_1.slackErrorMessage(`Code Transform${isDryRun(params) ? " (dry run)" : ""}`, `Code transform ${slack_messages_1.italic(codeTransformName)} failed while changing ${slack_messages_1.bold(slug(id))}:\n\n${editResult.error ? slack_messages_1.codeBlock(editResult.error.message) : ""}`, ctx), { id: params[handlerRegistrations_1.MsgIdParameter.name] });
        }
    });
}
function sendNoUpdateMessage(codeTransformName, id, params, ctx, ctr) {
    return __awaiter(this, void 0, void 0, function* () {
        if (isChatty(ctr)) {
            yield ctx.messageClient.respond(messages_1.slackInfoMessage(`Code Transform${isDryRun(params) ? " (dry run)" : ""}`, `Code transform ${slack_messages_1.italic(codeTransformName)} made no changes to ${slack_messages_1.bold(slug(id))}`), { id: params[handlerRegistrations_1.MsgIdParameter.name] });
        }
    });
}
function sendSuccessMessage(codeTransformName, id, params, ctx, ctr) {
    return __awaiter(this, void 0, void 0, function* () {
        if (isChatty(ctr)) {
            const msgId = params[handlerRegistrations_1.MsgIdParameter.name];
            yield ctx.messageClient.respond(messages_1.slackSuccessMessage("Code Transform", `Successfully applied code transform ${slack_messages_1.italic(codeTransformName)} to ${slack_messages_1.bold(slug(id))}`), { id: msgId });
        }
    });
}
function sendDryRunSummaryMessage(codeTransformName, id, diff, params, ctx, ctr) {
    return __awaiter(this, void 0, void 0, function* () {
        const msgId = params[handlerRegistrations_1.MsgIdParameter.name] || automation_client_1.guid();
        const applyAction = {
            actions: [
                automation_client_1.buttonForCommand({ text: "Apply Transform" }, codeTransformName, Object.assign(Object.assign({}, params), { "dry-run": false, "msgId": msgId, "targets.sha": params.targets.sha, "targets.owner": id.owner, "targets.repo": id.repo })),
            ],
        };
        yield ctx.messageClient.respond(messages_1.slackInfoMessage(`Code Transform (dry run)`, `Code transform ${slack_messages_1.italic(codeTransformName)} would make the following changes to ${slack_messages_1.bold(slug(id))}:
${slack_messages_1.codeBlock(diff)}
`, applyAction), { id: msgId });
    });
}
//# sourceMappingURL=chattyDryRunAwareEditor.js.map