"use strict";
/*
 * Copyright © 2020 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const configuration_1 = require("@atomist/automation-client/lib/configuration");
const _ = require("lodash");
const addressChannels_1 = require("../../api/context/addressChannels");
const preferenceStore_1 = require("../../api/context/preferenceStore");
const Goal_1 = require("../../api/goal/Goal");
const Goals_1 = require("../../api/goal/Goals");
const GoalWithFulfillment_1 = require("../../api/goal/GoalWithFulfillment");
const SdmGoalMessage_1 = require("../../api/goal/SdmGoalMessage");
const GoalImplementationMapper_1 = require("../../api/goal/support/GoalImplementationMapper");
const types_1 = require("../../typings/types");
const minimalClone_1 = require("./minimalClone");
const storeGoals_1 = require("./storeGoals");
/**
 * Choose and set goals for this push
 * @param {ChooseAndSetGoalsRules} rules: configuration for handling incoming pushes
 * @param parameters details of incoming request
 * @return {Promise<Goals | undefined>}
 */
function chooseAndSetGoals(rules, parameters) {
    return __awaiter(this, void 0, void 0, function* () {
        const { projectLoader, goalsListeners, goalSetter, implementationMapping, repoRefResolver, preferencesFactory } = rules;
        const { context, credentials, push } = parameters;
        const enrichGoal = !!rules.enrichGoal ? rules.enrichGoal : (g) => __awaiter(this, void 0, void 0, function* () { return g; });
        const tagGoalSet = !!rules.tagGoalSet ? rules.tagGoalSet : () => __awaiter(this, void 0, void 0, function* () { return []; });
        const id = repoRefResolver.repoRefFromPush(push);
        const addressChannels = addressChannels_1.addressChannelsFor(push.repo, context);
        const preferences = !!preferencesFactory ? preferencesFactory(parameters.context) : preferenceStore_1.NoPreferenceStore;
        const configuration = context.configuration;
        const goalSetId = automation_client_1.guid();
        const { determinedGoals, goalsToSave, tags } = yield determineGoals({ projectLoader, repoRefResolver, goalSetter, implementationMapping, enrichGoal, tagGoalSet }, {
            credentials, id, context, push, addressChannels, preferences, goalSetId, configuration,
        });
        if (goalsToSave.length > 0) {
            // First store the goals
            yield Promise.all(goalsToSave.map(g => storeGoals_1.storeGoal(context, g)));
            // And then store the goalSet
            yield storeGoals_1.storeGoalSet(context, storeGoals_1.constructGoalSet(context, goalSetId, determinedGoals.name, goalsToSave, tags, push));
        }
        // Let GoalSetListeners know even if we determined no goals.
        // This is not an error
        const gsi = {
            id,
            context,
            credentials,
            addressChannels,
            configuration,
            preferences,
            goalSetId,
            goalSetName: determinedGoals ? determinedGoals.name : undefined,
            goalSet: determinedGoals,
            push,
        };
        yield Promise.all(goalsListeners.map(l => l(gsi)));
        return determinedGoals;
    });
}
exports.chooseAndSetGoals = chooseAndSetGoals;
function determineGoals(rules, circumstances) {
    return __awaiter(this, void 0, void 0, function* () {
        const { enrichGoal, projectLoader, repoRefResolver, goalSetter, implementationMapping, tagGoalSet } = rules;
        const { credentials, id, context, push, addressChannels, goalSetId, preferences, configuration } = circumstances;
        return projectLoader.doWithProject({
            credentials,
            id,
            context,
            readOnly: true,
            cloneOptions: minimalClone_1.minimalClone(push, { detachHead: true }),
        }, (project) => __awaiter(this, void 0, void 0, function* () {
            const pli = {
                project,
                credentials,
                id,
                push,
                context,
                addressChannels,
                configuration,
                preferences: preferences || preferenceStore_1.NoPreferenceStore,
                facts: {},
            };
            const determinedGoals = yield chooseGoalsForPushOnProject({ goalSetter }, pli);
            if (!determinedGoals) {
                return { determinedGoals: undefined, goalsToSave: [], tags: [] };
            }
            const goalsToSave = yield sdmGoalsFromGoals(implementationMapping, push, repoRefResolver, pli, determinedGoals, goalSetId);
            // Enrich all goals before they get saved
            yield Promise.all(goalsToSave.map((g1) => __awaiter(this, void 0, void 0, function* () { return enrichGoal(g1, pli); })));
            // Optain tags for the goal set
            let tags = [];
            if (!!tagGoalSet) {
                tags = (yield tagGoalSet(goalsToSave, pli)) || [];
            }
            return { determinedGoals, goalsToSave, tags };
        }));
    });
}
exports.determineGoals = determineGoals;
function sdmGoalsFromGoals(implementationMapping, push, repoRefResolver, pli, determinedGoals, goalSetId) {
    return __awaiter(this, void 0, void 0, function* () {
        return Promise.all(determinedGoals.goals.map((g) => __awaiter(this, void 0, void 0, function* () {
            const ge = storeGoals_1.constructSdmGoal(pli.context, {
                goalSet: determinedGoals.name,
                goalSetId,
                goal: g,
                state: (Goal_1.hasPreconditions(g) ? types_1.SdmGoalState.planned :
                    (g.definition.preApprovalRequired ? types_1.SdmGoalState.waiting_for_pre_approval : types_1.SdmGoalState.requested)),
                id: pli.id,
                providerId: repoRefResolver.providerIdFromPush(pli.push),
                fulfillment: yield fulfillment({ implementationMapping }, g, pli),
            });
            if (ge.state === types_1.SdmGoalState.requested) {
                const cbs = implementationMapping.findFulfillmentCallbackForGoal(Object.assign(Object.assign({}, ge), { push })) || [];
                let ng = Object.assign(Object.assign({}, ge), { push });
                for (const cb of cbs) {
                    ng = yield cb.callback(ng, pli);
                }
                return Object.assign(Object.assign({}, ge), { data: ng.data });
            }
            else {
                return ge;
            }
        })));
    });
}
function fulfillment(rules, g, inv) {
    return __awaiter(this, void 0, void 0, function* () {
        const { implementationMapping } = rules;
        const plan = yield implementationMapping.findFulfillmentByPush(g, inv);
        if (GoalImplementationMapper_1.isGoalImplementation(plan)) {
            return storeGoals_1.constructSdmGoalImplementation(plan, inv.configuration.name);
        }
        else if (GoalImplementationMapper_1.isGoalSideEffect(plan)) {
            return {
                method: SdmGoalMessage_1.SdmGoalFulfillmentMethod.SideEffect,
                name: plan.sideEffectName,
                registration: plan.registration || configuration_1.configurationValue("name"),
            };
        }
        else {
            return { method: SdmGoalMessage_1.SdmGoalFulfillmentMethod.Other, name: "unknown", registration: "unknown" };
        }
    });
}
function chooseGoalsForPushOnProject(rules, pi) {
    return __awaiter(this, void 0, void 0, function* () {
        const { goalSetter } = rules;
        const { push, id } = pi;
        try {
            const determinedGoals = yield goalSetter.mapping(pi);
            if (!determinedGoals) {
                automation_client_1.logger.info("No goals set by push '%s' to '%s/%s/%s'", push.after.sha, id.owner, id.repo, push.branch);
                return determinedGoals;
            }
            else {
                const filteredGoals = [];
                const plannedGoals = yield planGoals(determinedGoals, pi);
                plannedGoals.goals.forEach(g => {
                    if (g.dependsOn) {
                        const preConditions = g.dependsOn;
                        if (preConditions) {
                            const filteredPreConditions = preConditions.filter(pc => plannedGoals.goals.some(ag => ag.uniqueName === pc.uniqueName &&
                                ag.environment === pc.environment));
                            if (filteredPreConditions.length > 0) {
                                filteredGoals.push(new Goal_1.GoalWithPrecondition(g.definition, ...filteredPreConditions));
                            }
                            else {
                                filteredGoals.push(new Goal_1.Goal(g.definition));
                            }
                        }
                        else {
                            filteredGoals.push(g);
                        }
                    }
                    else {
                        filteredGoals.push(g);
                    }
                });
                automation_client_1.logger.info("Goals for push '%s' on '%s/%s/%s' are '%s'", push.after.sha, id.owner, id.repo, push.branch, plannedGoals.name);
                return new Goals_1.Goals(plannedGoals.name, ...filteredGoals);
            }
        }
        catch (err) {
            automation_client_1.logger.error("Error determining goals: %s", err);
            automation_client_1.logger.error(err.stack);
            throw err;
        }
    });
}
function planGoals(goals, pli) {
    return __awaiter(this, void 0, void 0, function* () {
        const allGoals = [...goals.goals];
        const names = [];
        for (const dg of goals.goals) {
            if (!!dg.plan) {
                let planResult = yield dg.plan(pli, goals);
                if (!!planResult) {
                    // Check if planResult is a PlannedGoal or PlannedGoals instance
                    if (!_.some(planResult, v => !!v.goals)) {
                        planResult = { "#": { goals: planResult } };
                    }
                    const allNewGoals = [];
                    const goalMapping = new Map();
                    _.forEach(planResult, (planResultGoals, n) => {
                        names.push(n.replace(/_/g, " "));
                        const plannedGoals = [];
                        if (Array.isArray(planResultGoals.goals)) {
                            plannedGoals.push(...planResultGoals.goals);
                        }
                        else {
                            plannedGoals.push(planResultGoals.goals);
                        }
                        let previousGoals = [];
                        const newGoals = [];
                        plannedGoals.forEach(g => {
                            if (Array.isArray(g)) {
                                const gNewGoals = [];
                                for (const gg of g) {
                                    const newGoal = createGoal(gg, dg, planResultGoals.dependsOn, allNewGoals.length + gNewGoals.length, previousGoals, goalMapping);
                                    gNewGoals.push(newGoal);
                                }
                                allNewGoals.push(...gNewGoals);
                                newGoals.push(...gNewGoals);
                                previousGoals = [...gNewGoals];
                            }
                            else {
                                const newGoal = createGoal(g, dg, planResultGoals.dependsOn, allNewGoals.length, previousGoals, goalMapping);
                                allNewGoals.push(newGoal);
                                newGoals.push(newGoal);
                                previousGoals = [newGoal];
                            }
                        });
                        goalMapping.set(n, newGoals);
                    });
                    // Replace existing goal with new instances
                    const ix = allGoals.findIndex(g => g.uniqueName === dg.uniqueName);
                    allGoals.splice(ix, 1, ...allNewGoals);
                    // Replace all preConditions that point back to the original goal with references to new goals
                    allGoals.filter(Goal_1.hasPreconditions)
                        .filter(g => (g.dependsOn || []).some(gr => gr.uniqueName === dg.uniqueName))
                        .forEach(g => {
                        _.remove(g.dependsOn, gr => gr.uniqueName === dg.uniqueName);
                        g.dependsOn.push(...allNewGoals);
                    });
                }
            }
        }
        return new Goals_1.Goals(goals.name, ...allGoals);
    });
}
exports.planGoals = planGoals;
function createGoal(g, dg, preConditions, plannedGoalsCounter, previousGoals, goalMapping) {
    const uniqueName = `${dg.uniqueName}#sdm:${plannedGoalsCounter}`;
    const definition = _.merge({}, dg.definition, GoalWithFulfillment_1.getGoalDefinitionFrom(g.details, uniqueName));
    definition.uniqueName = uniqueName;
    definition.parameters = g.parameters;
    const dependsOn = [];
    if (Goal_1.hasPreconditions(dg)) {
        dependsOn.push(...dg.dependsOn);
    }
    if (!!previousGoals) {
        dependsOn.push(...previousGoals);
    }
    if (!!preConditions) {
        if (Array.isArray(preConditions)) {
            dependsOn.push(..._.flatten(preConditions.map(d => goalMapping.get(d)).filter(d => !!d)));
        }
        else {
            dependsOn.push(...goalMapping.get(preConditions));
        }
    }
    return new Goal_1.GoalWithPrecondition(definition, ..._.uniqBy(dependsOn.filter(d => !!d), "uniqueName"));
}
//# sourceMappingURL=chooseAndSetGoals.js.map