"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const stringify = require("json-stringify-safe");
const _ = require("lodash");
const sdmGoal_1 = require("./sdmGoal");
const storeGoals_1 = require("./storeGoals");
function fetchGoalsFromPush(sdmGoal) {
    if (sdmGoal.push && sdmGoal.push.goals) {
        const goals = _.cloneDeep(sumSdmGoalEvents(sdmGoal.push.goals.filter(g => g.goalSetId === sdmGoal.goalSetId)));
        const push = _.cloneDeep(sdmGoal.push);
        delete push.goals;
        goals.forEach(g => g.push = push);
        return goals;
    }
    return [];
}
exports.fetchGoalsFromPush = fetchGoalsFromPush;
function findSdmGoalOnCommit(ctx, id, providerId, goal) {
    return __awaiter(this, void 0, void 0, function* () {
        const sdmGoals = yield fetchGoalsForCommit(ctx, id, providerId);
        const matches = sdmGoals.filter(g => storeGoals_1.goalCorrespondsToSdmGoal(goal, g));
        if (matches && matches.length > 1) {
            automation_client_1.logger.warn("More than one match found for %s/%s; they are %j", goal.environment, goal.name, matches);
        }
        if (matches.length === 0) {
            automation_client_1.logger.debug("Did not find goal %s on commit %s#%s", sdmGoal_1.goalKeyString(goal), id.repo, id.sha);
            return undefined;
        }
        return matches[0];
    });
}
exports.findSdmGoalOnCommit = findSdmGoalOnCommit;
function fetchCommitForSdmGoal(ctx, goal) {
    return __awaiter(this, void 0, void 0, function* () {
        const variables = { sha: goal.sha, repo: goal.repo.name, owner: goal.repo.owner, branch: goal.branch };
        const result = yield ctx.graphClient.query({
            name: "CommitForSdmGoal",
            variables: { sha: goal.sha, repo: goal.repo.name, owner: goal.repo.owner, branch: goal.branch },
            options: Object.assign(Object.assign({}, automation_client_1.QueryNoCacheOptions), { log: automation_client_1.configurationValue("sdm.query.logging", false) }),
        });
        if (!result || !result.Commit || result.Commit.length === 0) {
            throw new Error("No commit found for goal " + stringify(variables));
        }
        return result.Commit[0];
    });
}
exports.fetchCommitForSdmGoal = fetchCommitForSdmGoal;
function fetchGoalsForCommit(ctx, id, providerId, goalSetId) {
    return __awaiter(this, void 0, void 0, function* () {
        const result = [];
        const size = 200;
        let offset = 0;
        const query = sdmGoalOffsetQuery(id, goalSetId, providerId, ctx);
        let pageResult = yield query(offset, size);
        while (pageResult && pageResult.SdmGoal && pageResult.SdmGoal.length > 0) {
            result.push(...pageResult.SdmGoal);
            offset += size;
            pageResult = yield query(offset, size);
        }
        if (!result) {
            throw new Error(`No result finding goals for commit ${providerId}/${id.owner}/${id.repo}#${id.sha} on ${id.branch}`);
        }
        if (result.length === 0) {
            automation_client_1.logger.warn("0 goals found for commit %j, provider %s", id, providerId);
        }
        if (result.some(g => !g)) {
            automation_client_1.logger.warn("Null or undefined goal found for commit %j, provider %s", id, providerId);
        }
        // only maintain latest version of SdmGoals from the current goal set
        const goals = sumSdmGoalEvents(result);
        // query for the push and add it in
        if (goals.length > 0) {
            const push = yield ctx.graphClient.query({
                name: "PushForSdmGoal",
                variables: {
                    owner: id.owner,
                    repo: id.repo,
                    providerId,
                    branch: goals[0].branch,
                    sha: goals[0].sha,
                },
                options: {
                    log: automation_client_1.configurationValue("sdm.query.logging", false),
                },
            });
            return goals.map(g => {
                const goal = _.cloneDeep(g);
                goal.push = push.Commit[0].pushes[0];
                return goal;
            });
        }
        return goals;
    });
}
exports.fetchGoalsForCommit = fetchGoalsForCommit;
function sumSdmGoalEvents(some) {
    // For some reason this won't compile with the obvious fix
    // tslint:disable-next-line:no-unnecessary-callback-wrapper
    const byKey = _.groupBy(some, sg => `${sg.goalSetId}-${sdmGoal_1.goalKeyString(sg)}`);
    const summedGoals = Object.keys(byKey).map(k => sumEventsForOneSdmGoal(byKey[k]));
    return summedGoals;
}
exports.sumSdmGoalEvents = sumSdmGoalEvents;
function sumEventsForOneSdmGoal(events) {
    if (events.length === 1) {
        return events[0];
    }
    // SUCCESS OVERRIDES ALL
    const success = events.find(e => e.state === "success");
    return success || _.maxBy(events, e => e.ts);
}
function sdmGoalOffsetQuery(id, goalSetId, providerId, ctx) {
    return (offset, size) => __awaiter(this, void 0, void 0, function* () {
        return ctx.graphClient.query({
            name: "SdmGoalsForCommit",
            variables: {
                owner: id.owner,
                repo: id.repo,
                branch: id.branch,
                sha: id.sha,
                providerId,
                goalSetId,
                qty: size,
                offset,
            },
            options: Object.assign(Object.assign({}, automation_client_1.QueryNoCacheOptions), { log: automation_client_1.configurationValue("sdm.query.logging", false) }),
        });
    });
}
//# sourceMappingURL=fetchGoalsOnCommit.js.map