"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const GoalInvocation_1 = require("../../api/goal/GoalInvocation");
const commonPushTests_1 = require("../../api/mapping/support/commonPushTests");
const storeGoals_1 = require("../goal/storeGoals");
const result_1 = require("../misc/result");
/**
 * ProjectLoader implementation that invokes GoalProjectListener instances on the loaded project.
 * Can be used to restore state into a project/workspace area.
 */
class ProjectListenerInvokingProjectLoader {
    constructor(gi, listeners) {
        this.gi = gi;
        this.listeners = listeners;
    }
    doWithProject(params, action) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.gi.configuration.sdm.projectLoader.doWithProject(params, (p) => __awaiter(this, void 0, void 0, function* () {
                let result;
                try {
                    // invoke the before listeners
                    const beforeResult = yield this.invokeListeners(p, GoalInvocation_1.GoalProjectListenerEvent.before);
                    updateGoalProperties(this.gi.goalEvent, beforeResult);
                    if (!!beforeResult && beforeResult.code !== 0) {
                        return beforeResult;
                    }
                    // invoke the wrapped action
                    result = yield action(p);
                }
                catch (err) {
                    throw err;
                }
                finally {
                    // invoke the after listeners
                    const afterResult = yield this.invokeListeners(p, GoalInvocation_1.GoalProjectListenerEvent.after);
                    updateGoalProperties(this.gi.goalEvent, afterResult);
                    if (!!afterResult && afterResult.code !== 0) {
                        result = afterResult;
                    }
                }
                return result;
            }));
        });
    }
    invokeListeners(p, event) {
        return __awaiter(this, void 0, void 0, function* () {
            const pli = {
                addressChannels: this.gi.addressChannels,
                preferences: this.gi.preferences,
                configuration: this.gi.configuration,
                context: this.gi.context,
                credentials: this.gi.credentials,
                id: this.gi.id,
                project: p,
                push: this.gi.goalEvent.push,
            };
            for (const lr of this.listeners) {
                const pushTest = lr.pushTest || commonPushTests_1.AnyPush;
                const events = lr.events || [GoalInvocation_1.GoalProjectListenerEvent.before, GoalInvocation_1.GoalProjectListenerEvent.after];
                if (events.includes(event) && (yield pushTest.mapping(pli))) {
                    this.gi.progressLog.write("/--");
                    this.gi.progressLog.write(`Invoking ${event} project listener: ${lr.name}`);
                    yield storeGoals_1.updateGoal(this.gi.context, this.gi.goalEvent, {
                        state: this.gi.goalEvent.state,
                        phase: lr.name,
                        description: this.gi.goalEvent.description,
                        url: this.gi.goalEvent.url,
                    });
                    const postResult = yield lr.listener(p, this.gi, event);
                    updateGoalProperties(this.gi.goalEvent, postResult);
                    this.gi.progressLog.write(`Result: ${postResult ? result_1.serializeResult(postResult) : "Success"}`);
                    this.gi.progressLog.write("\\--");
                    if (!!postResult && postResult.code !== 0) {
                        return postResult;
                    }
                }
            }
            return automation_client_1.Success;
        });
    }
}
exports.ProjectListenerInvokingProjectLoader = ProjectListenerInvokingProjectLoader;
function updateGoalProperties(goalEvent, result) {
    if (!!result) {
        goalEvent.description = !!result.description ? result.description : goalEvent.description;
        goalEvent.data = !!result.data ? result.data : goalEvent.data;
        goalEvent.externalUrls = !!result.externalUrls ? result.externalUrls : goalEvent.externalUrls;
        goalEvent.phase = !!result.phase ? result.phase : goalEvent.phase;
    }
}
//# sourceMappingURL=ProjectListenerInvokingProjectLoader.js.map