"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const GoalNameGenerator_1 = require("../GoalNameGenerator");
const GoalWithFulfillment_1 = require("../GoalWithFulfillment");
/**
 * Create a goal with basic information
 * and an action callback.
 * @deprecated use goal()
 */
function createGoal(egi, goalExecutor, options = {}) {
    const g = new GoalWithFulfillment_1.GoalWithFulfillment(Object.assign({ uniqueName: GoalNameGenerator_1.DefaultGoalNameGenerator.generateName(egi.displayName) }, egi));
    return g.with(Object.assign(Object.assign({}, options), { name: g.definition.uniqueName, goalExecutor }));
}
exports.createGoal = createGoal;
const DefaultWaitRules = {
    timeoutSeconds: 1,
    retries: 1000,
};
/**
 * Create a goal from the given executor, waiting until a condition is satisfied,
 * with a given timeout.
 * @param {EssentialGoalInfo} egi
 * @param {ExecuteGoal} goalExecutor
 * @param w rules for waiting
 * @return {Goal}
 */
function createPredicatedGoal(egi, goalExecutor, w) {
    return GoalWithFulfillment_1.goal(egi, createPredicatedGoalExecutor(egi.displayName, goalExecutor, w));
}
exports.createPredicatedGoal = createPredicatedGoal;
/**
 * Wrap provided ExecuteGoal instance with WaitRules processing
 * @param {string} uniqueName
 * @param {ExecuteGoal} goalExecutor
 * @param w rules for waiting
 * @return {ExecuteGoal}
 */
function createPredicatedGoalExecutor(uniqueName, goalExecutor, w, unref = true) {
    if (!!w.timeoutSeconds && !!w.timeoutMillis) {
        throw new Error("Invalid combination: Cannot specify timeoutSeconds and timeoutMillis: Choose one");
    }
    const waitRulesToUse = Object.assign(Object.assign({}, DefaultWaitRules), w);
    waitRulesToUse.timeoutMillis = waitRulesToUse.timeoutMillis || 1000 * w.timeoutSeconds;
    return (gi) => __awaiter(this, void 0, void 0, function* () {
        let tries = 1;
        while (true) {
            if (tries > waitRulesToUse.retries) {
                throw new Error(`Goal '${uniqueName}' timed out after max retries: ${JSON.stringify(waitRulesToUse)}`);
            }
            if (yield waitRulesToUse.condition(gi)) {
                return goalExecutor(gi);
            }
            tries++;
            automation_client_1.logger.debug("Waiting %dms for '%s'", waitRulesToUse.timeoutMillis, uniqueName);
            yield wait(waitRulesToUse.timeoutMillis, unref);
        }
    });
}
exports.createPredicatedGoalExecutor = createPredicatedGoalExecutor;
function createRetryingGoalExecutor(uniqueName, goalExecutor, retry) {
    return gi => automation_client_1.doWithRetry(() => __awaiter(this, void 0, void 0, function* () {
        const result = yield goalExecutor(gi);
        if (!!result && result.code !== 0) {
            throw new Error(`Goal '${uniqueName}' failed with non-zero code`);
        }
        return result;
    }), `Invoking goal '${uniqueName}'`, Object.assign({ log: false }, retry));
}
exports.createRetryingGoalExecutor = createRetryingGoalExecutor;
function wait(timeoutMillis, unref) {
    return new Promise(resolve => {
        const timer = setTimeout(resolve, timeoutMillis);
        if (unref) {
            timer.unref();
        }
    });
}
//# sourceMappingURL=createGoal.js.map