"use strict";
/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const _ = require("lodash");
const types_1 = require("../../../typings/types");
const PushTest_1 = require("../PushTest");
exports.ToDefaultBranch = PushTest_1.pushTest("Push to default branch", (p) => __awaiter(void 0, void 0, void 0, function* () {
    return p.push.branch === p.push.repo.defaultBranch ||
        ((!p.push.repo.defaultBranch || p.push.repo.defaultBranch.length === 0) && p.push.branch === "master");
}));
/**
 * Is this a push originated by Atomist? Note that we can't look at the committer,
 * as if a user invoked a command handler, their credentials will be used
 * @param {PushListenerInvocation} p
 * @return {boolean}
 * @constructor
 */
exports.FromAtomist = PushTest_1.pushTest("Push from Atomist", (p) => __awaiter(void 0, void 0, void 0, function* () { return p.push.after.message.includes("[atomist]"); }));
/**
 * Match on any push
 * @param {PushListenerInvocation} p
 * @constructor
 */
exports.AnyPush = PushTest_1.pushTest("Any push", () => __awaiter(void 0, void 0, void 0, function* () { return true; }));
/**
 * Return a PushTest testing for the existence of the given file
 * @param {string} path
 * @return {PushTest}
 */
function hasFile(path) {
    return PushTest_1.predicatePushTest(`HasFile(${path})`, (p) => __awaiter(this, void 0, void 0, function* () { return !!(yield p.getFile(path)); }));
}
exports.hasFile = hasFile;
/**
 * PushTest that returns true if project is non empty
 * @type {PredicatePushTest}
 */
exports.NonEmpty = PushTest_1.predicatePushTest("NonEmpty", (p) => __awaiter(void 0, void 0, void 0, function* () { return (yield p.totalFileCount()) > 0; }));
/**
 * Is there at least one file with the given extension?
 * @param {string} extension
 * @return {PredicatePushTest}
 */
function hasFileWithExtension(extension) {
    if (!extension) {
        return exports.NonEmpty;
    }
    const extensionToUse = extension.startsWith(".") ? extension : `.${extension}`;
    return PushTest_1.predicatePushTest(`HasFileWithExtension(${extensionToUse}})`, (p) => __awaiter(this, void 0, void 0, function* () { return automation_client_1.projectUtils.fileExists(p, `**/*${extensionToUse}`, () => true); }));
}
exports.hasFileWithExtension = hasFileWithExtension;
/**
 * Is this push to a non-default branch that has an open pull request?
 */
exports.IsPushToBranchWithPullRequest = PushTest_1.pushTest("Push to branch with open pull request", (p) => __awaiter(void 0, void 0, void 0, function* () {
    if (p.push.branch === p.push.repo.defaultBranch) {
        return false;
    }
    const result = yield p.context.graphClient.query({
        name: "PullRequestsForBranch",
        variables: {
            repo: p.push.repo.name,
            owner: p.push.repo.owner,
            branch: p.push.branch,
        },
    });
    const branch = _.get(result, "Repo[0].branches[0]");
    if (branch && branch.pullRequests && branch.pullRequests.some(pr => pr.state === "open")) {
        return true;
    }
    return false;
}));
/**
 * Return a push test that matches the repository owner/repo slug
 * against regular expression.
 * @param re Regular expression to match against using RegExp.test()
 * @return Push test performing the match
 */
function isRepo(re) {
    return PushTest_1.pushTest(`Project owner/name slug matches regular expression ${re.toString()}`, (pci) => __awaiter(this, void 0, void 0, function* () { return re.test(`${pci.id.owner}/${pci.id.repo}`); }));
}
exports.isRepo = isRepo;
/**
 * Return a push test that matches the repository branch
 * against regular expression.
 * @param re Regular expression to match against using RegExp.test()
 * @return Push test performing the match
 */
function isBranch(re) {
    return PushTest_1.pushTest(`Project branch matches regular expression ${re.toString()}`, (pci) => __awaiter(this, void 0, void 0, function* () { return re.test(pci.push.branch); }));
}
exports.isBranch = isBranch;
/**
 * Return a PushTest testing for the existence of the given file containing the pattern
 * @param {string} path
 * @param pattern regex to look for
 * @return {PushTest}
 */
function hasFileContaining(pattern, content = /.*/) {
    return PushTest_1.pushTest(`Project has files ${automation_client_1.toStringArray(pattern).join(", ")} with content ${content.toString()}`, (pci) => __awaiter(this, void 0, void 0, function* () {
        const files = yield pci.project.getFiles(automation_client_1.toStringArray(pattern));
        if (files.length === 0) {
            return false;
        }
        for (const file of files) {
            const fc = yield file.getContent();
            if (content.test(fc)) {
                return true;
            }
        }
        return false;
    }));
}
exports.hasFileContaining = hasFileContaining;
/**
 * Return a PushTest that checks if a certain resource provider exists in the graph
 */
function hasResourceProvider(type, name) {
    return PushTest_1.pushTest(`Workspace has resource provider of type '${type}'`, (pci) => __awaiter(this, void 0, void 0, function* () {
        switch (type) {
            case "docker":
                const dockerResult = yield pci.context.graphClient.query({
                    name: "DockerRegistryProvider",
                    variables: {
                        name,
                    },
                });
                const dockerProvider = dockerResult.DockerRegistryProvider[0];
                if (!!dockerProvider) {
                    return dockerProvider.state.name === types_1.ResourceProviderStateName.converged;
                }
                return false;
            case "npm":
            case "maven2":
                const binaryResult = yield pci.context.graphClient.query({
                    name: "BinaryRepositoryProvider",
                    variables: {
                        type: type,
                        name,
                    },
                });
                const binaryProvider = binaryResult.BinaryRepositoryProvider[0];
                if (!!binaryProvider) {
                    return binaryProvider.state.name === types_1.ResourceProviderStateName.converged;
                }
                return false;
            default:
                return false;
        }
    }));
}
exports.hasResourceProvider = hasResourceProvider;
//# sourceMappingURL=commonPushTests.js.map