"use strict";
/*
 * Copyright © 2018 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const PredicateMapping_1 = require("../PredicateMapping");
const defaultPredicateMappingCostAnalyzer_1 = require("./defaultPredicateMappingCostAnalyzer");
const PredicateMappingCostAnalyzer_1 = require("./PredicateMappingCostAnalyzer");
/**
 * Return the opposite of this predicate mapping
 */
function whenNot(t) {
    return {
        name: `not (${t.name})`,
        mapping: (pi) => __awaiter(this, void 0, void 0, function* () { return !(yield t.mapping(pi)); }),
        structure: {
            components: [t],
            compositionStyle: PredicateMapping_1.PredicateMappingCompositionStyle.Not,
        },
    };
}
exports.whenNot = whenNot;
/**
 * Wrap all these predicates in a single predicate
 * AND: Return true if all are satisfied
 * @param {PredicateMapping} predicates
 * @param analyzer analyzer to use for performance optimization
 * @return {PredicateMapping}
 */
function all(predicates, analyzer = defaultPredicateMappingCostAnalyzer_1.DefaultPredicateMappingCostAnalyzer) {
    return {
        name: predicates.map(g => g.name).join(" && "),
        mapping: (pci) => __awaiter(this, void 0, void 0, function* () { return optimizedAndEvaluation(predicates, analyzer)(pci); }),
        structure: {
            components: predicates,
            compositionStyle: PredicateMapping_1.PredicateMappingCompositionStyle.And,
        },
    };
}
exports.all = all;
/**
 * Wrap all these predicates in a single predicate
 * OR: Return true if any is satisfied
 * @param {PredicateMapping} predicates
 * @param analyzer analyzer to use for performance optimization
 * @return {PredicateMapping}
 */
function any(predicates, analyzer = defaultPredicateMappingCostAnalyzer_1.DefaultPredicateMappingCostAnalyzer) {
    return {
        name: predicates.map(g => g.name).join(" || "),
        mapping: (pci) => __awaiter(this, void 0, void 0, function* () {
            // Cannot short-circuit this
            const allResults = yield gatherResults(predicates)(pci);
            return allResults.includes(true);
        }),
        structure: {
            components: predicates,
            compositionStyle: PredicateMapping_1.PredicateMappingCompositionStyle.Or,
        },
    };
}
exports.any = any;
/**
 * Evaluate predicates for an AND, running non-expensive ones first
 * @param {Array<PredicateMapping<F>>} predicates
 * @param {PredicateMappingCostAnalyzer<F>} analyzer
 * @return {(f: F) => Promise<boolean>}
 */
function optimizedAndEvaluation(predicates, analyzer) {
    const cheap = [];
    const remaining = [];
    for (const p of predicates) {
        const cost = analyzer(p);
        if (cost !== PredicateMappingCostAnalyzer_1.ExpectedPredicateMappingCost.expensive) {
            cheap.push(p);
        }
        else {
            remaining.push(p);
        }
    }
    automation_client_1.logger.debug("Cheap: [%j], remaining: [%j]", cheap, remaining);
    return (pci) => __awaiter(this, void 0, void 0, function* () {
        const cheapResults = yield gatherResults(cheap)(pci);
        if (cheapResults.includes(false)) {
            return false;
        }
        const remainingResults = yield gatherResults(remaining)(pci);
        return !remainingResults.includes(false);
    });
}
function gatherResults(predicates) {
    return pci => {
        return Promise.all(predicates.map((pt) => __awaiter(this, void 0, void 0, function* () {
            const result = yield pt.mapping(pci);
            automation_client_1.logger.debug(`Result of PushTest '${pt.name}' was ${result}`);
            return result;
        })));
    };
}
//# sourceMappingURL=predicateUtils.js.map