/*
 * Copyright © 2019 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import {
    logger,
    Success,
} from "@atomist/automation-client";
import {
    ExecuteGoal,
    GoalInvocation,
} from "../../api/goal/GoalInvocation";
import { PushImpactListenerInvocation } from "../../api/listener/PushImpactListener";
import {
    PushImpactListenerRegisterable,
    PushImpactListenerRegistration,
    PushImpactResponse,
    toPushReactionRegistration,
} from "../../api/registration/PushImpactListenerRegistration";
import { minimalClone } from "../goal/minimalClone";
import { createPushImpactListenerInvocation } from "./createPushImpactListenerInvocation";
import { relevantCodeActions } from "./relevantCodeActions";

/**
 * Execute arbitrary code reactions against a codebase
 * @param {PushImpactListenerRegistration[]} registrations
 * @return {ExecuteGoal}
 */
export function executePushImpact(registrations: PushImpactListenerRegisterable[]): ExecuteGoal {
    return async (goalInvocation: GoalInvocation) => {
        if (registrations.length === 0) {
            return Success;
        }

        const { goalEvent, configuration, credentials, id, context } = goalInvocation;
        return configuration.sdm.projectLoader.doWithProject({
            credentials,
            id,
            context,
            readOnly: true,
            cloneOptions: minimalClone(goalEvent.push, { detachHead: true }),
        }, async project => {
            const cri: PushImpactListenerInvocation = await createPushImpactListenerInvocation(goalInvocation, project);
            const regs = registrations.map(toPushReactionRegistration);
            const relevantCodeReactions: PushImpactListenerRegistration[] = await relevantCodeActions<PushImpactListenerRegistration>(regs, cri);
            logger.debug("Will invoke %d eligible code reactions of %d to %j: [%s] of [%s]",
                relevantCodeReactions.length, registrations.length, cri.id,
                relevantCodeReactions.map(a => a.name).join(),
                regs.map(a => a.name).join());
            const allReactions: any[] = await Promise.all(relevantCodeReactions
                .map(reactionReg => reactionReg.action(cri)));
            const result = {
                code: allReactions.includes(PushImpactResponse.failGoals) ? 1 : 0,
                requireApproval: allReactions.includes(PushImpactResponse.requireApprovalToProceed),
            };
            logger.debug("PushReaction responses are %j, result=%j", allReactions, result);
            return result;
        });
    };
}
