import { Types, PluginFunction } from '@graphql-codegen/plugin-helpers';
import { GraphQLSchema } from 'graphql';
import { RawTypesConfig } from '@graphql-codegen/visitor-plugin-common';
import { TsVisitor } from './visitor';
import { TsIntrospectionVisitor } from './introspection-visitor';
import { AvoidOptionalsConfig } from './types';
export * from './typescript-variables-to-object';
export * from './visitor';
export { AvoidOptionalsConfig } from './types';
export interface TypeScriptPluginConfig extends RawTypesConfig {
    /**
     * @name avoidOptionals
     * @type boolean
     * @description This will cause the generator to avoid using TypeScript optionals (`?`) on types,
     * so the following definition: `type A { myField: String }` will output `myField: Maybe<string>`
     * instead of `myField?: Maybe<string>`.
     * @default false
     *
     * @example Override all definition types
     * ```yml
     * generates:
     * path/to/file.ts:
     *  plugins:
     *    - typescript
     *  config:
     *    avoidOptionals: true
     * ```
     *
     * @example Override only specific definition types
     * ```yml
     * generates:
     * path/to/file.ts:
     *  plugins:
     *    - typescript
     *  config:
     *    avoidOptionals:
     *      inputValue: true
     * ```
     *
     */
    avoidOptionals?: boolean | AvoidOptionalsConfig;
    /**
     * @name constEnums
     * @type boolean
     * @description Will prefix every generated `enum` with `const`, you can read more
     * about const enums {@link https://www.typescriptlang.org/docs/handbook/enums.html|here}.
     * @default false
     *
     * @example
     * ```yml
     * generates:
     * path/to/file.ts:
     *  plugins:
     *    - typescript
     *  config:
     *    constEnums: true
     * ```
     */
    constEnums?: boolean;
    /**
     * @name enumsAsTypes
     * @type boolean
     * @description Generates enum as TypeScript `type` instead of `enum`. Useful it you wish to genereate `.d.ts` declartion file instead of `.ts`
     * @default false
     *
     * @example
     * ```yml
     * generates:
     * path/to/file.ts:
     *  plugins:
     *    - typescript
     *  config:
     *    enumsAsTypes: true
     * ```
     */
    enumsAsTypes?: boolean;
    /**
     * @name immutableTypes
     * @type boolean
     * @description Generates immutable types by adding `readonly` to properties and uses `ReadonlyArray`.
     * @default false
     *
     * @example
     * ```yml
     * generates:
     * path/to/file.ts:
     *  plugins:
     *    - typescript
     *  config:
     *    immutableTypes: true
     * ```
     */
    immutableTypes?: boolean;
    /**
     * @name maybeValue
     * @type string
     * @description Allow to override the type value of `Maybe`.
     * @default T | null
     *
     * @example Allow undefined
     * ```yml
     * generates:
     * path/to/file.ts:
     *  plugins:
     *    - typescript
     *  config:
     *    maybeValue: T | null | undefined
     * ```
     */
    maybeValue?: string;
    /**
     * @name noExport
     * @type boolean
     * @description Set the to `true` in order to generate output without `export` modifier.
     * This is useful if you are generating `.d.ts` file and want it to be globally available.
     * @default false
     *
     * @example Disable all export from a file
     * ```yml
     * generates:
     * path/to/file.ts:
     *  plugins:
     *    - typescript
     *  config:
     *    noExport: true
     * ```
     */
    noExport?: boolean;
}
export declare const plugin: PluginFunction<TypeScriptPluginConfig>;
export declare function includeIntrospectionDefinitions(schema: GraphQLSchema, documents: Types.DocumentFile[], config: TypeScriptPluginConfig): string[];
export { TsVisitor, TsIntrospectionVisitor };
